# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import base64
import datetime
import hashlib
import hmac
import math
import os
import sys

import ndjson
import json
from enum import Enum
from jsonschema import validate

import azdata.cli.core.deploy as util
from azdata.cli.core.configuration import Configuration
from azdata.cli.commands.arc.azure.ad_auth_util import acquire_token
from azdata.cli.commands.arc.common_util import (
    get_config_file,
    get_output_file,
    get_resource_uri,
    write_file,
    write_output_file
)
from azdata.cli.commands.arc.azure import constants as azure_constants
from azdata.cli.commands.arc.azure.models.spn import Spn
from azdata.cli.commands.arc import export_instance_properties as instance_properties
from azdata.cli.core.exceptions import CliError, RequestTimeoutError, ServerError, http_status_codes
from azdata.cli.core.deploy import display
from azdata.cli.core.serialization import SanitizerRule
from azdata.cli.core.logging import get_logger

from knack.prompting import (
    prompt,
    prompt_y_n,
    NoTTYException
)

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util.retry import Retry
from urllib3.exceptions import NewConnectionError, MaxRetryError, TimeoutError
from http import HTTPStatus

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5

log = get_logger(__name__)

############################################################################
# Export Types
############################################################################


class ExportType(Enum):
    metrics = 'metrics'
    logs = 'logs'
    usage = 'usage'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @staticmethod
    def list():
        return list(map(lambda c: c.value, ExportType))


# #############################################################################
# Metric
# #############################################################################


class MetricsDataStructure(Enum):
    region_key = 'region'
    resource_id_key = 'resource_id'
    metrics_key = 'metrics'


class AzureResource(object):
    hostname = ''
    location = ''
    resource_id = ''
    instance_type = ''

    def __init__(self, **entries):
        self.__dict__.update(entries)


class AzureMetrics(object):
    location = ''
    resource_id = ''
    metrics = ''

    def __init__(self, **entries):
        self.__dict__.update(entries)


class MetricsQuery(object):
    def __init__(self, instance_type, metric, db, query):
        self.instance_type = instance_type
        self.metric = metric
        self.db = db
        self.query = query


# To support upload to Azure metrics, queries must return columns 'time' and 'count'.
# Aggregation columns 'sum', 'min', and 'max' are also supported by Azure metrics.
# These aggregations are useful because the Azure metrics UI only has 1 minute precision.
# Therefore queries should group by 1 minute. But in most cases, the data in InfluxDB
# is more precise (telegraf reports every 10 seconds). This gives us different ways of
# aggregating the data into a single data point per minute. The 'sum' aggregate is
# the most useful, since it combines with 'count' to produce 'avg' in the metrics UI.
# We also send 'min' and 'max' aggregates to indicate the range of values
# we saw during each minute.

SQL_METRICS_QUERIES = []

SQL_METRICS_QUERIES.append(MetricsQuery(
    'sql',
    'CPU Usage',
    'telegraf',
    'SELECT min("container_cpu_usage"), max("container_cpu_usage"), sum("container_cpu_usage"), count("container_cpu_usage") FROM (SELECT "cpu_usage_nanocores"/1000000000.0 AS "container_cpu_usage" FROM "kubernetes_pod_container" WHERE "pod_name" =~ /^{resource_name}-\d+$/ AND "container_name" = \'arc-sqlmi\' AND time > now() - 28m) group by time(1m)'))

SQL_METRICS_QUERIES.append(MetricsQuery(
    'sql',
    'Memory Usage',
    'telegraf',
    'SELECT min("container_memory_usage_bytes"), max("container_memory_usage_bytes"), sum("container_memory_usage_bytes"), count("container_memory_usage_bytes") FROM (SELECT "memory_usage_bytes" AS "container_memory_usage_bytes" FROM "kubernetes_pod_container" WHERE "pod_name" =~ /^{resource_name}-\d+$/ AND "container_name" = \'arc-sqlmi\' AND time > now() - 28m) group by time(1m)'))

SQL_METRICS_QUERIES.append(MetricsQuery(
    'sql',
    'Transactions/second',
    'collectd_db',
    'SELECT min(tran), max(tran), sum(tran), count(tran) from (SELECT non_negative_derivative("value", 1s) as tran FROM "dbi_perf_stats_value" WHERE ("type_instance" = \'Transactions_sec-_Total\' AND "host" =~ /^{resource_name}-\d+$/) AND time > now() -28m) group by time(1m)'))

POSTGRES_METRICS_QUERIES = []
POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'CPU Usage',
    'telegraf',
    'SELECT min("container_cpu_usage"), max("container_cpu_usage"), sum("container_cpu_usage"), count("container_cpu_usage") FROM (SELECT "cpu_usage_nanocores"/1000000000.0 AS "container_cpu_usage" FROM "kubernetes_pod_container" WHERE "namespace" = \'{resource_namespace}\' AND "pod_name" =~ /^{resource_name}[mprs]-\d+$/ AND "container_name" = \'postgres\' AND time > now() - 28m) GROUP BY time(1m), "pod_name"'))

POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'Memory Usage',
    'telegraf',
    'SELECT min("container_memory_usage_bytes"), max("container_memory_usage_bytes"), sum("container_memory_usage_bytes"), count("container_memory_usage_bytes") FROM (SELECT "memory_usage_bytes" AS "container_memory_usage_bytes" FROM "kubernetes_pod_container" WHERE "namespace" = \'{resource_namespace}\' AND "pod_name" =~ /^{resource_name}[mprs]-\d+$/ AND "container_name" = \'postgres\' AND time > now() - 28m) GROUP BY time(1m), "pod_name"'))

POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'Rows Fetched/second',
    'telegraf_postgres',
    'SELECT min(tup), max(tup), sum(tup), count(tup) from (SELECT non_negative_derivative("tup_fetched", 1s) as tup FROM "postgresql" WHERE "cluster_name" = \'{resource_name}\' AND "namespace" = \'{resource_namespace}\' AND time > now() - 28m GROUP BY "host", "db") GROUP BY time(1m), "host", "db"'))

POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'Rows Inserted/second',
    'telegraf_postgres',
    'SELECT min(tup), max(tup), sum(tup), count(tup) from (SELECT non_negative_derivative("tup_inserted", 1s) as tup FROM "postgresql" WHERE "cluster_name" = \'{resource_name}\' AND "namespace" = \'{resource_namespace}\' AND time > now() - 28m GROUP BY "host", "db") GROUP BY time(1m), "host", "db"'))

POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'Rows Updated/second',
    'telegraf_postgres',
    'SELECT min(tup), max(tup), sum(tup), count(tup) from (SELECT non_negative_derivative("tup_updated", 1s) as tup FROM "postgresql" WHERE "cluster_name" = \'{resource_name}\' AND "namespace" = \'{resource_namespace}\' AND time > now() - 28m GROUP BY "host", "db") GROUP BY time(1m), "host", "db"'))

POSTGRES_METRICS_QUERIES.append(MetricsQuery(
    'postgres',
    'Rows Deleted/second',
    'telegraf_postgres',
    'SELECT min(tup), max(tup), sum(tup), count(tup) from (SELECT non_negative_derivative("tup_deleted", 1s) as tup FROM "postgresql" WHERE "cluster_name" = \'{resource_name}\' AND "namespace" = \'{resource_namespace}\' AND time > now() - 28m GROUP BY "host", "db") GROUP BY time(1m), "host", "db"'))


# #############################################################################
# Metric export/upload functions --
# #############################################################################

ERROR_IGNORE = 'should not be older than 30 minutes and not more than 4 minutes in the future'
"""
Metric error msg could be ignored 
"""


def metrics_export(client, resources, data_controller):
    """
    Export metrics from data controller.
    :param client: controller client
    :param resources: custom resource
    :param data_controller:
    :return: metrics data
    """
    resources_metrics = []

    for r in resources:
        metrics = _get_metrics_from_backend(client, r)

        if len(metrics) == 0:
            continue
        resource_uri = get_resource_uri(r, data_controller)
        if resource_uri is None:
            continue
        else:
            metrics_exported = _convert_to_metrics_format(
                resource_uri, data_controller['location'], metrics)

            if metrics_exported:
                resources_metrics.append(metrics_exported)

    return resources_metrics


def _post_metrics(url, body, headers):
    response = requests.post(url, data=body, headers=headers)

    log.info("Metrics upload reponse header: {}".format(response.headers))
    try:
        response.raise_for_status()
    except HTTPError as ex:
        if response.status_code == http_status_codes.request_timeout:
            raise RequestTimeoutError(ex)
        elif response.status_code >= 500:
            raise ServerError(ex)
        else:
            raise

    return response


def metrics_upload(metrics):

    if metrics is None:
        display('No metrics need to upload.')
    else:
        display('\n')
        for data in metrics:
            resource_id_value = data[MetricsDataStructure.resource_id_key.value]
            region_value = data[MetricsDataStructure.region_key.value]
            metrics = data['metrics']
            if not resource_id_value or not metrics or not region_value:
                CliError(
                    '{} is not valid. Please check the file contains {} and '
                    'metrics fields. Or export the data to file and try again.')

            current_time = datetime.datetime.utcnow()

            filtered_metrics = list(filter(lambda metric: (current_time - datetime.datetime.strptime(metric["time"], "%Y-%m-%dT%H:%M:%SZ")).total_seconds()/60 <= 30, metrics))
            if filtered_metrics is None or not filtered_metrics:
                print("The metrics data are older than 30 minutes for {}, please export and upload again.".format(resource_id_value))
                continue
            log.info("Metrics data in file has {} records. {} are in last 30 mins".format(len(metrics), len(filtered_metrics)))
            display('Azure resource_id: {}'.format(resource_id_value))

            # Setup request header and url
            headers = _set_header()
            url = _set_url(region_value, resource_id_value)
            res = util.retry(
                lambda: _post_metrics(url, body=ndjson.dumps(filtered_metrics), headers=headers),
                retry_count=CONNECTION_RETRY_ATTEMPTS,
                retry_delay=RETRY_INTERVAL,
                retry_method="post metrics data",
                retry_on_exceptions=(NewConnectionError, MaxRetryError, TimeoutError, RequestTimeoutError, ServerError))

            display('Metrics upload pushed {} data points successfully.\n'.format(len(filtered_metrics)))


# Given a resource registered in Azure, returns the k8s namespace and name.
def _parse_resource(resource):
    name = resource[instance_properties.INSTANCE_NAME]
    namespace = resource[instance_properties.NAMESPACE]

    # Resources deployed outside the controller's namespace are named in the
    # format 'namespace_name'
    parts = name.split('_')
    if len(parts) == 2:
        name = parts[1]
    elif len(parts) > 2:
        raise Exception(
            "Cannot parse resource '{}'. Acceptable formats are 'namespace_name' or 'name'.".format(
                name))
    return namespace, name


def _get_queries(resource_type):
    if resource_type == azure_constants.INSTANCE_TYPE_SQL:
        return SQL_METRICS_QUERIES
    elif resource_type == azure_constants.INSTANCE_TYPE_POSTGRES:
        return POSTGRES_METRICS_QUERIES
    else:
        return None


def _get_metrics_from_backend(client, resource):
    resource_type = azure_constants.RESOURCE_TYPE_FOR_KIND[resource[instance_properties.KIND]]
    queries = _get_queries(resource_type)
    if queries is None:
        display(
            "Resource type '{}' is not supported. Please check the resources in your cluster".format(
                resource_type))
        return []

    resource_namespace, resource_name = _parse_resource(resource)
    all_metrics = []
    display("Collecting metrics for {} instance: '{}.{}'".format(
        resource_type.upper(), resource_namespace, resource_name))
    for query in queries:
        # Get result from InfluxDB
        request_body = {"db": query.db, "query": query.query.format(
            resource_namespace=resource_namespace, resource_name=resource_name)}
        try:
            # response = client.get_metrics(**kwargs)
            response = client.apis.controller.metrics_post(body=request_body)

            if response:
                body_content = response.get('body')
                if body_content is None:
                    client.stderr(
                        "\tFailed to get metric: '{}' with error: '{}'".format(
                            query.metric, response))
                    continue

                body = json.loads(body_content)
                for result in body.get('results', []):
                    metrics = _convert_to_azure_custom_metrics(result,
                                                               resource_type,
                                                               query.metric)
                    if metrics:
                        all_metrics.extend(metrics)
                        display("\t Successfully got metric: '{}'".format(
                            query.metric))
                    else:
                        display("\t No metric: '{}' found".format(query.metric))
        except Exception as e:
            client.stderr(
                "\tFailed to get metric: '{}' with exception: '{}'".format(
                    query.metric, e))
            continue
    display('\n')

    return all_metrics


def _set_header():
    """
    Authenticate using service principal w/ key.
    """

    headers = {'Authorization': 'Bearer ' + acquire_token(azure_constants.AZURE_METRICS_SCOPE),
               'Content-Type': 'application/x-ndjson'}
    return headers


def _set_url(location, resource_id):
    url = "https://{}.monitoring.azure.com{}/metrics".format(location,
                                                             resource_id)
    return url

# Converts the result of an InfluxDB query to a list of AzureMetricsDocuments.
# https://docs.microsoft.com/en-us/rest/api/monitor/metrics%20(data%20plane)/create#azuremetricsdocument
#
# Example result of an InfluxDB query that groups by 'pod.name':
# 'series': [
#     {
#         'columns': ['time', 'sum', 'count'],
#         'values': [['2019-10-25T04:30:00Z', 20, 1],
#                    ['2019-10-25T04:40:00Z', 21, 1]],
#         'tags': {'pod.name': 'p-0' }
#     },
#     {
#         'columns': ['time', 'sum', 'count'],
#         'values': [['2019-10-25T04:30:00Z', 30, 1],
#                    ['2019-10-25T04:40:00Z', 31, 1]],
#         'tags': {'pod.name': 'p-1' }
#     }
# ]
#
# The example would be converted into 2 AzureMetricsDocuments,
# one for each timestamp. The first one would look like this:
# {
#     'time': '2019-10-25T04:30:00Z',
#     'data': {
#         'baseData': {
#             'dimNames': ['pod.name'],
#             'series': [
#                 {
#                     'dimValues': ['p-0'],
#                     'sum': 20,
#                     'count': 1
#                 },
#                 {
#                     'dimValues': ['p-1'],
#                     'sum': 30,
#                     'count': 1
#                 }
#             ]
#         }
#     }
# },

def _convert_to_azure_custom_metrics(result, resource_type, metrics_type):
    if resource_type == azure_constants.INSTANCE_TYPE_SQL:
        namespace = azure_constants.METRICS_SQL_NAMESPACE
    elif resource_type == azure_constants.INSTANCE_TYPE_POSTGRES:
        namespace = azure_constants.METRICS_POSTGRES_NAMESPACE
    else:
        return None

    all_series = result.get('series', [])
    if len(all_series) == 0:
        return []

    # Assumption: all series have the same columns and tag keys
    columns = all_series[0]['columns']
    tag_keys = list(all_series[0].get("tags", {}).keys())

    for req_col in ["time", "count"]:
        if req_col not in columns:
            raise CliError(
                "Metric is missing required column '{}'".format(req_col))

    time_col = columns.index("time")
    count_col = columns.index("count")

    # Collect values and tags across all series
    values = []
    for series in all_series:
        tag_values = list(series.get("tags", {}).values())
        for value in (v for v in series["values"] if v[count_col] > 0):
            values.append(value + tag_values)

    # Sort by time so we can group values with the same timestamp into a
    # single metric
    values.sort(key=lambda v: v[time_col])
    metrics = []
    metric = None
    for value in values:
        if metric is None or metric["time"] != value[time_col]:
            if metric is not None:
                metrics.append(metric)

            # Timestamp has changed, create a new AzureMetricsDocument
            metric = {
                "time": value[time_col],
                "data": {
                    "baseData": {
                        "metric": metrics_type,
                        "namespace": namespace,
                        "series": []
                    }
                }
            }

            if len(tag_keys) > 0:
                metric["data"]["baseData"]["dimNames"] = tag_keys

        # Add a series to the AzureMetricsDocument
        series = {}
        for i, col in enumerate(columns):
            if i != time_col:
                series[col] = 0 if value[i] is None else round(value[i], 2)

        if len(tag_keys) > 0:
            series["dimValues"] = value[len(columns):]

        metric["data"]["baseData"]["series"].append(series)

    if metric is not None:
        metrics.append(metric)

    return metrics


def _convert_to_metrics_format(resource_id_value, azure_region, metrics):
    result = ''
    if metrics and resource_id_value:
        result = {
            MetricsDataStructure.region_key.value: azure_region,
            MetricsDataStructure.resource_id_key.value: resource_id_value,
            MetricsDataStructure.metrics_key.value: metrics
        }
    return result


# #############################################################################
# Log export/upload functions
# #############################################################################
LOGS_CONFIG_FILENAME = 'logs-config.json'
"""
Log config file name
"""

LOGS_DEFAULT_DURATION = 14
"""
Default log export duration
"""

LOGS_FILE_PATH = '/var/opt/mssql/log/errorlog'
"""
filepath of sql error logs
"""

SQL_MI_CONTAINER_NAME = 'arc-sqlmi'
"""
container name of SQL MI
"""

LOGS_MAXIMUM_POST_SIZE = 25 * 1024 * 1024
"""
Maximum post size to logs analytics workspace is 30M bytes. 5M for buffer
"""

LOGS_FILE_SIZE = 512 * 1024 * 1024
"""
Maximum log export file size
"""

LOGS_COLLECTION_COUNT_FROM_SERVER = 20
"""
Maximum logs collection count returned from server
"""

LOGS_BACKOFF_FACTOR = 1
"""
Upload backoff factor
"""

LOGS_POST_RETRY_CAP = 16
"""
Retry cap is 16 times for log analytics post request
"""

WORKSPACE_ENV_KEYS = {
    "customer_id": "WORKSPACE_ID",
    "shared_key": "WORKSPACE_SHARED_KEY"
}
"""
Environment variables' name of workspace for log upload
"""


def _generate_export_file_name(file_path, index):
    file_name_with_ext = os.path.basename(file_path)  # eds_report.csv
    dir_name = os.path.dirname(file_path)
    file_name = os.path.splitext(file_name_with_ext)[0]
    filename_suffix = os.path.splitext(file_name_with_ext)[1]
    base_filename = file_name + '-' + str(index)

    export_file_name = os.path.join(dir_name, base_filename + filename_suffix)
    return export_file_name


def _write_logs(data_controller, logs, file_name, resource, timestamp):

    resource_uri = get_resource_uri(resource, data_controller)
    if resource_uri is not None:
        content = []
        for log in logs:
            result = _convert_to_logs_format(
                log, 
                resource[instance_properties.INSTANCE_NAME],
                azure_constants.RESOURCE_TYPE_FOR_KIND[resource[instance_properties.KIND]], 
                resource_uri)
            content.append(result)
                
    write_file(file_name, content, ExportType.logs.value, timestamp)


def logs_upload(client, logs, data_timestamp, customer_id, shared_key):
    import zlib
    import base64

    logs_upload_success = True
    for log in logs:
        unzip = str(zlib.decompress(base64.b64decode(log['logs']), -zlib.MAX_WBITS), 'utf-8')
        log_table_name = log['instance_type'] + '_logs'
        if not _post_logs_to_logs_analytics(customer_id, 
                                            shared_key,
                                            log['instance_name'],
                                            unzip,
                                            log_table_name,
                                            log['instance_type'],
                                            log['resource_uri']):
            logs_upload_success = False

    # only update the last_logs_timestamp when logs for all the instance upload success
    # TODO this could be improved by adding the timestamp to each instance
    if logs_upload_success:
        update_upload_status_file(max_event_sequence_id=None, log_data_timestamp=data_timestamp)


def logs_export(client, file_path, resources, data_controller):
    content = []

    # End log collection 2 minutes ago. A 2 minute buffer is used because
    # logs emitted right now may not have been ingested by Elasticsearch yet.
    end_time = datetime.datetime.utcnow() - datetime.timedelta(minutes=2)

    # Start log collection at the last uploaded time,
    # if one exists and is more recent than the default duration.
    start_time = end_time - datetime.timedelta(
        days=LOGS_DEFAULT_DURATION)
    last_upload_time, schema_valid = _get_upload_status_value("last_logs_timestamp")
    if schema_valid and last_upload_time and last_upload_time > start_time:
        start_time = last_upload_time

    # Divide the time range into 1 day intervals
    interval = datetime.timedelta(days=1)

    # file_index is for all instances
    #
    file_index = 0
    files_list = []

    # Export file has size limit
    #
    for resource in resources:
        instance_type = azure_constants.RESOURCE_TYPE_FOR_KIND[resource[instance_properties.KIND]]
        instance_name = resource[instance_properties.INSTANCE_NAME]

        if instance_type == azure_constants.INSTANCE_TYPE_SQL:
            body = {
                "termQueries": {
                    "custom_resource_name.keyword": instance_name,
                    "kubernetes_container_name.keyword": SQL_MI_CONTAINER_NAME,
                    "log_filename.keyword": LOGS_FILE_PATH}}
        elif instance_type == azure_constants.INSTANCE_TYPE_POSTGRES:
            resource_namespace, resource_name = _parse_resource(resource)
            body = {
                "termQueries": {
                    "kubernetes_namespace.keyword": resource_namespace,
                    "custom_resource_name.keyword": resource_name,
                    "service_name.keyword": "postgres"
                }
            }
        else:
            client.stderr(
                '\tInstance: "{}" resource type: "{}" is not a supported instance type'.format(
                    instance_name, instance_type))
            continue

        days = math.ceil((end_time - start_time) / datetime.timedelta(days=1))
        display(
            'Start retrieving last {} days of logs for {} instance "{}", this may take some time...'.format(
                days,
                instance_type,
                instance_name))

        # Query each interval
        t_start = start_time

        while t_start < end_time:
            body["startTime"] = t_start.strftime('%Y%m%d%H%M%S')
            t_end = min(t_start + interval, end_time)
            body["endTime"] = t_end.strftime('%Y%m%d%H%M%S')

            try:
                curlogs = client.apis.controller.logs_post(body)
                logs = []
                logs_count = 0
                upload_count = 0
                
                for curlog in curlogs:
                    count = curlog.get('Count')
                    if count > 0:
                        logs.append(curlog.get('Logs'))
                        logs_count += count
                        upload_count += 1
            except Exception as e:
                client.stderr(e)
                continue
            display("\tCollected {:d} logs from UTC '{}' to '{}'".format(
                logs_count, t_start.strftime('%Y-%m-%d %H:%MZ'),
                t_end.strftime('%Y-%m-%d %H:%MZ')))

            if logs_count > 0:
                files_count_per_interval = math.ceil(upload_count/LOGS_COLLECTION_COUNT_FROM_SERVER)
                i = 0
                
                while i < files_count_per_interval:
                    file_name = _generate_export_file_name(file_path, file_index)
                    logs_index = i * LOGS_COLLECTION_COUNT_FROM_SERVER
                    _write_logs(data_controller, logs[logs_index: logs_index + LOGS_COLLECTION_COUNT_FROM_SERVER], file_name, resource, end_time.strftime('%Y%m%d%H%M%S'))
                    files_list.append(file_name)
                    file_index += 1
                    i += 1
            t_start = t_end


    # return files_list for all instances
    #
    return files_list, end_time.strftime('%Y%m%d%H%M%S')


"""
https://docs.microsoft.com/en-us/azure/azure-monitor/platform/data-collector-api#data-limites
limit 2: Maximum of 30 MB per post to Azure Monitor Data Collector API, so we have to slice the data
Each log size is different, starts with the average log count per post pre-calculated
"""


def _slice_logs_to_post_size(logs, start, avg_logs_count_per_post):
    count = avg_logs_count_per_post
    while count > 0:
        logs_slice = logs[start: start + count]
        data_string = json.dumps(logs_slice)
        data_size_byte = sys.getsizeof(data_string)
        if data_size_byte < LOGS_MAXIMUM_POST_SIZE:
            break
        else:
            count -= 1
    return count


def _convert_to_logs_format(logs, instance, instance_type, resource_uri):
    result = ''
    if logs and instance and instance_type:
        result = {
            "instance_name": instance,
            "instance_type": instance_type,
            "resource_uri": resource_uri,
            "logs": logs
        }
    return result


# #############################################################################
# Log workspace functions
# #############################################################################
def _get_log_workspace_credentials_from_env(client):
    """
    Get the shared key and customer id for the given
    client's log workspace
    """
    customer_id = None
    shared_key = None
    get_workspace_from_env = True

    # First try to get workspace info from ENV
    for workspace_env in WORKSPACE_ENV_KEYS.values():
        if workspace_env not in os.environ or os.environ[workspace_env] is None:
            get_workspace_from_env = False
            break
    if get_workspace_from_env:
        customer_id = os.environ[WORKSPACE_ENV_KEYS["customer_id"]]
        shared_key = os.environ[WORKSPACE_ENV_KEYS["shared_key"]]
    else:
        # Prompt user to provide workspace info
        succeed = False
        while not succeed:
            display(
                'Please provide Azure Log Analytics workspace information')
            customer_id = prompt('Workspace ID: ')
            shared_key = prompt('Primary Key: ')
            try:
                succeed = _test_log_workspace_key(customer_id, shared_key)
            except Exception as e:
                client.stderr(
                    'Can\'t upload to the workspace: "{} Exception: {}". Please try again'.format(
                        customer_id, e))
                display('\n')

        _store_log_workspace_credentials_in_env(customer_id, shared_key)
    return customer_id, shared_key


def _update_last_upload_logs_timestamp(config_file_name, last_logs_timestamp=None):
    """
    Writes or updates the last logs timestamp
    :param config_file_name: string
    :param last_logs_timestamp: datetime
    """
    content = {
        "last_logs_timestamp": last_logs_timestamp
    }
    try:
        with open(config_file_name, 'w', encoding='utf-8') as conf_file:
            json.dump(content, conf_file)
    except BaseException:
        raise ValueError("Could not save logs timestamp.")


def _store_log_workspace_credentials_in_env(customer_id, shared_key):
    """
    Stores logs workspace credentials in environment
    :param customer_id: string
    :param shared_key: string
    """
    if customer_id and shared_key:
        os.environ[WORKSPACE_ENV_KEYS["customer_id"]] = customer_id
        os.environ[WORKSPACE_ENV_KEYS["shared_key"]] = shared_key
    else:
        raise ValueError("Failed to store one or more of the following: {}".format(list(WORKSPACE_ENV_KEYS)))


def _build_log_request_header(customer_id, shared_key, content_length, log_type,
                              resource_uri):
    rfc1123date = datetime.datetime.utcnow().strftime(
        '%a, %d %b %Y %H:%M:%S GMT')
    method = 'POST'
    content_type = 'application/json'

    signature = _build_log_request_signature(customer_id, shared_key,
                                             rfc1123date, content_length,
                                             method, content_type,
                                             azure_constants.API_LOG)
    headers = {
        'content-type': content_type,
        'Authorization': signature,
        'Log-Type': log_type,
        'x-ms-date': rfc1123date,
        'x-ms-AzureResourceId': resource_uri
    }
    return headers


def _build_log_request_uri(customer_id):
    return 'https://' + customer_id + '.ods.opinsights.azure.com' + \
        azure_constants.API_LOG + '?api-version=2016-04-01'


# Build the API signature
def _build_log_request_signature(customer_id, shared_key, date, content_length,
                                 method, content_type, resource):
    x_headers = 'x-ms-date:' + date
    string_to_hash = method + "\n" + str(
        content_length) + "\n" + content_type + "\n" + x_headers + "\n" + resource
    bytes_to_hash = bytes(string_to_hash, 'UTF-8')
    decoded_key = base64.b64decode(shared_key)
    encoded_hash = base64.b64encode(
        hmac.new(decoded_key, bytes_to_hash, digestmod=hashlib.sha256).digest())
    encoded_hash = encoded_hash.decode("utf-8")
    authorization = "SharedKey {}:{}".format(customer_id, encoded_hash)
    return authorization


def _test_log_workspace_key(customer_id, shared_key):
    uri = _build_log_request_uri(customer_id)
    body = '[{"Property 1": "test1"}]'
    table_name = 'test_workspace_logs'
    headers = _build_log_request_header(customer_id, shared_key, len(body),
                                        table_name, uri)
    succeed = False
    try:
        response = requests.post(uri, data=body, headers=headers)
        if 200 <= response.status_code <= 299:
            succeed = True
    finally:
        return succeed


def _post_logs_to_logs_analytics(customer_id, shared_key, instance_name,
                                 body, log_type, instance_type, resource_uri):
    uri = _build_log_request_uri(customer_id)
    headers = _build_log_request_header(customer_id, shared_key, len(body),
                                        log_type, resource_uri)

    response = _requests_retry_session().post(uri, data=body, headers=headers)
    if 200 <= response.status_code <= 299:
        display(
            '\tSuccessfully upload "{}" records for resource type "{}", instance: "{}\'s" log to table: "{}" '.format(
                len(json.loads(body)),
                instance_type,
                instance_name,
                log_type))
        upload_success = True
    else:
        raise Exception(
            '\tFail to upload "{}" to table: "{}" with status code: "{}" and error msg: "{}"'.format(
                instance_name,
                log_type,
                response.status_code,
                response.text))

    return upload_success


# https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html
# Explains how retry works
def _requests_retry_session(
        retries=LOGS_POST_RETRY_CAP,
        backoff_factor=LOGS_BACKOFF_FACTOR,
        session=None, ):
    """
    Get a retry request session
    :arg:
      - retries: number of retries
      - backoff_factor: for exponential backoff
      - status_forcelist: server error code list
        Now retry on three cases based on the following doc:
        https://docs.microsoft.com/en-us/azure/azure-monitor/platform/data-collector-api#return-codes
      - session: https request retry session
    :return: retry session
    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=(
            HTTPStatus.TOO_MANY_REQUESTS, HTTPStatus.INTERNAL_SERVER_ERROR,
            HTTPStatus.SERVICE_UNAVAILABLE),
        method_whitelist=frozenset(['POST']),
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('https://', adapter)
    return session

# #############################################################################
# Export data JSON schema
# #############################################################################


EXPORT_DATA_JSON_SCHEMA = {
    "type": "object",
    "properties": {
        "exportType": {"type": "string"},
        "dataTimestamp": {"type": "string"},
        "dataController": {
            "type": "object",
            "properties": {
                "instanceName": {"type": "string"},
                "subscriptionId": {"type": "string"},
                "resourceGroupName": {"type": "string"},
                "location": {"type": "string"},
                "publicKey": {"type": "string"},
                "k8sRaw": {"type": "object"},
            },
            "required": [
                "instanceName",
                "subscriptionId",
                "resourceGroupName",
                "location",
                "publicKey",
                "k8sRaw",
            ],
        },
        "deletedInstances": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "uid": {"type": "string"},
                    "instanceName": {"type": "string"},
                    "instanceNamespace": {"type": "string"},
                    "kind": {"type": "string"},
                },
                "required": [
                    "uid",
                    "instanceName",
                    "instanceNamespace",
                    "kind",
                ],
            },
        },
        "instances": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "instanceName": {"type": "string"},
                    "instanceNamespace": {"type": "string"},
                    "kind": {"type": "string"},
                    "creationTimestamp": {"type": "string"},
                    "k8sRaw": {"type": "object"}
                },
                "required": [
                    "instanceName",
                    "instanceNamespace",
                    "kind",
                    "creationTimestamp",
                    "k8sRaw"
                ],
            },
        },
        "data": {"type": "array"},
    },
    "required": [
        "exportType",
        "dataTimestamp",
        "dataController",
        "deletedInstances",
        "instances",
    ],
}

EXPORT_FILE_DICT_KEY = {'exportType', 'data', 'dataTimestamp'}
"""
Export file dictionary keys
"""

EXPORT_SANITIZERS = []
"""
Export file sanitizer rules.  
example: SanitizerRule(".*externalEndpoint")
"""


# #############################################################################
# Upload status functions
# #############################################################################
UPLOAD_STATUS_FILENAME = 'upload-status.json'
"""
Upload status file name
"""

def _get_upload_status_value(value_type):
    """
    :param value_type: string value for "last_logs_timestamp" or "max_event_sequence_id"
    Returns the specified value from the upload status file and schema_valid, which represents 
    whether or not the file schema was valid when checked.
    """
    try:
        upload_status_file_path, schema_valid = get_upload_status_file_path(UPLOAD_STATUS_FILENAME)
        if os.path.exists(upload_status_file_path):
            with open(upload_status_file_path, encoding='utf-8') as upload_status_file:
                upload_status_json = json.load(upload_status_file)
                if upload_status_json:

                    if value_type == "max_event_sequence_id":
                        value = upload_status_json.get('usage').get('max_event_sequence_id')
                    if value_type == "last_logs_timestamp":
                        str_timestamp = upload_status_json.get('logs').get('last_logs_timestamp')
                        value = datetime.datetime.strptime(str_timestamp, '%Y%m%d%H%M%S')

                    return value, schema_valid
    except BaseException:
        raise ValueError("Could not retrieve data from the upload status file.")


def update_upload_status_file(max_event_sequence_id=None, log_data_timestamp=None):
    """
    Updates either the max_event_sequence_id or the log_data_timestamp given the input parameters.
    Also updates the upload status file's upload time for usage or logs respectively.
    If the JSON schema was not valid, the upload status file will be reset to the default and
    will not update the input values.
    :param max_event_sequence_id: int
    :param log_data_timestamp: string
    """
    try:
        if max_event_sequence_id and log_data_timestamp:
            raise ValueError("Should not have value for both parameters.")

        upload_status_file_path, schema_valid = get_upload_status_file_path(UPLOAD_STATUS_FILENAME)
        if schema_valid:
            with open(upload_status_file_path, 'r+', encoding='utf-8') as upload_status_file:
                upload_status_json = json.load(upload_status_file)
                log_msg = ""

                if max_event_sequence_id:
                    upload_status_json['usage']['max_event_sequence_id'] = max_event_sequence_id
                    upload_status_json['usage']['upload_timestamp'] = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')
                    log_msg = "Max event sequence ID was updated to: " + str(max_event_sequence_id)
                if log_data_timestamp:
                    upload_status_json['logs']['last_logs_timestamp'] = log_data_timestamp
                    upload_status_json['logs']['upload_timestamp'] = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')
                    log_msg = "Last logs timestamp was updated to: " + str(log_data_timestamp)
                
                upload_status_file.seek(0)
                json.dump(upload_status_json, upload_status_file, indent=4)
                upload_status_file.truncate
                log.info("Upload status file was updated. {}".format(log_msg))

    except BaseException:
        log.info("Upload status file was not able to be updated.")
        raise ValueError("Upload status file was not able to be updated.")


def get_upload_status_file_path(filename):
    """
    Returns the upload status file's path and schema_valid, which represents whether
    or not the input file was valid. If the input file did not follow the expected schema, 
    then the file will be rewritten with default values.
    """
    cli_name = Configuration().CLI_NAME
    config_dir = os.path.expanduser(Configuration().CLI_CONFIG_DIR)
    query_file = os.path.join(config_dir, '{}-{}'.format(cli_name, filename))
    query_file_exists = os.path.exists(query_file)
    schema_valid = False
    if query_file_exists:
        schema_valid = validate_upload_status_file(query_file)
    if not query_file_exists or not schema_valid:
        new_upload_status_file_content = {
            "logs": {
                "upload_timestamp": "11111111111111",
                "last_logs_timestamp": "11111111111111"
            },
            "usage": {
                "upload_timestamp": "11111111111111",
                "max_event_sequence_id": 0
            }
        }
        with open(query_file, mode='w', encoding='utf-8') as new_upload_status_file:
            json.dump(new_upload_status_file_content, new_upload_status_file)
            log.info("Upload status file was created or recreated.")
    
    return query_file, schema_valid


def validate_upload_status_file(path):
    """
    :param path: string value path of the upload status file
    Returns whether or not the input upload status file follows the expected JSON schema.
    """
    try:
        if not os.path.exists(path):
            raise FileNotFoundError('Cannot find file: "{}". Please provide the correct file name and try again'.format(
                path))
        with open(path, encoding='utf-8') as input_file:
            data = json.load(input_file)
            validate(data, UPLOAD_STATUS_FILE_SCHEMA)
        return True
    except Exception as e:
        return False


UPLOAD_STATUS_FILE_SCHEMA = {
        "type": "object",
        "properties": {
            "logs": {
                "type": "object",
                "properties": {
                    "upload_timestamp": {"type": "string", "maxLength": 14, "minLength": 14},
                    "last_logs_timestamp": {"type": "string", "maxLength": 14, "minLength": 14},
                },
                "required": [
                    "upload_timestamp",
                    "last_logs_timestamp"
                ],
            },
            "usage": {
                "type": "object",
                "properties": {
                    "upload_timestamp": {"type": "string", "maxLength": 14, "minLength": 14},
                    "max_event_sequence_id": {"type": "integer"},
                },
                "required": [
                    "upload_timestamp",
                    "max_event_sequence_id"
                ],
            }
        },
        "required": [
            "logs",
            "usage"
        ],
    }