import re
import ssl
import json
import urllib
import urllib.request
import urllib_kerberos
from enum import Enum
from types import MethodType
from six.moves import http_client
from urllib.parse import urlsplit, urlunsplit
from azdata.cli.commands.hdfs.exceptions import *
from azdata.cli.core.util import is_windows
try:
    from urllib.parse import quote, quote_plus
except ImportError:
    from urllib import quote, quote_plus

#
# Extended WebHdfs Class.
#
# This started out as an extension of the pywebhdfs package to add methods that it didn't provide.
# Specifically around ACLs and extended attributes.
#
# The base library uses the requests package to perform REST calls to the server.  When operating
# against a Kubernetes secured hdfs cluster there were problems with REST calls that contacted
# the name server and were redirected to a data node.  The authentication token was reused
# between the calls.  The security proxy Knox would close the connection because it detected
# the redirect authentication as a replay attack.
#
# No solution could be found and the pywebhdfs GitHub project hasn't had any activity in several
# years.  As a result this class now contains the original base class functionality with a
# change to use urllib.request instead of requests package.  The original pywebhdfs package
# needs to continue to be attributed in any products that use this class.
#
# Original project location: https://github.com/pywebhdfs/pywebhdfs
#


# WebHDFS Operations
#
class Operation(Enum):
    CREATE            = 'CREATE'
    APPEND            = 'APPEND'
    OPEN              = 'OPEN'
    MKDIRS            = 'MKDIRS'
    RENAME            = 'RENAME'
    DELETE            = 'DELETE'
    GETFILESTATUS     = 'GETFILESTATUS'
    LISTSTATUS        = 'LISTSTATUS'
    GETFILECHECKSUM   = 'GETFILECHECKSUM'
    GETCONTENTSUMMARY = 'GETCONTENTSUMMARY'
    GETXATTRS         = 'GETXATTRS'
    LISTXATTRS        = 'LISTXATTRS'
    REMOVEXATTR       = 'REMOVEXATTR'
    SETXATTR          = 'SETXATTR'
    GETHOMEDIRECTORY  = 'GETHOMEDIRECTORY'
    SETPERMISSION     = 'SETPERMISSION'
    SETOWNER          = 'SETOWNER'
    MODIFYACLENTRIES  = 'MODIFYACLENTRIES'
    REMOVEACLENTRIES  = 'REMOVEACLENTRIES'
    REMOVEDEFAULTACL  = 'REMOVEDEFAULTACL'
    REMOVEACL         = 'REMOVEACL'
    SETACL            = 'SETACL'
    GETACLSTATUS      = 'GETACLSTATUS'
    CHECKACCESS       = 'CHECKACCESS'


# HTTP Verbs
#
class HttpVerb(Enum):
    GET    = "GET"
    PUT    = "PUT"
    POST   = "POST"
    DELETE = "DELETE"


class ExtendedPyWebHdfsClient():
    """
    PyWebHdfsClient is a Python wrapper for the Hadoop WebHDFS REST API

    To use this client:

    >>> from pywebhdfs.webhdfs import PyWebHdfsClient
    """

    def __init__(self, host='localhost', port='50070', user_name=None,
                 path_to_hosts=None, timeout=None,
                 base_uri_pattern="http://{host}:{port}/webhdfs/v1/"):
        """
        Create a new client for interacting with WebHDFS

        :param host: the ip address or hostname of the HDFS namenode
        :param port: the port number for WebHDFS on the namenode
        :param user_name: WebHDFS user.name used for authentication
        :param path_to_hosts: mapping paths to hostnames for federation
        :param timeout: timeout for the underlying HTTP request
        :param base_uri_pattern: format string for base URI

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')

        Via a secure Knox gateway:

        >>> hdfs = PyWebHdfsClient(base_uri_pattern=
        >>>     "https://knox.mycluster.local:9443/gateway/default/webhdfs/v1")
        >>> hdfs.enable_basic_auth('username', 'password')
        >>> hdfs.enable_verify_mutual_auth('/etc/ssl/myrootca.crt')
        """

        self.host = host
        self.port = port
        self.user_name = user_name
        self.timeout = timeout
        self.path_to_hosts = path_to_hosts
        if self.path_to_hosts is None:
            self.path_to_hosts = [('.*', [self.host])]

        self.base_uri_pattern = base_uri_pattern.format(
            host="{host}", port=port)
        self.verify_mutual_auth = False
        self.use_basic_auth = False
        self.use_kerberos_auth = False
        self.negotiate_token = None
        return

    def enable_verify_mutual_auth(self, ca=None):
        self.verify_mutual_auth = True
        return

    def enable_basic_auth(self, username, password):
        self.use_basic_auth = True
        self.password_manager = urllib.request.HTTPPasswordMgrWithPriorAuth()

        # The URL pattern when using Knox is like "https://knox.mycluster.local:9443/gateway/default/webhdfs/v1"
        # But when redirecting to a data node it is changed like this:
        #        "https://knox.mycluster.local:9443/gateway/default/webhdfs/data/v1"
        # Because the 'data' comes before 'v1' the password manager doesn't see the data URL as a subset of
        # original URL so the password lookup fails.  Registering the password happens for the base URL.
        # So need to strip off the path before registering the password so all parts of WebHDFS can be
        # authenticated.
        #
        split_url = urlsplit(self.base_uri_pattern)
        split_url = split_url._replace(path="")
        url = urlunsplit(split_url)

        self.password_manager.add_password(None, url, username, password, is_authenticated=True)
        self.password_manager.add_password("application", url, username, password, is_authenticated=True)
        self.password_manager.add_password("Server", url, username, password, is_authenticated=True)
        return

    def enable_kerberos_auth(self):
        self.use_kerberos_auth = True

        if is_windows():
            from azdata.cli.core.clients.kerberos_client import KerberosClient
            from urllib.parse import urlsplit

            # Extract scheme + host only w/no trailing `/`
            # ----------------------------------------------
            # https://knox.tdeupgrade.aris.local:30443/gateway/default/webhdfs/v1/
            # https://knox.tdeupgrade.aris.local
            hostname = "{0.scheme}://{0.hostname}/". \
                format(urlsplit(self.base_uri_pattern))
            hostname = hostname[:-1] if hostname.endswith("/") else hostname

            # acquire GSSAPI token based on principal
            token = KerberosClient(
                endpoint=hostname
            ).acquire_gssapi_token()

            self.negotiate_token = token

    def create_file(self, path, file_data, **kwargs):
        """
        Creates a new file on HDFS

        :param path: the HDFS file path
        :param file_data: the initial data to write to the new file

        The function wraps the WebHDFS REST call:

        PUT http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=CREATE

        [&overwrite=<true|false>][&blocksize=<LONG>][&replication=<SHORT>]
        [&permission=<OCTAL>][&buffersize=<INT>]

        The function accepts all WebHDFS optional arguments shown above

        Example:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_data = '01010101010101010101010101010101'
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.create_file(my_file, my_data)

        Example with optional args:

        >>> hdfs.create_file(my_file, my_data, overwrite=True, blocksize=64)

        Or for sending data from file like objects:

        >>> with open('file.data') as file_data:
        >>>     hdfs.create_file(hdfs_path, data=file_data)


        Note: The create_file function does not follow automatic redirects but
        instead uses a two step call to the API as required in the
        WebHDFS documentation
        """

        # make the initial CREATE call to the HDFS namenode
        optional_args = kwargs
        init_response = self._resolve_host(HttpVerb.PUT, False,
                                           path, Operation.CREATE,
                                           **optional_args)
        if not init_response.status_code == http_client.TEMPORARY_REDIRECT:
            _raise_pywebhdfs_exception(
                init_response.status_code, init_response.content)

        # Get the address provided in the location header of the
        # initial response from the namenode and make the CREATE request
        # to the datanode
        uri = init_response.headers['location']
        if type(file_data) is str:
            file_data = bytes(file_data, 'utf-8')
        extra_args = { "headers": {'content-type': 'application/octet-stream'},
                       "data": file_data }
        response = self._make_web_call(HttpVerb.PUT, False, uri,
                            self.timeout,
                            **extra_args)

        if not response.status_code == http_client.CREATED:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def append_file(self, path, file_data, **kwargs):
        """
        Appends to an existing file on HDFS

        :param path: the HDFS file path
        :param file_data: data to append to existing file

        The function wraps the WebHDFS REST call:

        POST http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=APPEND

        [&buffersize=<INT>]

        The function accepts all WebHDFS optional arguments shown above

        Example:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_data = '01010101010101010101010101010101'
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.append_file(my_file, my_data)

        Example with optional args:

        >>> hdfs.append_file(my_file, my_data, overwrite=True, buffersize=4096)

        Note: The append_file function does not follow automatic redirects but
        instead uses a two step call to the API as required in the
        WebHDFS documentation

        Append is not supported in Hadoop 1.x
        """

        # make the initial APPEND call to the HDFS namenode
        optional_args = kwargs

        init_response = self._resolve_host(HttpVerb.POST, False,
                                           path, Operation.APPEND,
                                           **optional_args)
        if not init_response.status_code == http_client.TEMPORARY_REDIRECT:
            _raise_pywebhdfs_exception(
                init_response.status_code, init_response.content)

        # Get the address provided in the location header of the
        # initial response from the namenode and make the APPEND request
        # to the datanode
        uri = init_response.headers['location']
        extra_args = { "headers": {'content-type': 'application/octet-stream'},
                       "data": file_data }
        response = self._make_web_call(HttpVerb.POST, True, uri,
                            self.timeout,
                            **extra_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def read_file(self, path, **kwargs):
        """
        Reads from a file on HDFS  and returns the content

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=OPEN

        [&offset=<LONG>][&length=<LONG>][&buffersize=<INT>]

        Note: this function follows automatic redirects

        Example:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.read_file(my_file)
        01010101010101010101010101010101
        01010101010101010101010101010101
        01010101010101010101010101010101
        01010101010101010101010101010101
        """

        optional_args = kwargs

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.OPEN,
                                      **optional_args)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.content

    def make_dir(self, path, **kwargs):
        """
        Create a new directory on HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        PUT http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=MKDIRS

        [&permission=<OCTAL>]

        Example:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_dir = 'user/hdfs/data/new_dir'
        >>> hdfs.make_dir(my_dir)

        Example with optional args:

        >>> hdfs.make_dir(my_dir, permission=755)
        """

        optional_args = kwargs

        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.MKDIRS,
                                      **optional_args)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def rename_file_dir(self, path, destination_path):
        """
        Rename an existing directory or file on HDFS

        :param path: the HDFS file path
        :param destination_path: the new file path name

        The function wraps the WebHDFS REST call:

        PUT <HOST>:<PORT>/webhdfs/v1/<PATH>?op=RENAME&destination=<PATH>

        Example:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> current_dir = 'user/hdfs/data/my_dir'
        >>> destination_dir = 'user/hdfs/data/renamed_dir'
        >>> hdfs.rename_file_dir(current_dir, destination_dir)
        """

        destination_path = '/' + destination_path.lstrip('/')

        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.RENAME,
                                      destination=destination_path)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def delete_file_dir(self, path, recursive=False):
        """
        Delete an existing file or directory from HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        DELETE "http://<host>:<port>/webhdfs/v1/<path>?op=DELETE

        [&recursive=<true|false>]

        Example for deleting a file:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.delete_file_dir(my_file)

        Example for deleting a directory:

        >>> hdfs.delete_file_dir(my_file, recursive=True)
        """

        response = self._resolve_host(HttpVerb.DELETE, True,
                                      path, Operation.DELETE,
                                      recursive=recursive)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def get_file_dir_status(self, path):
        """
        Get the file_status of a single file or directory on HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETFILESTATUS

        Example for getting file status:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.get_file_dir_status(my_file)
        {
            "FileStatus":{
                "accessTime":1371737704282,
                "blockSize":134217728,
                "group":"hdfs",
                "length":90,
                "modificationTime":1371737704595,
                "owner":"hdfs",
                "pathSuffix":"",
                "permission":"755",
                "replication":3,
                "type":"FILE"
            }
        }

        Example for getting directory status:

        >>> my_dir = 'user/hdfs/data/'
        >>> hdfs.get_file_dir_status(my_file)
        {
            "FileStatus":{
                "accessTime":0,
                "blockSize":0,
                "group":"hdfs",
                "length":0,
                "modificationTime":1371737704208,
                "owner":"hdfs",
                "pathSuffix":"",
                "permission":"755",
                "replication":0,
                "type":"DIRECTORY"
            }
        }
        """

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETFILESTATUS)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def get_content_summary(self, path):
        """
        Get the content summary of a directory on HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETCONTENTSUMMARY

        Example for getting a directory's content summary:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_folder = 'user/hdfs/data/'
        >>> hdfs.get_file_dir_status(my_folder)
        {
            "ContentSummary":
            {
                "directoryCount": 2,
                "fileCount": 1,
                "length": 24930,
                "quota": -1,
                "spaceConsumed": 24930,
                "spaceQuota": -1
            }
        }
        """

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETCONTENTSUMMARY)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def get_file_checksum(self, path):
        """
        Get the file_checksum of a single file on HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETFILECHECKSUM

        Example for getting file status:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.get_file_checksum(my_file)
        {
            "FileChecksum":{
                "algorithm": "MD5-of-1MD5-of-512CRC32",
                "bytes": "000002000000000000000000729a144ad5e9399f70c9bed...",
                "length": 28
            }
        }
        """

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETFILECHECKSUM)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def list_dir(self, path):
        """
        Get a list of file_status for all files and directories
        inside an HDFS directory

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=LISTSTATUS

        Example for listing a directory:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_dir = 'user/hdfs'
        >>> hdfs.list_dir(my_dir)
        {
            "FileStatuses":{
                "FileStatus":[
                    {
                        "accessTime":1371737704282,
                        "blockSize":134217728,
                        "group":"hdfs",
                        "length":90,
                        "modificationTime":1371737704595,
                        "owner":"hdfs",
                        "pathSuffix":"example3.txt",
                        "permission":"755",
                        "replication":3,
                        "type":"FILE"
                    },
                    {
                        "accessTime":1371678467205,
                        "blockSize":134217728,
                        "group":"hdfs","length":1057,
                        "modificationTime":1371678467394,
                        "owner":"hdfs",
                        "pathSuffix":"example2.txt",
                        "permission":"700",
                        "replication":3,
                        "type":"FILE"
                    }
                ]
            }
        }

        """

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.LISTSTATUS)
        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def exists_file_dir(self, path):
        """
        Checks whether a file or directory exists on HDFS

        :param path: the HDFS file path

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETFILESTATUS

        and returns a bool based on the response.

        Example for checking whether a file exists:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.exists_file_dir(my_file)
        True
        """
        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETFILESTATUS)
        if response.status_code == http_client.OK:
            return True
        elif response.status_code == http_client.NOT_FOUND:
            return False
        _raise_pywebhdfs_exception(response.status_code, response.content)

    def get_xattr(self, path, xattr=None):
        """
        Get extended attributes set on an HDFS path

        :param path: the HDFS file path without a leading '/'
        :param xattr: the extended attribute to get

        This function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETXATTRS

        [&xattr.name=<string>]

        Example for getting an extended attribute:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.get_xattr('user/hdfs/data.txt', xattr='user.important')
        {
            "XAttrs": [
                {
                    "name":"user.important",
                    "value":"very"
                }
            ]
        }
        """
        kwd_params = {}
        if xattr:
            kwd_params['xattr.name'] = xattr

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETXATTRS,
                                      **kwd_params)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return response.json()

    def set_xattr(self, path, xattr, value, replace=False):
        """
        Set an extended attribute on an HDFS path

        :param path: the HDFS file path without a leading '/'
        :param xattr: the extended attribute to set
        :param value: the value of the extended attribute
        :param replace: replace the extended attribute

        This function wraps the WebHDFS REST call:

        PUT http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETXATTR

        [&xattr.name=<string>][&xattr.value=<string][&flag=<REPLACE|CREATE>]

        Example for setting an extended attribute:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.set_xattr('user/hdfs/data.txt', 'user.important', 'very')
        """
        kwd_params = {
            'xattr.name': xattr,
            'xattr.value': value
        }
        if replace:
            kwd_params['flag'] = "REPLACE"
        else:
            kwd_params['flag'] = "CREATE"

        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.SETXATTR,
                                      **kwd_params)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return True

    def list_xattrs(self, path):
        """
        List all the extended attributes set on an HDFS path

        :param path: the HDFS file path without a leading '/'

        This function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=LISTXATTRS

        Example for getting an extended attribute:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.list_xattrs('user/hdfs/data.txt')
        {
            "XAttrNames": "[\"XATTRNAME1\",\"XATTRNAME2\",\"XATTRNAME3\"]"
        }
        """
        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.LISTXATTRS)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return response.json()

    def delete_xattr(self, path, xattr):
        """
        Delete the extended attribute set on an HDFS path

        :param path: the HDFS file path without a leading '/'
        :param xattr: the extended attribute to delete

        This function wraps the WebHDFS REST call:

        PUT http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEXATTR

        [&xattr.name=<string>]

        Example for getting an extended attribute:

        >>> hdfs = PyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.delete_xattr('user/hdfs/data.txt', 'user.important')
        """
        kwd_params = {
            'xattr.name': xattr
        }

        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.REMOVEXATTR,
                                      **kwd_params)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return True

    def _create_uri(self, path, operation, **kwargs):
        """
        internal function used to construct the WebHDFS request uri based on
        the <PATH>, <OPERATION>, and any provided optional arguments
        """

        no_root_path = (path[1:] if path[0] == '/' else path)
        path_param = quote(no_root_path.encode('utf8'))

        # setup the parameter represent the WebHDFS operation
        operation_param = '?op={operation}'.format(operation=operation.value)

        # configure authorization based on provided credentials
        auth_param = str()
        if self.user_name:
            auth_param = '&user.name={user_name}'.format(
                user_name=self.user_name)

        # setup any optional parameters
        keyword_params = str()
        for key in kwargs:
            try:
                value = quote_plus(kwargs[key].encode('utf8'))
            except:
                value = str(kwargs[key]).lower()
            keyword_params = '{params}&{key}={value}'.format(
                params=keyword_params, key=key, value=value)

        base_uri = self.base_uri_pattern.format(host="{host}")

        # build the complete uri from the base uri and all configured params
        uri = '{base_uri}{path}{operation}{keyword_args}{auth}'.format(
            base_uri=base_uri, path=path_param,
            operation=operation_param, keyword_args=keyword_params,
            auth=auth_param)

        return uri

    def _resolve_federation(self, path):
        """
        internal function used to resolve federation
        """
        for path_regexp, hosts in self.path_to_hosts:
            if re.match(path_regexp, path):
                return hosts
        raise CorrespondHostsNotFound(
            msg="Could not find hosts corresponds to /{0}".format(path))


    def _make_web_call(self, http_verb, allow_redirects, uri, timeout, **extra_opts):
        """
        Internal function that performs the REST request using urllib.  Urllib
        uses a list of handlers to talk to the server.  Depending on the options
        selection various handlers are added to the list.
        """
        context = ssl.create_default_context()
        if not self.verify_mutual_auth:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        handlers = [
            NoThrowDefaultErrorHandler(),
            urllib.request.HTTPSHandler(context=context)
        ]

        if self.use_kerberos_auth:
            if is_windows() and self.negotiate_token:
                headers = extra_opts.get("headers", {})
                headers["Authorization"] = "Negotiate {value}".format(
                    value=self.negotiate_token
                )
                extra_opts["headers"] = headers
            else:
                import urllib_gssapi
                handlers.append(urllib_gssapi.HTTPSPNEGOAuthHandler())
                handlers.append(urllib_kerberos.HTTPKerberosAuthHandler())

        if self.use_basic_auth:
            handlers.append(urllib.request.HTTPBasicAuthHandler(self.password_manager))

        if allow_redirects:
            handlers.append(urllib.request.HTTPRedirectHandler())
        else:
            handlers.append(NoRedirectionHandler())

        opener = urllib.request.build_opener(*handlers)
        request = urllib.request.Request(uri, method=http_verb.value)
        if "data" in extra_opts:
            request.data = extra_opts["data"]

        if "headers" in extra_opts:
            for header, value in extra_opts["headers"].items():
                request.add_header(header, value)

        response = opener.open(request)
        response.status_code = response.getcode()
        response.content = response.read()
        response.json = MethodType(_json_parse_body, response)
        return response

    def _resolve_host(self, request_type, allow_redirect,
                      path, operation, **kwargs):
        """
        internal function used to resolve federation and HA and
        return response of resolved host.
        """
        uri_without_host = self._create_uri(path, operation, **kwargs)
        hosts = self._resolve_federation(path)
        for host in hosts:
            uri = uri_without_host.format(host=host)
            response = self._make_web_call(request_type, allow_redirect, uri,
                                self.timeout)

            if not _is_standby_exception(response):
                _move_active_host_to_head(hosts, host)
                return response
        raise ActiveHostNotFound(msg="Could not find active host")

    """
    This class extends the PyWebHdfsClient with additional calls that the standard library doesn't
    implement.  By extending the class the same method implementation pattern can be used and the
    code reuse is maximized.

    Note there are no parts of the protocol implemented by this class that are different than
    the standard Apache Hadoop WebHdfs documentation.  No customization should eb done as part
    of this class.
    """

    def get_home_directory(self):
        """
	    Gets the calling users home directory.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/?op=GETHOMEDIRECTORY

        and returns a directory path.

        Example for getting home directory:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.get_home_directory()
        {"Path": "/user/username"}

    	"""
        path = "/"
        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETHOMEDIRECTORY)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def set_permission(self, path, **kwargs):
        """
	    Sets the permission on a file or directory.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETPERMISSION

        Example for setting permission:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_permission(my_file, '755')

        """
        optional_args = kwargs
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.SETPERMISSION, **optional_args)

        if response.status_code == http_client.OK:
            return True
        elif response.status_code == http_client.NOT_FOUND:
            return False
        _raise_pywebhdfs_exception(response.status_code, response.content)

    def set_owner(self, path, **kwargs):
        """
    	Sets the user and group owners for a file or directory

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETOWNER[&owner=<USER>][&group=<GROUP>]

        Example for setting user and group owner:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_owner(my_file, owner='hdfs', group='hdfs')

        """
        optional_args = kwargs
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.SETOWNER, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def modify_acl_entries(self, path, **kwargs):
        """
        Can add a new ACL entries or modify existing ACL entries.  Any ACL not specified is left unchanged.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=MODIFYACLENTRIES&aclspec=<ACLSPEC>

        Example for modifying ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.modify_acl_entries(my_file, aclspec='user:bob:rw-')

        """
        optional_args = kwargs
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.MODIFYACLENTRIES, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_acl_entries(self, path, **kwargs):
        """
        Remove existing ACL entries.  Any ACL not specified is left unchanged.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEACLENTRIES&aclspec=<ACLSPEC>

        Note the aclspec doesn't have read/write/execute flags.

        Example for removing ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_acl_entries(my_file, aclspec='user:bob:')

        """
        optional_args = kwargs
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.REMOVEACLENTRIES, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_default_acl_entries(self, path):
        """
        Remove default ACL entries: default:user, default:group, default:other

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEDEFAULTACL

        Example for removing default ACL entries:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_default_acl_entries(my_file)

        """
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.REMOVEDEFAULTACL)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def remove_acl(self, path):
        """
        Remove all but the base ACL entries.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=REMOVEACL

        Example for removing an ACL:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.remove_acl(my_file)

        """
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.REMOVEACL)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def set_acl(self, path, **kwargs):
        """
        Fully replaces the ACL entries with new values.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=SETACL&aclspec=<ACLSPEC>

        Example for setting an ACL:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.set_acl(my_file, aclspec='user:bob:rw-')

        """
        optional_args = kwargs
        response = self._resolve_host(HttpVerb.PUT, True,
                                      path, Operation.SETACL, **optional_args)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return True

    def get_acl_status(self, path):
        """
        Returns the AclStatus JSON object for the given path.

        The function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETACLSTATUS

        Example for getting an ACL's status:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> my_file = 'user/hdfs/data/myfile.txt'
        >>> hdfs.get_acl_status(my_file)

        """
        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETACLSTATUS)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)

        return response.json()

    def get_xattr(self, path, xattr=None):
        """
        Get extended attributes set on an HDFS path.

        This function wraps the WebHDFS REST call:

        GET http://<HOST>:<PORT>/webhdfs/v1/<PATH>?op=GETXATTRS

        Example for getting an extended attribute:

        >>> hdfs = ExtendedPyWebHdfsClient(host='host',port='50070', user_name='hdfs')
        >>> hdfs.get_xattr('user/hdfs/data.txt', xattr='user.important')

        """
        kwd_params = {}
        if xattr:
            kwd_params['xattr.name'] = xattr
        kwd_params['encoding'] = 'text'

        response = self._resolve_host(HttpVerb.GET, True,
                                      path, Operation.GETXATTRS,
                                      **kwd_params)

        if not response.status_code == http_client.OK:
            _raise_pywebhdfs_exception(response.status_code, response.content)
        return response.json()


def _raise_pywebhdfs_exception(resp_code, message=None):

    if resp_code == http_client.BAD_REQUEST:
        raise BadRequest(msg=message)
    elif resp_code == http_client.UNAUTHORIZED:
        raise Unauthorized(msg=message)
    elif resp_code == http_client.NOT_FOUND:
        raise FileNotFound(msg=message)
    elif resp_code == http_client.METHOD_NOT_ALLOWED:
        raise MethodNotAllowed(msg=message)
    else:
        raise PyWebHdfsException(msg=message)


def _is_standby_exception(response):
    """
    check whether response is StandbyException or not.
    """
    if response.status_code == http_client.FORBIDDEN:
        try:
            body = response.json()
            exception = body["RemoteException"]["exception"]
            if exception == "StandbyException":
                return True
        except:
            pass
    return False


def _move_active_host_to_head(hosts, active_host):
    """
    to improve efficiency move active host to head
    """
    hosts.remove(active_host)
    hosts.insert(0, active_host)


def _json_parse_body(self):
    """
    Helper that is added to response objects to decode the json body.
    """
    # TODO: Use headers to decode
    decoded_body = json.loads(self.content.decode('utf-8'))
    return decoded_body


class NoRedirectionHandler(urllib.request.HTTPRedirectHandler):
    '''
    Redirection handler that doesn't follow redirection responses.
    This is necessary for PUT operations where a PUT is needed against
    the redirection target.  The normal behavior is that a redirection
    is always a GET operation.
    '''
    def redirect_request(self, req, fp, code, msg, hdrs, newurl):
        pass

class NoThrowDefaultErrorHandler(urllib.request.HTTPDefaultErrorHandler):
    """
    The default error handler throws an exception on a non-200 result.
    This overrides the behavior to just return the response object.
    """
    def http_error_default(self, req, fp, code, msg, hdrs):
        response = urllib.response.addinfourl(fp, hdrs, req.full_url, code=code)
        return response


