# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import os
import json
import tempfile
import sys

# On some terminals the width can't be determined.  This causes an error message from MDV.  Since GCI is one
# place where the terminal width check fails it causes an unexpected stderr message and fails some tests checking
# for empty stderr.  As a work around set COLUMNS environment variable before loading mdv.  Longer term
# should probably do the detection and use the width where it is available instead of a fixed value.
#
os.environ["COLUMNS"] = "80"

#
# Needed for Windows Python 3.8, otherwise the tornado dependency will give
# errors. https://github.com/tornadoweb/tornado/issues/2608
#
# Important! These lines needs to happen before the imports below which pull
# in tornado.
#
if sys.version_info[0] == 3 \
        and sys.version_info[1] >= 8 \
        and sys.platform.startswith('win'):
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

from nbformat import from_dict as nbformat_from_dict
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError
from azdata.cli.core.clients.cli_client import CliClient
from nbconvert import HTMLExporter
from azdata.cli.commands.notebook.util import CellFormatter
from azdata.cli.commands.notebook.interactive import InteractiveNotebook


import codecs
import nbformat


def beget(_):
    """Client factory"""
    return NotebookClientMixin(check_auth=False, check_eula=True)


class NotebookClientMixin(CliClient):
    def __init__(self, check_auth=False, check_eula=True):
        """
        :param check_auth: Set to `True` to have this command enforce authentication
                           before running.
        :param check_auth: Set to `True` to have this command enforce eula
                           acceptance before running.
        """
        if os.name == 'nt':
            from colorama import init
            init()

        super(NotebookClientMixin, self).__init__(
            check_auth=check_auth, check_eula=check_eula)

        # Persist modified profile items immediately before moving onto any
        # notebook executions. The reason for this is a supplied notebook could
        # potentially contain azdata commands that modify the profile itself.
        # This could cause this profile to be out-of-date here with the one on
        # disk modified by the notebook resulting in overwriting that content
        # with this profile content.
        self.profile.persist()

    def _convert_notebook(self, notebook_path, noteobok_content):
        path = os.path.dirname(notebook_path)
        filename = os.path.basename(notebook_path)
        file_prefix, file_ext = os.path.splitext(filename)

        notebook_html = os.path.join(path, "{0}.{1}".format(file_prefix, "html"))

        exporter = HTMLExporter()
        exporter.template_file = 'full'

        (body, resources) = exporter.from_notebook_node(noteobok_content)

        with codecs.open(notebook_html, mode='w', encoding='utf-8') as f:
            f.write(body)

        return notebook_html

    def _decode_dictionary(self, value):
        """
        Some parameters need to be dictionary of name value pairs.  For example configuration dictionary.
        This function checks to see if the content is a json dictionary and if so decodes it.

        Example:
            { "name": "value", "another_name": "another_value" }
        """
        if value.strip().startswith('{'):
            result = json.loads(value)
        else:
            raise ValueError("Value must be JSON encoded dictionary [{0}].".format(value))

        return result

    def cells_with_tag(self, notebook, tag):
        tagged_cell_indexes = [notebook.cells.index(x) for x in notebook.cells if x.cell_type == 'code' and hasattr(x.metadata, "tags") and tag in x.metadata.tags]
        return tagged_cell_indexes

    def generate_parameters_cell(self, arguments):
        cell = dict(cell_type="code", execution_count=None, outputs=[], source=[])

        cell["metadata"] = nbformat_from_dict(
            {
                "tags": [
                    "generated"
                ]
            }
        )
        cell["source"].append("#----------------------------------------------------------------------------------------\n")
        cell["source"].append("# Generated Cell\n")
        cell["source"].append("#\n")
        cell["source"].append("# This cell contains parameters passed into the notebook execution command line.\n")
        cell["source"].append("# The command line should not contain sensitive data.  To pass sensitive data use an\n")
        cell["source"].append("# environment variable or use secure storage.\n")
        cell["source"].append("#----------------------------------------------------------------------------------------\n")
        cell["source"].append("#\n")
        for key in sorted(arguments):
            cell["source"].append("{0} = '{1}'\n".format(key, arguments[key]))
        cell["source"] = CellFormatter.convert_list_to_string(cell["source"])

        cell = nbformat_from_dict(cell)

        return cell

    # TODO: Remove this after interactive code is also updated.
    def execute_notebook_batch_deprecated(self, notebook_path, output_path=None, output_html=None, arguments=None, timeout=600, suppress_generated_cells=False):
        msg = ""
        exception = None
        notebook_html = None
        notebook_directory, notebook_filename = os.path.split(notebook_path)

        with open(notebook_path, "r", encoding="utf-8") as f:
            if output_path == None:
                notebook_working_folder = tempfile.mkdtemp()
                notebook_output = os.path.join(notebook_working_folder, "output-" + notebook_filename)
            else:
                notebook_working_folder = output_path
                notebook_output = os.path.join(output_path, "output-" + notebook_filename)

            nb = nbformat.read(f, as_version=4)

            if arguments is not None:
                # If the user has passed in arguments then see if the notebook has a cell with a 'parameters' tag.
                # If it does generate a new cell after the parameters cell with the passed in parameters.
                #
                # NOTE for now this only supports Python cells.
                #
                arguments = self._decode_dictionary(arguments)
                cells = self.cells_with_tag(nb, "parameters")
                if len(cells) > 0:
                    gen_cell = self.generate_parameters_cell(arguments)
                    nb.cells.insert(cells[0] + 1, gen_cell)
                else:
                    # None of the cells were marked for parameters.
                    msg += "Warning: Parameters were passed in but no cells in the target notebook were marked with parameters tag.  Parameters not inserted.\n"
            try:
                # Check and see if the notebook has a kernelspec.  If not default to python
                if hasattr(nb, "metadata") and hasattr(nb.metadata, "kernelspec") and hasattr(nb.metadata.kernelspec, "name"):
                    kernel = nb.metadata.kernelspec.name
                else:
                    kernel = "python3"

                # Note if the timeout is -1 convert it to None.  The traitlet
                # throws an error if you pass a negative number.
                #
                if timeout == '-1':
                    timeout = None
                else:
                    timeout = int(timeout)
                ep = ExecutePreprocessor(timeout=timeout, kernel_name=kernel)
                out = ep.preprocess(nb, {'metadata': {'path': notebook_working_folder}})
            except CellExecutionError as cee:
                exception = cee
                out = None
                msg += 'Error executing the notebook "%s".\n\n' % notebook_filename
                msg += cee.traceback + '\n\n'
                msg += 'See notebook "%s" for details of the execution.' % notebook_output
            except TimeoutError as te:
                exception = te
                out = None
                msg += 'Timeout executing the notebook "%s".\n\n' % notebook_filename
                msg += 'See notebook "%s" for details of the execution.' % notebook_output
            finally:
                if suppress_generated_cells:
                    # If there were generated cells remove them
                    generated_cells = self.cells_with_tag(nb, "generated")
                    if len(generated_cells) > 0:
                        generated_cells.sort(reverse=True)
                        for index in generated_cells:
                            nb.cells.remove(nb.cells[index])

                with open(notebook_output, mode='w', encoding='utf-8') as of:
                    nbformat.write(nb, of)

                if output_html:
                    notebook_html = self._convert_notebook(notebook_output, nb)

        return notebook_output, notebook_html, msg, exception

    def execute_notebook_interactive(self, notebook_path, clear=None):
        InteractiveNotebook.launch_interactive_console_notebook(self, notebook_path, clear)
        return

    def render_notebook(self, notebook_path, continue_on_error):
        with open(notebook_path, "r", encoding="utf-8") as f:
            output = ""
            notebook = json.load(f)
            i = 1
            found_error = False

            # Check and see if the notebook has a kernelspec.  If not default to python
            if hasattr(notebook, "metadata") and hasattr(notebook.metadata, "kernelspec") and hasattr(notebook.metadata.kernelspec, "name"):
                kernel = notebook.metadata.kernelspec.name
            else:
                kernel = "python3"

            for cell in notebook['cells']:
                cell_output, found_error = CellFormatter.render_cell(cell, kernel, i)
                output += cell_output
                i += 1
                if found_error and not continue_on_error:
                    break

            return output



