# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import sys
import time
import json
import queue
import signal
import nbformat
from jupyter_client.consoleapp import JupyterConsoleApp
from jupyter_core.application import JupyterApp
from traitlets import (
    Instance, Dict, Unicode, Bool, List, CUnicode, Any, Float
)
from azdata.cli.commands.notebook.util import CellFormatter
from azdata.cli.commands.notebook import _


OUTPUT_TIMEOUT = 60

class InteractiveNotebook(JupyterApp, JupyterConsoleApp):
    name = 'azdata notebook'
    description = "Interactive CLI Notebook"
    kernel_timeout = Float(60, config=True)

    def initialize(self, argv=None):
        super(InteractiveNotebook, self).initialize(argv)
        JupyterConsoleApp.initialize(self)
        signal.signal(signal.SIGINT, self.handle_sigint)
        self.init_kernel_info()
        return

    def handle_sigint(self, *args):
        if self.kernel_manager:
            self.kernel_manager.interrupt_kernel()
        else:
            print("", file=sys.stderr)
            raise Exception("Cannot interrupt kernel.\n")

    def init_kernel_info(self):
        timeout = self.kernel_timeout
        tic = time.time()
        self.kernel_client.hb_channel.unpause()
        msg_id = self.kernel_client.kernel_info()
        while True:
            try:
                reply = self.kernel_client.get_shell_msg(timeout=1)
            except queue.Empty:
                if (time.time() - tic) > timeout:
                    raise RuntimeError("Kernel didn't respond to kernel_info_request")
            else:
                if reply['parent_header'].get('msg_id') == msg_id:
                    self.kernel_info = reply['content']
                    return

    def print_help(self):
        print(_("Interactive Notebook Experience\n"))
        print(_("This allows for an experience of walking through a notebook a cell at a time.  Each"))
        print(_("cell will be displayed and depending on the type of cell various commands can be"))
        print(_("used.  Cells can be executed, navigation to next and previous cells, and various"))
        print(_("other operations like saving the notebook to a file.\n"))
        print(_("Command List:"))
        print(_("    ?    Display this help message."))
        print(_("    x    Execute the current cell.  Only applies to code cells."))
        print(_("    n    Move to the next cell."))
        print(_("    p    Move to the previous cell."))
        print(_("    a    Enter and run an adhoc code snippet."))
        print(_("    q    Quit."))
        return

    def move_to_next_cell(self, notebook, current_cell):
        number_of_cells = len(notebook["cells"])
        next_cell = current_cell + 1
        if next_cell >= number_of_cells:
            next_cell = current_cell
        else:
            if self.clear:
                os.system("cls||clear")

            # Render the current cell.
            output = CellFormatter.render_cell(notebook["cells"][next_cell], self.kernel, next_cell)
            print(output[0])

        return next_cell

    def move_to_previous_cell(self, notebook, current_cell):
        number_of_cells = len(notebook["cells"])
        previous_cell = current_cell - 1
        if previous_cell < 0:
            previous_cell = 0
        else:
            if self.clear:
                os.system("cls||clear")

            # Render the current cell.
            output = CellFormatter.render_cell(notebook["cells"][previous_cell], self.kernel, previous_cell)
            print(output[0])

        return previous_cell

    def notebook_main_loop(self, notebook_path):
        with open(notebook_path, "r", encoding="utf-8") as f:
            notebook = json.load(f)

            # Check and see if the notebook has a kernelspec.  If not default to python
            if hasattr(notebook, "metadata") and hasattr(notebook.metadata, "kernelspec") and hasattr(
                    notebook.metadata.kernelspec, "name"):
                self.kernel = notebook.metadata.kernelspec.name
            else:
                self.kernel = "python3"

            if self.clear:
                os.system("cls||clear")

            done = False
            current_cell = 0
            cell = notebook["cells"][current_cell]

            # Render the current cell.
            output = CellFormatter.render_cell(notebook["cells"][current_cell], self.kernel, current_cell)
            print(output[0])

            while not done:
                if cell['cell_type'] == "code":
                    print(_("\nEnter Command : [?] Help, [x] Execute Cell, [n] Next Cell, [p] Previous Cell, [a] Adhoc code, [q] Quit"))
                else:
                    print(_(
                        "\nEnter Command : [?] Help, [n] Next Cell, [p] Previous Cell, [a] Adhoc code, [q] Quit"))
                command = input("> ")

                # Execute command
                if command == "?":
                    self.print_help()
                elif command == "x":
                    print(_("\nOutput:\n"))
                    reply = self.kernel_client.execute_interactive(CellFormatter.convert_list_to_string(cell['source']),
                                                                   timeout=OUTPUT_TIMEOUT)

                elif command == "n":
                    current_cell = self.move_to_next_cell(notebook, current_cell)
                    cell = notebook["cells"][current_cell]
                elif command == "p":
                    current_cell = self.move_to_previous_cell(notebook, current_cell)
                    cell = notebook["cells"][current_cell]
                elif command == "a":
                    print(_(
                        "\nEnter cell content to execute.  Complete editing with two empty lines in a row."))

                    cell_contents = []
                    adhoc_done = False
                    empty_line_count = 0
                    while not adhoc_done:
                        command = input("> ")
                        if not len(command) == 0:
                            cell_contents.append(command + "\n")
                            empty_line_count = 0
                        else:
                            if empty_line_count == 0:
                                cell_contents.append(command + "\n")
                                empty_line_count += 1
                            else:
                                adhoc_done = True

                    print(_("\nOutput:\n"))
                    reply = self.kernel_client.execute_interactive(CellFormatter.convert_list_to_string(cell_contents),
                                                                   timeout=OUTPUT_TIMEOUT)
                elif command == "q":
                    done = True
                else:
                    print(_("Error unknown command [{0}].".format(command)), file=sys.stderr)

        return

    def start(self):
        super(InteractiveNotebook, self).start()
        self.notebook_main_loop(self.filename)
        return

    @staticmethod
    def detect_kernelspec(notebook_path):
        kernel = "python3"
        with open(notebook_path, "r", encoding="utf-8") as f:
            nb = nbformat.read(f, as_version=4)

            if hasattr(nb, "metadata") and hasattr(nb.metadata, "kernelspec") and hasattr(nb.metadata.kernelspec, "name"):
                kernel = nb.metadata.kernelspec.name
        return kernel

    @staticmethod
    def launch_interactive_console_notebook(client, notebook_path, clear):
        kernel_name = InteractiveNotebook.detect_kernelspec(notebook_path)
        notebook_app = InteractiveNotebook()
        notebook_app.client = client
        notebook_app.filename = notebook_path
        notebook_app.kernel_name = kernel_name
        notebook_app.clear = clear
        notebook_app.initialize(argv=["azdata"])
        notebook_app.start()

        # Note: the Juypter client class leaks around 30 socket handles by default on every create/delete
        # instance.  This code reaches into the class to shut down connections manually because they don't
        # get cleaned up on class deletion.  The class holds weak references so things aren't shut down normally.
        #
        notebook_app.kernel_manager.shutdown_kernel(False, False)
        notebook_app.kernel_manager.cleanup_resources()
        notebook_app.kernel_manager = None
        notebook_app.kernel_client.stop_channels()
        notebook_app.kernel_client = None
        del notebook_app

        return


