# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import json
import codecs
import logging
import tempfile
from nbclient import NotebookClient
from nbclient.exceptions import CellExecutionError, CellTimeoutError
from nbformat import NotebookNode, from_dict
from nbconvert import NotebookExporter, HTMLExporter
from azdata.cli.commands.notebook.util import CellFormatter, WorkingDirectoryContext, get_livy_endpoint
from azdata.cli.commands.notebook import _
from azdata.cli.commands.spark.custom import (
    _get_knox_username,
    _get_knox_password
)
from azdata.cli.core.enums import SecurityStrategy
from azdata.cli.core.exceptions import CliError

# Uncomment to enable debug logging
#
#logging.basicConfig(level=logging.DEBUG)

# Configure Sparkmagic to clean up sessions when azdata exits.
#
try:
    import sparkmagic.utils.configuration as conf
    conf.override('cleanup_all_sessions_on_exit', True)
except Exception:
    pass

class NotebookDriver:
    def __init__(self):
        self.retry_count = 3
        return

    def run_notebook(self, client, notebook_path, output_path=None, output_html=None, arguments=None, timeout=60, suppress_generated_cells=False):
        execution_exception = None
        html_notebook_output_file = None

        # Setup the output directory and file names based on the input parameters
        #
        notebook_path = os.path.abspath(notebook_path)
        notebook_directory, notebook_filename = os.path.split(notebook_path)
        output_notebook_filename = "output-" + notebook_filename
        if output_path == None:
            working_folder = tempfile.mkdtemp()
            notebook_output_file = os.path.join(working_folder, output_notebook_filename)
        else:
            working_folder = output_path
            notebook_output_file = os.path.join(output_path, output_notebook_filename)

        resources = {'metadata': {'path': working_folder}}

        if output_html:
            file_prefix, file_ext = os.path.splitext(output_notebook_filename)
            html_notebook_output_file = os.path.join(working_folder, "{0}.{1}".format(file_prefix, "html"))

        with WorkingDirectoryContext(working_folder) and open(notebook_path, "r", encoding="utf-8") as f:
            # Decode the file into nested dictionary data structure
            #
            notebook = json.load(f)

            # Load the name of the Jupyter kernel configured in the notebook.  If the notebook doesn't have a kernel
            # configured then Python3 is used as the default.
            #
            kernel_name = self._get_kernel_name(notebook)

            # Process passed in arguments
            #
            # NOTE: for now this only supports Python based kernels.
            #
            if kernel_name == "python3" or kernel_name == "pysparkkernel":
                if arguments is not None:
                    # If the user has passed in arguments then see if the notebook has a cell with a 'parameters' tag.
                    # If it does generate a new cell after the parameters cell with the passed in parameters.
                    #
                    arguments = self._decode_dictionary(arguments)
                    self._generate_parameters_cell(notebook, kernel_name, arguments)

                # Generate kernel specific cells
                #
                self._generate_kernel_specific_cells(client, notebook, kernel_name)

            # Canonicalize the notebook json to what the client expects.
            #
            notebook = self._canonicalize_code_cells(notebook)

            # Convert from json to NotebookNode
            #
            nbnode = self._convert_json_to_notebook(notebook)

            # Execute the notebook.  The output is a new NotebookNode containing the results.
            #

            retry_count = 0
            done = False
            completed_notebook = None

            while not done:
                nbclient = NotebookClient(nbnode)
                nbclient.timeout = timeout
                nbclient.kernel_name = kernel_name
                nbclient.resources = resources

                try:
                    completed_notebook = nbclient.execute()
                    done = True
                except CellExecutionError as cee:
                    execution_exception = cee
                    completed_notebook = nbclient.nb
                    done = True
                except CellTimeoutError as cte:
                    execution_exception = cte
                    completed_notebook = nbclient.nb
                    done = True
                except RuntimeError as rte:
                    # The kernel died.  Try again
                    if retry_count < self.retry_count:
                        done = False
                        retry_count += 1
                    else:
                        execution_exception = rte
                        if nbclient and hasattr(nbclient, 'nb'):
                            completed_notebook = nbclient.nb
                        done = True
                finally:
                    del nbclient

            # If there were generated cells remove them
            #
            if suppress_generated_cells:
                completed_notebook = self._remove_generated_cells(completed_notebook)

            # Convert output
            # TODO: handle resources correctly.  Right now they aren't saved to the output directory
            #
            output_notebook, resources = self._convert_notebook_to_json(completed_notebook)

            # If specified convert the notebook to HTML
            #
            notebook_html = None
            if output_html:
                (html_notebook, html_resources) = self._convert_notebook_to_html(completed_notebook)

        # Write the output files after input file is closed to support overwrite
        #
        self._write_to_file(notebook_output_file, output_notebook)
        if output_html:
            self._write_to_file(html_notebook_output_file, html_notebook)

        return notebook_output_file, html_notebook_output_file, execution_exception

    def _get_kernel_name(self, notebook):
        kernel = "python3"
        if "metadata" in notebook and "kernelspec" in notebook["metadata"] and "name" in notebook["metadata"]["kernelspec"]:
            kernel = notebook["metadata"]["kernelspec"]["name"]
        return kernel

    def _convert_json_to_notebook(self, notebook):
        # This uses a converter from the nbformat library to convert from JSON document to NotebookNode
        #
        notebook_node = from_dict(notebook)
        return notebook_node

    def _convert_notebook_to_json(self, notebook_node):
        exporter = NotebookExporter()
        (output_notebook, resources) = exporter.from_notebook_node(notebook_node)
        return output_notebook, resources

    def _convert_notebook_to_html(self, notebook_node):
        exporter = HTMLExporter()
        exporter.template_file = 'full'
        (output_notebook, resources) = exporter.from_notebook_node(notebook_node)
        return output_notebook, resources

    def _remove_generated_cells(self, notebook):
        generated_cells = self._cells_with_tag(notebook, "generated")
        if len(generated_cells) > 0:
            generated_cells.sort(reverse=True)
            for index in generated_cells:
                notebook["cells"].remove(notebook["cells"][index])
        return notebook

    def _canonicalize_code_cells(self, notebook_json):
        """
        The Notebook executor expects code source to be in a string.  JSON formatters
        use multi-line to make reading the code easier.  This method converts the
        source lists to strings.
        """
        for cell in notebook_json["cells"]:
            # If the input type is a list merge them together since the formatter takes a multiline string.
            #
            if "source" in cell:
                source = cell["source"]
                if type(source) is list:
                    # If the array has embedded new lines then concatenate without adding new lines.
                    #
                    if '\n' in source[0]:
                        cell["source"] = "".join(source)
                    else:
                        cell["source"] = "\n".join(source)

        return notebook_json

    def _write_to_file(self, filename, body):
        with codecs.open(filename, mode='w', encoding='utf-8') as f:
            f.write(body)
        return

    def _cells_with_tag(self, notebook, tag):
        """
        This is used to find cells with a particular tag.  Examples would be to find where parameters should be
        inserted or to find all the generated cells in a notebook.
        """
        tagged_cell_indexes = [notebook["cells"].index(x) for x in notebook["cells"] if x["cell_type"] == 'code' and "tags" in x["metadata"] and tag in x["metadata"]["tags"]]
        return tagged_cell_indexes

    def _generate_cell(self, cell_body):
        # NOTE: Currently only python cells can be generated
        #
        cell = dict(cell_type="code", execution_count=None, outputs=[], source=[])

        cell["metadata"] = (
            {
                "tags": [
                    "generated"
                ]
            }
        )

        cell["source"].append("#----------------------------------------------------------------------------------------\n"
                              "# Generated Cell\n"
                              "#\n"
                              "# This cell contains generated code.  These generated blocks can be created for several\n"
                              "# reasons.  Examples include command line parameter injection or kernel specific bootstrap\n"
                              "# code.  The generated cells should not contain sensitive data.  To pass sensitive data\n"
                              "# to a notebook use an environment variable or use secure storage.\n"
                              "#----------------------------------------------------------------------------------------\n"
                              "#\n")
        cell["source"].append("{0}".format(cell_body))

        cell["source"] = CellFormatter.convert_list_to_string(cell["source"])

        return cell

    def _generate_parameters_cell(self, notebook, kernel_spec, arguments):
        msg = ""
        if arguments is not None and kernel_spec=="python3":
            # If the user has passed in arguments then see if the notebook has a cell with a 'parameters' tag.
            # If it does generate a new cell after the parameters cell with the passed in parameters.
            #
            # NOTE for now this only supports Python cells.
            #
            cells = self._cells_with_tag(notebook, "parameters")
            if len(cells) > 0:
                cell_body = ""
                for key in sorted(arguments):
                    cell_body += "{0} = '{1}'\n".format(key, arguments[key])

                gen_cell = self._generate_cell(CellFormatter.convert_list_to_string(cell_body))
                notebook['cells'].insert(cells[0] + 1, gen_cell)
            else:
                # None of the cells were marked for parameters.
                #
                msg += "Warning: Parameters were passed in but no cells in the target notebook were marked with parameters tag.  Parameters not inserted.\n"
        return msg

    def _generate_kernel_specific_cells(self, client, notebook, kernel_spec):
        if kernel_spec == "pysparkkernel":
            # Need to generate the _do_not_call_change_endpoint in order to setup the kernel connection to Spark.
            # This uses the Spark command module code to get the connection properties.
            #
            endpoint = get_livy_endpoint(client)

            if client.profile.active_context.security_strategy == SecurityStrategy.basic:
                username = _get_knox_username()
                password = _get_knox_password(client)
                spark_magic = "%_do_not_call_change_endpoint --username={0} --password={1} --server={2} --auth=Basic_Access".format(
                    username, password, endpoint
                )
            elif client.profile.active_context.security_strategy == SecurityStrategy.ad:
                spark_magic = "%_do_not_call_change_endpoint --server={0} --auth=Kerberos".format(endpoint)

            else:
                raise CliError("Invalid security strategy.  Use azdata login to choose supported strategy.")

            generated_cell = self._generate_cell("")

            # Replace the source of the generated cell.  The only thing in the cell can be the spark magic.
            # By having the cell comment the client tries to send the source to the server and can't connect.
            #
            generated_cell["source"] = spark_magic

            # Execute the cell instead of inserting it into the notebook.  This is because this generated cell
            # can contain passwords and that shouldn't be persisted to disk.
            #
            had_error = self.execute_cell(self.generated_cell, True, False)
            if had_error:
                raise Exception("Error initializing pyspark kernel. Check pyspark client configuration and login information.")

        return

    def _decode_dictionary(self, value):
        """
        Some parameters need to be dictionary of name value pairs.  For example configuration dictionary.
        This function checks to see if the content is a json dictionary and if so decodes it.

        Example:
            { "name": "value", "another_name": "another_value" }
        """
        if value.strip().startswith('{'):
            result = json.loads(value)
        else:
            raise ValueError("Value must be JSON encoded dictionary [{0}].".format(value))

        return result



