# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import os
import mdv
from pygments import highlight
from pygments.lexers import PythonLexer, MarkdownLexer, HtmlLexer
from pygments.formatters import TerminalFormatter, Terminal256Formatter, HtmlFormatter
from azdata.cli.commands.notebook import _
from azdata.cli.core.exceptions import CliError


class CellFormatter(object):
    @staticmethod
    def convert_list_to_string(cell_content):
        # If the input type is a list merge them together since the formatter takes a multiline string.
        if type(cell_content) is list:
            # If the array has embedded new lines then concatenate without adding new lines.
            if '\n' in cell_content[0]:
                cell_content = "".join(cell_content)
            else:
                cell_content = "\n".join(cell_content)
        return cell_content

    @staticmethod
    def _render_markdown_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)
        mdv.term_columns = 60
        output = mdv.main(cell_content, theme="pastel", c_theme="pastel")
        return output

    @staticmethod
    def _render_python_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)
        output = highlight(cell_content, PythonLexer(), Terminal256Formatter())
        return output

    @staticmethod
    def _render_output_cell(cell_content):
        cell_content = CellFormatter.convert_list_to_string(cell_content)

        # Note that there might be other types of output content and right now it's not getting formatted.
        output = "\n[Output]:\n"
        output += cell_content
        return output

    @staticmethod
    def _render_cell_boundary(cell_type, cell_number):
        return _("\n[{0: <8}] CELL [{1: >4}] ---------------------------------------------------------------------------------------\n".format(cell_type, cell_number))

    @staticmethod
    def render_cell(cell, kernel_type, cell_number):
        found_error = False
        output = CellFormatter._render_cell_boundary(cell['cell_type'], cell_number)

        if cell['cell_type'] == 'markdown':
            output += CellFormatter._render_markdown_cell(cell['source'])
        elif cell['cell_type'] == 'code':
            if kernel_type.startswith("python"):
                output += CellFormatter._render_python_cell(cell['source'])
            else:
                # If the cell type isn't known add the content without syntax highlighting
                output += cell['source']

            if len(cell['outputs']) > 0:
                for output_cell in cell['outputs']:
                    if 'text' in output_cell:
                        output += CellFormatter._render_output_cell(output_cell['text'])
                    if 'traceback' in output_cell:
                        output += CellFormatter._render_output_cell(output_cell['traceback'])
                    if output_cell['output_type'] == 'error':
                        found_error = True

        return output, found_error


def dump_buffer(buffer):
    output = ""
    length = len(buffer)
    for offset in range(0, length, 16):
        data = buffer[offset:offset + 16 if offset + 16 <= length else length - offset]
        output += '{:08X} | {:47} | {}\n'.format(offset,
                                           ' '.join('{:02X}'.format(i) for i in data),
                                           ''.join(31 < i < 127 and chr(i) or '.' for i in data))
    return output

def diff_buffers(buffer1, buffer2, show_equal = True):
    output = ""
    length = max(len(buffer1), len(buffer2))
    for offset in range(0, length, 16):
        data_buffer1 = buffer1[offset:offset + 16 if offset + 16 <= length else length - offset]
        data_buffer2 = buffer2[offset:offset + 16 if offset + 16 <= length else length - offset]

        if data_buffer1 == data_buffer2:
            equal = True
        else:
            equal = False

        if equal == False or (show_equal == True and equal == True):
            output += '{:08X} | {:47} | {:16} {:4} {:47} | {:16}\n'.format(offset,
                                           ' '.join('{:02X}'.format(i) for i in data_buffer1),
                                           ''.join(31 < i < 127 and chr(i) or '.' for i in data_buffer1),
                                           ' == ' if equal else ' XX ',
                                           ' '.join('{:02X}'.format(i) for i in data_buffer2),
                                           ''.join(31 < i < 127 and chr(i) or '.' for i in data_buffer2))

    return output


def get_livy_endpoint(client):
    """
    Call the controller to get the webhdfs endpoint URL
    """
    try:
        endpoint = client.apis.controller.get_endpoints('livy')['endpoint']

        # The endpoint must end with '/' in order for the APIs to succeed
        if not endpoint[len(endpoint) - 1] == '/':
            endpoint = endpoint + '/'
    except Exception as e:
        raise CliError(e)

    return endpoint


class WorkingDirectoryContext(object):
    def __init__(self, working_path):
        self.working_path = os.path.expanduser(working_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.working_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)

