# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import re
from decimal import Decimal
from typing import Union
from azdata.cli.core.models.custom_resource import CustomResource
from azdata.cli.commands.postgres.constants import (
    POSTGRES_MIN_CORES_SIZE,
    POSTGRES_MIN_MEMORY_SIZE)
from azdata.cli.core.models.dict_utils import SerializationUtils
from azdata.cli.core.constants import DNS_NAME_REQUIREMENTS
from azdata.cli.core.models.kube_quantity import KubeQuantity
from azdata.cli.core.util import name_meets_dns_requirements
from azdata.cli.commands.postgres.constants import RESOURCE_KIND_12, RESOURCE_KIND_11
from azdata.cli.core.class_utils import (
    validator,
    validatedclass)
from typing import TYPE_CHECKING
# KubernetesClient is only needed for typehints, but causes a circular import. This is the python provided workaround
if TYPE_CHECKING:
    from azdata.cli.core.clients.kubernetes_client import KubernetesClient


@validatedclass
class PostgresqlCustomResource(CustomResource):
    """
    Internal Postgres Custom Resource object to be used for deployments.
    """
    def __init__(self):
        super().__init__()

    class Spec(CustomResource.Spec):
        """
        CustomResource.spec
        @override
        """
        def __init__(self):
            super().__init__()
            self.scheduling = self.Scheduling()
            self.engine = self.Engine()
            self.scale = self.Scale()

        class Engine(SerializationUtils):
            """
            CustomResource.spec.engine
            """
            def __init__(self):
                self.extensions = []
                # Need to initialize directly rather than going through setter, since setting this value typically
                # uses the existing value
                self._settings = {"default":{}}

            @property
            def extensions(self) -> list:
                return self._extensions

            @extensions.setter
            def extensions(self, e: Union[list, str]):
                if type(e) is str:
                    self._extensions = self._parse_engine_extensions_string(e)
                elif type(e) is list:
                    self._extensions = e
                else:
                    raise ValueError("Type '{}' is not compatible with property 'extentions'. Need 'str', or 'list'".format(type(e)))

            @property
            def settings(self) -> dict:
                return self._settings

            @settings.setter
            def settings(self, s: Union[str, dict]):
                self.set_engine_settings(s, True)

            def set_engine_settings(self, s: Union[str,  dict], overwrite=False):
                """
                Sets the engine settings from either a settings string or a dictionary, and optionally overwrites the
                existing settings completely. (Merges settings by default)
                :param s: the new settings
                :param overwrite: False by default. If True eliminates all existing settings before replacing them, if
                False this function retains existing settings and will overwrite or add any provided keys where appropriate
                """
                if type(s) is str:
                    self._settings["default"] = self._get_new_engine_settings(self._parse_engine_settings_string(s), overwrite)
                elif type(s) is dict:
                    self._settings["default"] = self._get_new_engine_settings(
                        s["default"].items() if "default" in s else s.items(), overwrite)
                else:
                    raise TypeError(
                        "Invalid type '{}' for CustomResource.engine.settings. valid types are 'str' and 'dict'".format(type(s)))

            def get_settings_string(self):
                """
                Returns any settings in a similar format to how they're provided via commandline e.g. "key=val,...,"
                format.
                """
                if not hasattr(self, "_settings"):
                    return ""
                else:
                    return ",".join(map(lambda x: "{}={}".format(x[0], x[1]), self._settings["default"].items()))

            def _get_new_engine_settings(self, engine_settings, overwrite_engine_settings: bool):
                """
                Computes new engine settings based on the provided engine settings, the existing settings, and the overwrite
                flag. Note: Does NOT mutate the underlying dictionary
                :param engine_settings: the settings that are being provided
                :param overwrite_engine_settings: If True existing settings are wiped out, if False they are retained and
                overridden where appropriate
                :returns: A dictionary with what the new settings would be based on the provided settings and the overwrite
                flag.
                """
                base_settings = self.settings["default"] if not overwrite_engine_settings else {}
                # Must be careful to avoid mutating underlying dict or this method could cause undocumented side effects
                #
                new_settings = base_settings.copy()
                for k, v in engine_settings:
                    if not v:
                        new_settings[k] = None
                    else:
                        new_settings[k] = v
                for k in list(new_settings.keys()):
                    if new_settings[k] is None:
                        del new_settings[k]

                return new_settings

            @staticmethod
            def _parse_engine_settings_string(engine_settings: str):
                """
                Given a settings string in format "key=val,key2=val2,..." returns the settings as a list of tuples of
                the form [(key, val),(key2, val2),...]
                :returns: A list of tuples representing the provided settings string
                """
                return re.findall(r'\s*([^\s=,]+)\s*=\s*("(?:[^"]|"")*"|[^\s",]*),?',
                                  engine_settings.replace("'", "").replace('"', ""))

            @staticmethod
            def _parse_engine_extensions_string(extensions: str):
                return [e.strip() for e in extensions.replace("\"", "").strip().split(",")]

            def _to_dict(self) -> dict:
                return {
                    "extensions": [{"name": e} for e in self.extensions],
                    "settings": self.settings
                }

            def _hydrate(self, d: dict):
                if "extensions" in d:
                    exts = d["extensions"]
                    if type(exts) is list and len(exts) > 0 and type(exts[0]) is dict:
                        self.extensions = [e["name"] for e in exts]
                    else:
                        self.extensions = d["extensions"]

                if "settings" in d:
                    self.settings = d["settings"]

        @property
        def engine(self) -> Engine:
            return self._engine

        @engine.setter
        def engine(self, e: Engine):
            self._engine = e

        class Scheduling(SerializationUtils):
            """
            CustomResource.spec.scheduling
            """
            def __init__(self):
                # If you add a nested class, please create a property and initialize it here
                #
                self.default = self.Default()

            class Default(SerializationUtils):
                """
                CustomResource.spec.scheduling.default
                """
                def __init__(self):
                    self.resources = self.Resources()

                class Resources:
                    """
                    CustomResource.spec.scheduling.default.resources
                    """
                    def __init__(self):
                        self.requests = self.Requests()
                        self. limits = self.Limits()

                    class Requests(SerializationUtils):
                        """
                        CustomResource.spec.scheduling.default.resources.requests
                        """
                        @property
                        def memory(self) -> KubeQuantity:
                            return self._memory

                        @memory.setter
                        def memory(self, m: Union[str, KubeQuantity]):
                            if type(m) is str and m == '':
                                self._memory = None
                                return

                            val = KubeQuantity(m)
                            if val < POSTGRES_MIN_MEMORY_SIZE:
                                raise ValueError("Postgres memory request must be at least '{}'".format(POSTGRES_MIN_MEMORY_SIZE.quantity))
                            self._memory = KubeQuantity(m)

                        @property
                        def cpu(self) -> KubeQuantity:
                            return self._cpu

                        @cpu.setter
                        def cpu(self, c: Union[str, KubeQuantity]):
                            if type(c) is str and c == '':
                                self._cpu = None
                                return

                            val = KubeQuantity(c)
                            if val < POSTGRES_MIN_CORES_SIZE:
                                raise ValueError("Postgres cores request must be at least '{}'".format(POSTGRES_MIN_CORES_SIZE.quantity))
                            self._cpu = val

                        def _to_dict(self):
                            """
                            @override
                            """
                            mem = getattr(self, "memory", None)
                            cores = getattr(self, "cpu", None)
                            return {
                                "memory": mem.quantity if mem is not None else mem,
                                "cpu": cores.quantity if cores is not None else cores
                            }

                        def _hydrate(self, d: dict):
                            if "memory" in d:
                                self.memory = d["memory"]

                            if "cpu" in d:
                                self.cpu = d["cpu"]

                    @property
                    def requests(self) -> Requests:
                        return self._requests

                    @requests.setter
                    def requests(self, r: Requests):
                        self._requests = r

                    class Limits(SerializationUtils):
                        """
                        CustomResource.spec.scheduling.default.resources.limits
                        """
                        @property
                        def memory(self) -> KubeQuantity:
                            return self._memory

                        @memory.setter
                        def memory(self, m: Union[str, KubeQuantity]):
                            if type(m) is str and m == '':
                                self._memory = None
                                return

                            val = KubeQuantity(m)
                            if val < POSTGRES_MIN_MEMORY_SIZE:
                                raise ValueError("Postgres memory limit must be at least '{}'".format(POSTGRES_MIN_MEMORY_SIZE.quantity))
                            self._memory = val

                        @property
                        def cpu(self) -> KubeQuantity:
                            return self._cpu

                        @cpu.setter
                        def cpu(self, c: Union[str, KubeQuantity]):
                            if type(c) is str and c == '':
                                self._cpu = None
                                return

                            val = KubeQuantity(c)
                            if val < POSTGRES_MIN_CORES_SIZE:
                                raise ValueError("Postgres memory limit must be at least '{}'".format(POSTGRES_MIN_CORES_SIZE.quantity))
                            self._cpu = val

                        def _to_dict(self):
                            """
                            @override
                            """
                            mem = getattr(self, "memory", None)
                            cores = getattr(self, "cpu", None)
                            return {
                                "memory": mem.quantity if mem is not None else mem,
                                "cpu": cores.quantity if cores is not None else cores
                            }

                        def _hydrate(self, d: dict):
                            """
                            @override
                            """
                            if "memory" in d:
                                self.memory = d["memory"]

                            if "cpu" in d:
                                self.cpu = d["cpu"]

                    @property
                    def limits(self) -> Limits:
                        return self._limits

                    @limits.setter
                    def limits(self, r: Limits):
                        self._limits = r

                    def _to_dict(self):
                        return {
                            "limits": self.limits._to_dict(),
                            "requests": self.requests._to_dict()
                        }

                    def _hydrate(self, d: dict):
                        if "limits" in d:
                            self.limits._hydrate(d["limits"])
                        if "requests" in d:
                            self.requests._hydrate(d["requests"])

                @property
                def resources(self) -> Resources:
                    return self._resources

                @resources.setter
                def resources(self, r: Resources):
                    self._resources = r

                def _to_dict(self) -> dict:
                    return {
                        "resources": self.resources._to_dict()
                    }

                def _hydrate(self, d: dict):
                    if "resources" in d:
                        self.resources._hydrate(d["resources"])

            @property
            def default(self) -> Default:
                return self._default

            @default.setter
            def default(self, d: Default):
                self._default = d

            def _to_dict(self) -> dict:
                return {
                    "default": self.default._to_dict()
                }

            def _hydrate(self, d: dict):
                if "default" in d:
                    self.default._hydrate(d["default"])

        @property
        def scheduling(self) -> Scheduling:
            return self._scheduling

        @scheduling.setter
        def scheduling(self, s: Scheduling):
            self._scheduling = s

        class Scale(SerializationUtils):
            """
            Namespaces custom resource spec properties relating to how resources should scale

            CustomResource.spec.scale
            """
            def __init__(self, shards: int = None):
                self._shards = shards

            @property
            def shards(self) -> int:
                """
                A database custom resource property - How many partitions a table should be spread out across
                """
                return self._shards

            @shards.setter
            def shards(self, s: Union[int, str]):
                self._shards = int(s)

            def _hydrate(self, d: dict):
                """
                @override
                """
                if "workers" in d:
                    self.shards = d["workers"]

            def _to_dict(self):
                """
                @override
                """
                return {
                    "workers": self.shards,
                }

        @property
        def scale(self) -> Scale:
            return self._scale

        @scale.setter
        def scale(self, s: Scale):
            self._scale = s

        def _to_dict(self) -> dict:
            base = super()._to_dict()
            base["engine"] = self.engine._to_dict()
            base["scheduling"] = self.scheduling._to_dict()
            base["scale"] = self.scale._to_dict()
            return base

        def _hydrate(self, d: dict):
            super()._hydrate(d)
            if "engine" in d:
                self.engine._hydrate(d["engine"])

            if "scheduling" in d:
                self.scheduling._hydrate(d["scheduling"])

            if "scale" in d:
                self.scale._hydrate(d["scale"])

    class Metadata(CustomResource.Metadata):
        """
        CustomResource.metadata
        @override
        """
        arc_psql_name_max_length = 11

        @CustomResource.Metadata.name.setter
        def name(self, n: str):
            if not n:
                raise ValueError("Azure Arc enabled PostgreSQL Hyperscale server group name cannot be empty")

            if len(n) > self.arc_psql_name_max_length:
                raise ValueError("Azure Arc enabled PostgreSQL Hyperscale server group name '{}' exceeds {} character length limit"
                                .format(n, self.arc_psql_name_max_length))

            if not name_meets_dns_requirements(n):
                raise ValueError("Azure Arc enabled PostgreSQL Hyperscale server group name '{}' does not follow DNS requirements: {}"
                                .format(n, DNS_NAME_REQUIREMENTS))

            if not name_meets_dns_requirements(n):
                raise ValueError(
                    "Postgres server name '{}' does not follow DNS requirements: {}".format(n, DNS_NAME_REQUIREMENTS))

            self._name = n

    class Status(CustomResource.Status):
        """
        @override CustomResource.Status
        """
        def __init__(self):
            super().__init__()

        @property
        def readyPods(self) -> int:
            return self._readyPods

        @readyPods.setter
        def readyPods(self, rp: int):
            self._readyPods = rp

        def _hydrate(self, d: dict):
            """
            @override
            """
            super()._hydrate(d)
            if "readyPods" in d:
                self.readyPods = d["readyPods"]

        def _to_dict(self):
            """
            @override
            """
            base = super()._to_dict()
            base["readyPods"] = getattr(self, "readyPods", None)
            return base



    def apply_args(self, replace_engine_settings=False, **kwargs):
        super().apply_args(**kwargs)
        self._set_if_provided(self.spec.scheduling.default.resources.requests, "memory", kwargs, "memory_request")
        self._set_if_provided(self.spec.scheduling.default.resources.requests, "cpu", kwargs, "cores_request")
        self._set_if_provided(self.spec.scheduling.default.resources.limits, "memory", kwargs, "memory_limit")
        self._set_if_provided(self.spec.scheduling.default.resources.limits, "cpu", kwargs, "cores_limit")

        if "engine_version" in kwargs:
            kind = self._get_resource_kind_by_engine_version(kwargs["engine_version"])
            if not kind:
                raise ValueError("Unsupported engine version '{}'".format(kwargs["engine_version"]))
            else:
                self.kind = kind

        self._set_if_provided(self.spec.scale, "shards", kwargs, "workers")
        self._set_if_provided(self.spec.engine, "extensions", kwargs, "extensions")
        if "engine_settings" in kwargs and kwargs["engine_settings"]:
            self.spec.engine.set_engine_settings(kwargs["engine_settings"], replace_engine_settings)

    def _to_dict(self) -> dict:
        return super()._to_dict()

    def _hydrate(self, d: dict):
        super()._hydrate(d)

    @validator
    def _validate_cores(self, client: 'KubernetesClient'):
        """
        Verifies the cores request does not exceed the limit
        """
        request = getattr(self.spec.scheduling.default.resources.requests, "cpu", None)
        lim = getattr(self.spec.scheduling.default.resources.limits, "cpu", None)

        # Lim and request could be empty strings if they're being deleted from the settings
        if lim and request and type(lim) is KubeQuantity and type(request) is KubeQuantity:
            if lim < request:
                raise ValueError(
                    "Cores request of %s cannot exceed cores limit of %s" % (request.quantity, lim.quantity))

    @validator
    def _validate_memory(self, client: 'KubernetesClient'):
        """
        verifies the memory request does not exceed the limit
        """
        request = getattr(self.spec.scheduling.default.resources.requests, "memory", None)
        lim = getattr(self.spec.scheduling.default.resources.limits, "memory", None)

        # Lim and request could be empty strings if they're being deleted from the settings
        if lim and request and type(lim) is KubeQuantity and type(request) is KubeQuantity:
            if request > lim:
                raise ValueError(
                    "Memory request of %s cannot exceed memory limit of %s" % (request.quantity, lim.quantity))

    @staticmethod
    def _get_resource_kind_by_engine_version(engine_version: int = None):
        """
        Returns the resource kind defined in the corresponding CRD for the specified postgresql engine version.
        It's possible to create multiple PostgreSQL server groups with the same name but different engine version.
        :param engine_version: The postgresql engine version. If not specified, the latest supported engine version will be used.
        :return:
        """
        if engine_version == 11:
            return RESOURCE_KIND_11
        elif engine_version == 12 or engine_version is None:  # Default to the latest supported version
            return RESOURCE_KIND_12

        return None
