from azdata.cli.core.models.custom_resource import CustomResource
from azdata.cli.core.models.kube_quantity import KubeQuantity
from azdata.cli.core.class_utils import (validator, validatedclass)
from azdata.cli.commands.sqlmi.constants import (
    SQLMI_MIN_MEMORY_SIZE,
    SQLMI_MIN_CORES_SIZE)
from azdata.cli.core.constants import (
    DNS_NAME_REQUIREMENTS)
from azdata.cli.core.util import name_meets_dns_requirements
from azdata.cli.core.class_utils import enforcetype
from azdata.cli.core.models.volume_claim import VolumeClaim
from azdata.cli.core.models.dict_utils import SerializationUtils
from typing import Union, TYPE_CHECKING
# KubernetesClient is only needed for typehints, but causes a circular import. This is the python provided workaround
if TYPE_CHECKING:
    from azdata.cli.core.clients.kubernetes_client import KubernetesClient

TYPE_ERROR = "Type \'{}\' is incompatible with property \'{}\'"


@validatedclass
class SqlmiCustomResource(CustomResource):
    """
    Internal Sqlmi Custom Resource object to be used for deployments.
    """
    def __init__(
            self,
            spec: 'SqlmiCustomResource.Spec' = None,
            metadata: 'SlqlmiCustomResource.Metadata' = None,
            *args,
            **kwargs):
        """
        Initializes a CR object with the given json.
        """
        super().__init__(*args, **kwargs)
        self.spec = spec if spec else self.Spec()
        self.metadata = metadata if metadata else self.Metadata()

    class Spec(CustomResource.Spec):
        """
        @override CustomResource.spec
        """
        def __init__(
                self,
                serviceType: str = None,
                limits: 'SqlmiCustomResource.Limits' = None,
                requests: 'SqlmiCustomResource.Requests' = None,
                *args,
                **kwargs):
            super().__init__(*args, **kwargs)
            self.serviceType = serviceType
            self.limits = limits if limits else self.Limits()
            self.requests = requests if requests else self.Requests()

        @property
        def serviceType(self) -> str:
            return self._serviceType

        @serviceType.setter
        def serviceType(self, s: str):
            self._serviceType = s

        class Storage(CustomResource.Spec.Storage):
            """
            @override CustomResource.spec.storage
            """
            def __init__(self, datalogs: VolumeClaim = None, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.datalogs = datalogs if datalogs else VolumeClaim()

            @property
            def datalogs(self) -> VolumeClaim:
                return self._datalogs

            @datalogs.setter
            @enforcetype(VolumeClaim)
            def datalogs(self, vc):
                self._datalogs = vc

            def _hydrate(self, d: dict):
                """
                @override
                """
                super()._hydrate(d)
                if "datalogs" in d:
                    self.datalogs._hydrate(d["datalogs"])

            def _to_dict(self):
                """
                @override
                """
                base = super()._to_dict()
                base["datalogs"] = self.datalogs._to_dict()
                return base

        class Limits(SerializationUtils):
            """
            Namespaces the resource limits part of this custom resource spec. The custom resource is not guaranteed these
            resources, but will never exceed them
            @see https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#requests-and-limits
            CustomResource.spec.limits
            """

            def __init__(self, vcores: Union[str, KubeQuantity] = None, memory: Union[str, KubeQuantity] = None):
                if vcores:
                    self.vcores = vcores
                if memory:
                    self.memory = memory

            @property
            def vcores(self) -> KubeQuantity:
                """
                cores limit
                """
                return self._vcores

            @vcores.setter
            def vcores(self, vc: Union[str, KubeQuantity]):
                val = KubeQuantity(vc)
                if val < SQLMI_MIN_CORES_SIZE:
                    raise ValueError("Sqlmi cores limit must be at least '{}'".format(SQLMI_MIN_CORES_SIZE.quantity))

                self._vcores = val

            @property
            def memory(self) -> KubeQuantity:
                """
                memory limit
                """
                return self._memory

            @memory.setter
            def memory(self, m: Union[str, KubeQuantity]):
                val = KubeQuantity(m)
                if val < SQLMI_MIN_MEMORY_SIZE:
                    raise ValueError("Sqlmi memory limit must be at least '{}'".format(SQLMI_MIN_MEMORY_SIZE.quantity))

                self._memory = val

            def _to_dict(self):
                mem = getattr(self, "memory", None)
                cores = getattr(self, "vcores", None)
                return {
                    "memory": mem.quantity if mem is not None else mem,
                    "vcores": cores.quantity if cores is not None else cores
                }

            def _hydrate(self, d: dict):
                if "memory" in d:
                    self.memory = d["memory"]

                if "vcores" in d:
                    self.vcores = d["vcores"]

        @property
        def limits(self) -> Limits:
            return self._limits

        @limits.setter
        @enforcetype(Limits)
        def limits(self, spec: Limits):
            self._limits = spec

        class Requests(SerializationUtils):
            """
            Namespaces the resource requests part of this custom resource spec. The custom resource is guaranteed these
            minimum resources from kubernetes
            @see https://kubernetes.io/docs/concepts/configuration/manage-resources-containers/#requests-and-limits
            CustomResource.spec.requests
            """
            def __init__(self, vcores: Union[str, KubeQuantity] = None, memory: Union[str, KubeQuantity] = None):
                if vcores:
                    self.vcores = vcores
                if memory:
                    self.memory = memory

            @property
            def vcores(self) -> KubeQuantity:
                """
                Cores request
                """
                return self._vcores

            @vcores.setter
            def vcores(self, vc: Union[str, KubeQuantity]):
                val = KubeQuantity(vc)
                if val < SQLMI_MIN_CORES_SIZE:
                    raise ValueError("Sqlmi cores request must be at least '{}'".format(SQLMI_MIN_CORES_SIZE.quantity))

                self._vcores = val

            @property
            def memory(self) -> KubeQuantity:
                """
                Memory request
                """
                return self._memory

            @memory.setter
            def memory(self, m: Union[str, KubeQuantity]):
                val = KubeQuantity(m)
                if val < SQLMI_MIN_MEMORY_SIZE:
                    raise ValueError(
                        "Sqlmi memory request must be at least '{}'".format(SQLMI_MIN_MEMORY_SIZE.quantity))

                self._memory = val

            def _to_dict(self):
                mem = getattr(self, "memory", None)
                cores = getattr(self, "vcores", None)
                return {
                    "memory": mem.quantity if mem is not None else mem,
                    "vcores": cores.quantity if cores is not None else cores
                }

            def _hydrate(self, d: dict):
                if "memory" in d:
                    self.memory = d["memory"]

                if "vcores" in d:
                    self.vcores = d["vcores"]

        @property
        def requests(self):
            """
            @see Limits
            """
            return self._requests

        @requests.setter
        @enforcetype(Requests)
        def requests(self, spec: Requests):
            """
            @see Requests
            """
            self._requests = spec

        def _hydrate(self, d: dict):
            super()._hydrate(d)
            if "serviceType" in d:
                self.serviceType = d["serviceType"]
            if "limits" in d:
                self.limits._hydrate(d["limits"])
            if "requests" in d:
                self.requests._hydrate(d["requests"])

        def _to_dict(self):
            base = super()._to_dict()
            base["serviceType"] = getattr(self, "serviceType", None)
            base["limits"] = self.limits._to_dict()
            base["requests"] = self.requests._to_dict()
            return base

    class Metadata(CustomResource.Metadata):
        """
        @override CustomResource.metadata
        """
        arc_sql_name_max_length = 13

        def __init__(self, name: str = None, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @CustomResource.Metadata.name.setter
        def name(self, n: str):
            """
            @override CustomResource.metadata.name.setter
            """
            if not n:
                raise ValueError("SQL MI name cannot be empty")

            if len(n) > self.arc_sql_name_max_length:
                raise ValueError(
                    "SQL MI name '{}' exceeds {} character length limit".format(n, self.arc_sql_name_max_length))

            if not name_meets_dns_requirements(n):
                raise ValueError(
                    "SQL MI name '{}' does not follow DNS requirements: {}".format(n, DNS_NAME_REQUIREMENTS))

            self._name = n

        def _hydrate(self, d: dict):
            super()._hydrate(d)

        def _to_dict(self):
            return super()._to_dict()

    class Status(CustomResource.Status):
        """
        @override CustomResource.Status
        """
        def __init__(self):
            super().__init__()

        @property
        def readyReplicas(self) -> int:
            """
            If this resource is a ReplicaSet, then how many replicas of this resource are available
            @see https://kubernetes.io/docs/concepts/workloads/controllers/replicaset/
            """
            return self._readyReplicas

        @readyReplicas.setter
        def readyReplicas(self, rp: int):
            self._readyReplicas = rp

        def _hydrate(self, d: dict):
            """
            @override
            """
            super()._hydrate(d)
            if "readyReplicas" in d:
                self.readyReplicas = d["readyReplicas"]

        def _to_dict(self):
            """
            @override
            """
            base = super()._to_dict()
            base["readyReplicas"] = getattr(self, "readyReplicas", None)
            return base

    def _hydrate(self, d: dict):
        """
        @override
        """
        super()._hydrate(d)

    def _to_dict(self):
        """
        @override
        """
        return super()._to_dict()


    @validator
    def _validate_storage_classes(self, client: 'KubernetesClient'):
        """
        Extends the super implementation of this to account for the datalogs volume
        @override CustomResource._validate_storage_classes
        """
        super()._validate_storage_classes(client)
        STORAGE_CLASS_ERROR = "Storage class '{}' does not exist"

        datalogs = getattr(self.spec.storage, "datalogs", None)
        if datalogs and datalogs.className and not client.storage_class_exists(datalogs.className):
            raise ValueError(STORAGE_CLASS_ERROR.format(datalogs.className))

    @validator
    def _validate_cores(self, client: 'KubernetesClient'):
        """
        Verifies the cores request does not exceed the limit
        """
        request = getattr(self.spec.requests, "vcores", None)
        lim = getattr(self.spec.limits, "vcores", None)

        # Lim and request could be empty strings if they're being deleted from the settings
        if lim and request and type(lim) is KubeQuantity and type(request) is KubeQuantity:
            if lim < request:
                raise ValueError("Cores request of %s cannot exceed cores limit of %s" % (request.quantity, lim.quantity))

    @validator
    def _validate_memory(self, client: 'KubernetesClient'):
        """
        verifies the memory request does not exceed the limit
        """
        request = getattr(self.spec.requests, "memory", None)
        lim = getattr(self.spec.limits, "memory", None)

        # Lim and request could be empty strings if they're being deleted from the settings
        if lim and request and type(lim) is KubeQuantity and type(request) is KubeQuantity:
            if request > lim:
                raise ValueError("Memory request of %s cannot exceed memory limit of %s" % (request.quantity, lim.quantity))

    def apply_args(self, **kwargs):
        super().apply_args(**kwargs)
        self._set_if_provided(self.spec.requests, "memory", kwargs, "memory_request")
        self._set_if_provided(self.spec.requests, "vcores", kwargs, "cores_request")
        self._set_if_provided(self.spec.limits, "memory", kwargs, "memory_limit")
        self._set_if_provided(self.spec.limits, "vcores", kwargs, "cores_limit")
        self._set_if_provided(self.spec.storage.datalogs, "className", kwargs, "storage_class_datalogs")
        self._set_if_provided(self.spec.storage.datalogs, "size", kwargs, "storage_size_datalogs")