# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
import re
from collections import OrderedDict
from azdata.cli.commands.sqlmi.constants import (
    SQLMI_PASSWORD_CHARS,
    SQLMI_PASSWORD_MIN_LENGTH,
    SQLMI_PASSWORD_REQUIRED_GROUPS)


def is_valid_sql_password(pw, user):
    """
    Checks if the provided pw is a valid sql server password i.e. is at least
    eight characters long and contains a char from at least three of these groups
    -Uppercase letters
    -Lowercase letters
    -Base 10 digits
    -Non-alphanumeric characters
    :param pw: the password
    :param user: username for the sql instance
    :return: True if pw meets requirements, False otherwise
    """
    if not pw:
        return False

    if user in pw:
        return False

    if len(pw) < SQLMI_PASSWORD_MIN_LENGTH:
        return False

    lower = 0
    upper = 0
    special = 0
    digit = 0

    for c in pw:
        if c.isdigit():
            digit = 1
        elif c.isalpha():
            if c.isupper():
                upper = 1
            else:
                lower = 1

    special_chars = re.compile(SQLMI_PASSWORD_CHARS)
    if special_chars.search(pw) is not None:
        special = 1

    return (lower + upper + special + digit) >= SQLMI_PASSWORD_REQUIRED_GROUPS


def order_endpoints():
    """
    Order SQL instance `dict` sections to the same order the server API handed us.
    NOTE: This is redundant in Python 3.7 however needed for earlier versions.

    :return: A well defined `OrderedDict` of the given SQL instance endpoints.
    """

    def get_endpoints(endpoints):
        """
        Creates ordered dictionaries for the given endpoints to be used in the BoxLayout.
        :param endpoints:
        :return:
        """
        def new_endpoint(e):
            return OrderedDict([
                ("description", e["description"]),
                ("endpoint", e["endpoint"]),
                ("options", [])
            ])

        return [new_endpoint(endpoint) for endpoint in endpoints]

    def get_instances(obj):
        """
        Returns all instances and their endpoints.
        :param obj:
        :return:
        """
        obj = obj if obj else []
        return [OrderedDict([
            ("instanceName", instance["name"]),
            ("endpoints", get_endpoints(instance.get("endpoints")))
        ]) for instance in obj]

    def get_arc_sql_endpoints(obj):
        """
        Retrieves all SQL instance endpoints in an ordered dictionary to be used in the BoxLayout.
        :param obj:
        :return:
        """
        return None if "namespace" not in obj else OrderedDict([
            ("clusterName", obj["namespace"]),
            ("instance", get_instances(obj["instances"]))
        ])

    return get_arc_sql_endpoints


def hierarchical_output(command_result):
    """
    Callback for formatting complex custom-output.
    :parm_am command_result: The command's high-level result object.
    :return: Complex BoxLayout otherwise flat json.
    """
    from azdata.cli.core.layout import BoxLayout

    try:
        raw_result = command_result.result
        result = order_endpoints()(raw_result)

        return BoxLayout(result, config={
            'headers': {
                'left': {
                    'label': '',
                    'id': None
                },
                'right': {
                    'label': '',
                    'id': None
                }
            },
            'identifiers': []
        }, bdc_config=True)
    except Exception as e:  # -- fallback --
        from knack.output import format_json
    return format_json(command_result)
