# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""
Client for all CLI actions.
"""

from azdata.cli.core.output import OutputStream
from azdata.cli.core.profile import CliProfile
from azdata.cli.core.enums import ContextType
from azdata.cli.core.clients.process_client import Command as ProcessClient
from azdata.cli.core.clients.controller_client import ControllerClient
from azdata.cli.core.clients.control_deploy_client import ControlDeployClient
from azdata.cli.core.clients.kerberos_client import KerberosClient
from azdata.cli.core.clients.kubernetes_client import KubernetesClient
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.util import (eula_prompt, is_windows)
from abc import ABCMeta
from six import add_metaclass
from importlib import import_module
from inspect import getmodule

__all__ = ['client', 'CliClient']


def beget_cli_client(check_auth=True, check_eula=True):
    """
    The factory function used to apply the common `CliClient` to a custom
    commands's command group.

    :param check_auth: Check for authentication on this command group.
    :param check_eula: Check for eula acceptance on this command group.
    :return: A function.
    """
    def beget(_):
        return CliClient(check_auth=check_auth, check_eula=check_eula)

    return beget

client = beget_cli_client  # Export

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


@add_metaclass(ABCMeta)
class BaseCliClient(object):
    def __init__(self, check_auth=True, check_eula=True, cmd_group=None):
        self._namespace = getmodule(self).__package__
        self._profile = CliProfile()
        if check_eula:
            accept_eula = self.profile.accept_eula
            if not accept_eula:
                self.profile.accept_eula = eula_prompt(self)

        if check_auth:
            if not self.profile.active_context or \
                    not self.profile.active_context.endpoint or \
                    not self.profile.active_context.access_token:
                err = 'Please run \'{0} login\' to first authenticate, or ' \
                      'set a valid active context with \'{0} context set\'.'.\
                    format(Configuration().CLI_NAME)
                raise CliError(err)
            if cmd_group == ContextType.bdc.name and \
                            self.profile.active_context.context_type != \
                            ContextType.bdc.name:
                err = 'The context type you are logged into is currently' \
                      ' \'{}\'. This does not support SQL Server Big Data ' \
                      'Cluster authenticated commands.'\
                    .format(self.profile.active_context.context_type)
                raise CliError(err)
            if cmd_group == ContextType.arc.name and \
                            self.profile.active_context.context_type != \
                            ContextType.arc.name:
                err = 'The context type you are logged into is currently ' \
                      '\'{}\'. This does not support Arc authenticated ' \
                      'commands.'.\
                    format(self.profile.active_context.context_type)
                raise CliError(err)

    @property
    def profile(self):
        return self._profile

    @property
    def stdout(self):
        return OutputStream().stdout.write

    @property
    def stderr(self):
        return OutputStream().stderr.write

    @property
    def namespace(self):
        return self._namespace

    @property
    def plugins(self):
        """
        Discoverers all plugins that a command might implement. This allows for
        the ability to provide customizations or extra features depending on the
        presence of reserved modules in a command's namespace package.

        Supported plugins are:
        - api_service: Communicating API server through different HttpClient
                       methods.
        """

        def _import_module(name):
            try:
                return import_module('{}.{}'.format(self.namespace, name))
            except ImportError:
                return None

        return {
            'services': _import_module('controller_service')
            # -- Add more here when needed --
        }

    def __str__(self):
        """
        Returns the base string representation of attributes. Sub-class should
        override and implement.
        """
        return '<BaseCliClient>'

    def __repr__(self):
        """For `print` and `pprint`. Sub-class should override and implement."""
        return self.__str__()

# ============================================================================ #
# ============================================================================ #
# ============================================================================ #


class CliClient(BaseCliClient):
    """
    Default client injected in every command group. For further command
    customization extend this class.
    """
    def __init__(self, check_auth=True, check_eula=True, cmd_group=None):
        super(CliClient, self).__init__(check_auth, check_eula, cmd_group)

        ac = self.profile.active_context

        # -- exposed client APIS --
        # follow the same pattern when more "clients" need to be introduced
        self._apis = type('', (object,), {
            'process': ProcessClient,
            'control_deploy': ControlDeployClient(self, ac.namespace),
            'kerberos': KerberosClient(ac.endpoint, principal=ac.username),
            'kubernetes': KubernetesClient(),
            'controller': ControllerClient(
                ac.endpoint,
                access_token=ac.access_token,
                services=self.plugins['services']
            )
        })

        self._utils = None
        self._terminal = None

    @property
    def apis(self):
        """
        Gets the reference to the different API resource clients.
        """
        return self._apis

    @property
    def request(self):
        """
        Utility to acquire a low-level request function to preform general HTTP
        operations.

        :return: A request function.
        """

        def _request(resource_path,
                     method,
                     path_params=None,
                     query_params=None,
                     header_params=None,
                     body=None,
                     post_params=None,
                     files=None,
                     response_type=None,
                     auth_settings=None,
                     return_http_data_only=None,
                     collection_formats=None,
                     preload_content=True,
                     request_timeout=None):
            """
            Makes the HTTP request (synchronous) and returns deserialized data.

            :param resource_path: Path to method endpoint.
            :param method: Method to call.
            :param path_params: Path parameters in the url.
            :param query_params: Query parameters in the url.
            :param header_params: Parameters to be placed in the request header.
            :param body: body: Request body.
            :param post_params: Request post form parameters, for
                                `application/x-www-form-urlencoded`,
                                `multipart/form-data`.
            :param files: key -> filename, value -> filepath, for
                          `multipart/form-data`.
            :param response_type: Response data type.
            :param auth_settings: Auth Settings names for the request.
            :param return_http_data_only: response data without head status code
                                          and headers
            :param collection_formats: dict of collection formats for path,
            query,
                                       header, and post parameters.
            :param preload_content: if False, the urllib3.HTTPResponse object
            will
                                    be returned without reading/decoding
                                    response
                                    data. Default is True.
            :param request_timeout: timeout setting for this request. If one
                                    number provided, it will be total request
                                    timeout. It can also be a pair (tuple) of
                                    (connection, read) timeouts.
            :return: The response directly.
            """

            return self.apis.controller.api_client.call_api(
                resource_path,
                method,
                path_params=path_params,
                query_params=query_params,
                header_params=header_params,
                body=body,
                post_params=post_params,
                files=files,
                response_type=response_type if response_type else object,
                auth_settings=auth_settings if auth_settings else [],
                _return_http_data_only=return_http_data_only,
                collection_formats=collection_formats,
                _preload_content=preload_content,
                _request_timeout=request_timeout
            )

        return _request

    @property
    def terminal(self):
        """
        Object mapping to supported public `terminal` operations.

         Supported:
        - `progress_indicator`

        Example:

        ```
        progress = client.terminal.progress_indicator
        ...
        ...
        ```

        :return: The Object mapping to supported terminal operations.
        """
        if self._terminal:
            return self._terminal

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------

        class Progress(object):
            """
            Show a spinner on the terminal that automatically
            starts animating around a provided worker function.

            Example:

            ```
            add = lambda a, b: return a + b
            args = {'a': 1, 'b': 2}

            progress = client.terminal.progress_indicator
            result = progress.message('Downloading').worker(
            add, args).start()
            ```

            :return: A `Progress` instance to load-up and start.
            """
            def __init__(self):
                self._defaults()

            def _defaults(self):
                self._show_time = True
                self._worker = {'fn': None, 'args': {}}
                self._message = ''

            def worker(self, fn, args):
                 self._worker = {'fn': fn, 'args': args}
                 return self

            def message(self, message):
                self._message = message
                return self

            def show_time(self, show_time):
                self._show_time = show_time
                return self

            def start(self):
                from humanfriendly.terminal.spinners import AutomaticSpinner

                message = self._message
                worker_fn = self._worker.get('fn')
                arguments = self._worker.get('args')
                show_time = self._show_time

                assert worker_fn

                try:
                    if not is_windows():
                        with AutomaticSpinner(message, show_time=show_time):
                            return worker_fn(**arguments)
                    else:
                        # TODO: Make the same experience for windows ps1/dos
                        OutputStream().stdout.write(message)
                        result = worker_fn(**arguments)
                        return result
                finally:
                    self._defaults()  # reset

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        from azdata.cli.core.text import Text

        self._terminal = type('', (object,), {
            'progress_indicator': Progress(),
            'text': Text()
        })

        return self._terminal

    @property
    def utils(self):
        """
        Object mapping to supported public `utils` operations.

        Supported:
        - `download`
        - `import_api`

        Example:
        ```
        client.utils.download(...)
        client.utils.import_api(...)
        ```

        :return: The Object mapping to supported `utils` operations.
        """
        if self._utils:
            return self._utils

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------

        def download(url, filename, destination=None, label='Downloading',
                     show_progress=True):
            """
            Helper to download a file given the url and a write destination.
            If no
            destination is given the file download is sent to a temporary
            location.

            :param url: The URL to the file to be downloaded.
            :param filename: Name the downloaded file.
            :param destination: Location where to save file.
            :param label: Work with `show_progress` to define the label for the
                   optional spinner (a string or None, defaults to Downloading).
            :param show_progress: To display a progress spinner on the terminal.
            :return: The path to the downloaded file.
            """
            import urllib
            import time
            import tempfile
            import shutil
            import os

            def _download(uri, name, dest):
                stage_dir = tempfile.mkdtemp()
                try:
                    file_path = os.path.join(stage_dir, name)
                    num_blocks = 0
                    chunk_size = 4096
                    req = urllib.request.urlopen(uri)

                    with open(file_path, 'wb') as f:
                        while True:
                            data = req.read(chunk_size)
                            time.sleep(.5)
                            num_blocks += 1

                            if not data:
                                break

                            f.write(data)

                    return shutil.copyfile(file_path,
                                           dest) if dest else file_path
                except IsADirectoryError:
                    from urllib.error import URLError
                    raise URLError('Not able to download file', filename=uri)
                finally:
                    if dest:
                        shutil.rmtree(stage_dir, ignore_errors=True)

            if destination and not os.path.isdir(destination):
                raise ValueError('Destination directory does not exist {0}'.
                                 format(destination))

            # -- download and show progress indicator --
            if not show_progress:
                return _download(url, filename, destination)
            else:
                return self.terminal.progress_indicator.worker(_download, {
                    'uri': url,
                    'name': filename,
                    'dest': destination
                }).message(label).start()

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------

        def import_api(api_module):
            """
            Gets the instance of the API class from the given `api_module`.
            This is
            a common pattern for instantiated APIs given a API module name.

            :param api_module: The name of API module.
            :return: The instance of the API.
            """
            from importlib import import_module
            from inspect import (getmembers, isclass)

            mod = import_module('azdata.cli.core.api.{}'.format(api_module))
            name, klass = getmembers(mod, isclass)[1]
            return klass(api_client=self.apis.controller.api_client)

        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------
        # ----------------------------------------------------------------------

        self._utils = type('', (object,), {
            'download': download,
            'import_api': import_api
        })

        return self._utils
