# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------
from azdata.cli.core.profile import CliProfile
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.exceptions import (
    ControllerError,
    ApiError,
    OPERATION_TIMED_OUT,
    CONNECTION_REFUSED,
    http_status_codes
)
from azdata.cli.core.api.app_router_api import AppRouterApi
from azdata.cli.core.api.cluster_router_api import ClusterRouterApi
from azdata.cli.core.api.token_router_api import TokenRouterApi
from azdata.cli.core.api.usage_router_api import UsageRouterApi
from azdata.cli.core.api.registration_router_api import RegistrationRouterApi
from azdata.cli.core.clients.api_client import ApiClient
from azdata.cli.core.logging import get_logger
from urllib3.exceptions import (NewConnectionError, MaxRetryError)
from abc import ABCMeta
from six import add_metaclass

import uuid
import json
import types
import inspect

logger = get_logger(__name__)

__all__ = ['ControllerClient']


class ControllerClient(object):
    """
    The client for managing HTTP API calls to the controller.
    """

    def __new__(cls, *args, **kwargs):
        """
        Resource object composition. We do this to open up the opportunity to
        make dependencies explicit, enforce a stronger level of encapsulation,
        and can be built upon for richer forms of method decoration.
        :return: This with all subclassed `BaseResource` functions composed in.
        """
        bases = []
        resources = [
            TokenResource,
            HomeResource,
            # TODO: Move each of these below to their respective control and
            # load as as an optional service plugin.
            ControlResource,
            AppResource,
            BdcResource,
            LogsResource,
            MetricsResource,
            StorageResource,
            RegistrationResource,
            ArcResource
        ]
        bases = []

        # -- identify public service functions to be mixed in as resources --
        if kwargs.get('services'):
            mod = kwargs.get('services')
            members = {}
            for public in getattr(mod, '__all__'):
                obj = getattr(mod, public)
                if isinstance(obj, types.FunctionType):  # include functions
                    members[public] = obj
                else:  # include classes
                    resources.append(obj)

            # dynamic class to warp arbitrary functions in, then are then added
            # to the ControlerClient instance
            if members:
                resources.append(
                    type('Service', tuple([BaseResource]), members))

        # -- compose all resource services to `this` --
        for Resource in resources:
            resource = Resource()
            for mixin in (getattr(resource, name) for name in dir(resource)):
                try:
                    if mixin and inspect.isclass(mixin):
                        bases.append(mixin)
                    else:
                        pass
                except Exception:
                    pass

        name = '{}Composed'.format(cls.__name__)
        bases.append(cls)
        return object.__new__(type(name, tuple(bases), dict(cls.__dict__)))

    def __init__(self, endpoint, access_token=None, services=None):
        logger.debug('Controller endpoint %s', endpoint)
        logger.debug('Attached optional services %s', services)
        logger.debug(dir(self))

        config = Configuration()
        self._api_client = ApiClient(
            configuration=config,
            header_name='version',
            header_value=config.server_version,
            model_namespace='{}.models'.format(services.__package__)
            if services else None
        )

        self.endpoint = endpoint

        # If we have an authenticated token set it in the header and use it for
        # all future protected resource HTTP invocations in this session
        if access_token:
            self.set_authorization_header('Bearer', access_token)

    @property
    def endpoint(self):
        """
        Gets the endpoint.
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        """
        Sets the endpoint.
        :param endpoint: The endpoint.
        """
        self._endpoint = Configuration().host = endpoint

        # keep singleton configuration properties in-sync
        self.api_client.configuration = Configuration()

    @property
    def api_client(self):
        """
        Gets the internal swagger client reference.
        """
        return self._api_client

    def set_authorization_header(self, prefix, value):
        """
        Apply across all HTTP requests
        :param prefix: The authorization api-key `Basic, Bearer, Negotiate`
        :param value: api-key value `base64, access-token, kerberos-ticket`
        """

        # keep singleton configuration properties in-sync
        self.api_client.configuration = Configuration()

        self.api_client.set_default_header(
            'Authorization',
            '{prefix} {value}'.format(prefix=prefix, value=value)
        )

# ---------------------------------------------------------------------------- #
# HTTP 401/403 Retry decorator
# ---------------------------------------------------------------------------- #


class RefreshRetry(object):
    """
    If credentials have a refresh_token, in cases of HTTP 401 and 403 responses,
    it automatically asks for a new access token and replay the unsuccessful
    request.

    @NOTE: We can do better than this implementation with some request plugin
    middleware of similar in the the future. Keeping it simple for now.
    """

    MAX_RETRY = 1
    """ The maximum number of attempts to retry the given function."""

    def __init__(self, fn):
        self.fn = fn
        self.retry_attempts = 0

    def __call__(self, *args, **kwargs):
        try:
            ret = self.fn(*args, **kwargs)
            return ret
        except ApiError as e:
            refresh_token = CliProfile().active_context.refresh_token

            logger.debug('Checking error for refresh/retry candidacy.')
            logger.debug('Retry attempt count: %s', self.retry_attempts)
            logger.debug('Retry attempts exceeded: %s',
                         self.retry_attempts >= RefreshRetry.MAX_RETRY)
            logger.debug('Refresh-Token: %s', refresh_token is not None)
            logger.debug('HTTP status: %s', e.status)

            self.retry_attempts = self.retry_attempts + 1
            if self.retry_attempts <= RefreshRetry.MAX_RETRY \
                    and refresh_token \
                    and e.status == http_status_codes.unauthorized \
                    or e.status == http_status_codes.forbidden:

                logger.debug('Attempting to acquire a new access-token with '
                             'the given refresh-token.')
                logger.debug('Incremented retry for route=%s', self.fn.__name__)
                #ret = self.fn(*args, **kwargs)
                #return ret
                raise e
            else:
                raise e

    def __get__(self, instance, owner):
        from functools import partial
        return partial(self.__call__, instance)

# ---------------------------------------------------------------------------- #
# BaseResource
# ---------------------------------------------------------------------------- #


@add_metaclass(ABCMeta)
class BaseResource(object):
    """Base HTTP Controller Resource"""
    pass

# ---------------------------------------------------------------------------- #
# Home: /
# ---------------------------------------------------------------------------- #


class HomeResource(BaseResource):
    """Home: /"""

    def is_ready(self, loggingEnabled=True):
        """
        Check if controller is ready
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            request_id = str(uuid.uuid4())

            if loggingEnabled:
                logger.debug('Sending Get request.')

            api.get_home(
                x_request_id=request_id,
                connection='close',
                _request_timeout=30
            )
            return True
        except ApiError as e:
            if e.status not in [http_status_codes.bad_gateway,
                                http_status_codes.service_unavailable,
                                OPERATION_TIMED_OUT, CONNECTION_REFUSED]:
                if loggingEnabled:
                    logger.info('Failed to retrieve home with status code '
                                '[%d]' % e.status)
                try:
                    logger.error('Reason: %s' % (json.loads(e.body)['reason']))
                except ValueError:
                    pass
                msg = 'Failed to retrieve home with status ' \
                      'code [%d]' % e.status
                logger.debug(e)
                raise ControllerError(Exception(msg), status=e.status)
            return False
        except (NewConnectionError, MaxRetryError, Exception) as e:
            return False


# ---------------------------------------------------------------------------- #
# Token: /token
# ---------------------------------------------------------------------------- #


class TokenResource(BaseResource):
    """Token: /token"""

    def acquire_token(self):
        api = TokenRouterApi(api_client=self.api_client)
        response = api.token_post()

        # -- save for future protected HTTP requests --
        self.set_authorization_header('Bearer',  response.access_token)

        return response


# ---------------------------------------------------------------------------- #
# Arc: /arc
# ---------------------------------------------------------------------------- #
class ArcResource(BaseResource):
    """
    ArcResource /arc
    """
    @RefreshRetry
    def resource_types_get(self):
        """
        Lists the Arc custom resource types available.
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.resource_types_get()
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    @RefreshRetry
    def resource_spec_get(self, resource_type):
        """
        Lists the Arc custom resource spec.
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.resource_spec_get(resource_type)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    @RefreshRetry
    def resource_crd_get(self, resource_type):
        """
        Lists the Arc custom resource CRD.
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.resource_crd_get(resource_type)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    @RefreshRetry
    def signing_certificate_public_key_get(self):
        """
        Gets the public key of the usage log signing certificate.
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.signing_certificate_public_key_get(version="1")
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    @RefreshRetry
    def usages_post(self, body):
        """
        Gets usages for billing.
        :return:
        """
        try:
            api = UsageRouterApi(api_client=self.api_client)
            return api.usages_post(version="1", body=body)
        except ApiError as e:
            raise ControllerError(e, status=e.status)

    @RefreshRetry
    def list_deleted_resources(self, start_date):
        """
        Gets resources deleted since the start date.
        :param datetime start_date: The start date of time range within which the resources were deleted.
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.deleted_resources_get(start_date=start_date, version="1")
        except ApiError as e:
            raise ControllerError(e, status=e.status)

# ---------------------------------------------------------------------------- #
# Control: /control
# ---------------------------------------------------------------------------- #


class ControlResource(BaseResource):
    def upgrade_control_plane(self, target_version):
        """
        Upgrade control plane
        """

        api = ClusterRouterApi(api_client=self.api_client)

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            api.upgrade_control_plane(
                x_request_id=request_id,
                connection='close',
                data=target_version
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def get_control_status(self, resource_name=None, control_plane_state=False, loggingEnabled = True, **kwargs):
        """
        Get control plane status
        """
        api = ClusterRouterApi(api_client=self.api_client)

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            time = 30
            conn = 'close'
            if resource_name:
                return api.get_control_resource_status(
                    **kwargs,
                    resource_name=resource_name,
                    x_request_id=request_id,
                    connection=conn,
                    _request_timeout=time
                )

            else:
                result = api.get_control_status(
                    **kwargs,
                    x_request_id=request_id,
                    connection=conn,
                    _request_timeout=time
                )

                if not control_plane_state:
                    return result
                else:
                    return result['state']
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

# ---------------------------------------------------------------------------- #
# Apps: /app/*
# ---------------------------------------------------------------------------- #


class AppResource(BaseResource):
    """Apps: /app/*"""

    @RefreshRetry
    def create_app(self, spec, archive):
        try:
            api = AppRouterApi(api_client=self.api_client)
            return api.create_app(spec=spec, package=archive)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def update_app(self, spec, archive):
        try:
            api = AppRouterApi(api_client=self.api_client)
            return api.update_app(spec=spec, package=archive)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def delete_app(self, name, version=None):
        try:
            api = AppRouterApi(api_client=self.api_client)
            return api.app_by_name_by_version_delete(name, version)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def list_apps(self, name=None):
        try:
            api = AppRouterApi(api_client=self.api_client)
            if not name:
                res = api.app_get()
            else:
                res = api.app_by_name_get(name=name)
            apps = []
            for app in res:
                apps.append(app.to_dict())
            return apps
        except ApiError as e:
            if e.status == http_status_codes.not_found:
                return []
            else:
                raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def list_app(self, name, version=None):
        try:
            api = AppRouterApi(api_client=self.api_client)
            return [api.app_by_name_by_version_get(name, version).to_dict()]
        except ApiError as e:
            if e.status == http_status_codes.not_found:
                return []
            else:
                raise ControllerError(e, status=e.status, response_json=e.response_json)

# ---------------------------------------------------------------------------- #
# BDC: /bdc/*
# ---------------------------------------------------------------------------- #


class BdcResource(BaseResource):
    """BDC: /bdc/* """

    def create_cluster(self, cluster_config):
        """
        Create cluster
        """
        api = ClusterRouterApi(api_client=self.api_client)

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            return api.create_cluster(
                x_request_id=request_id,
                connection='close',
                data=json.loads(cluster_config)
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def delete_cluster(self):
        """
        Delete cluster
        """
        api = ClusterRouterApi(api_client=self.api_client)

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            api.delete_cluster(
                x_request_id=request_id,
                connection='close'
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def upgrade_bdc(self, target_version):
        """
        Upgrade Big Data Cluster
        """
        api = ClusterRouterApi(api_client=self.api_client)

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            api.upgrade_bdc(
                x_request_id=request_id,
                connection='close',
                data=target_version
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def get_controller_logs(self, offset):
        """
        Streams the progress of cluster creation
        """
        api = ClusterRouterApi(api_client=self.api_client)

        try:
            return api.get_logs(
                x_request_id=str(uuid.uuid4()),
                connection='close',
                offset=offset,
                _request_timeout=60
            )
        except Exception as e:
            logger.debug(e)
            return None

    @RefreshRetry
    def get_cluster(self):
        """
        Get cluster config
        """

        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            response = api.get_cluster(
                x_request_id=request_id,
                connection='close',
                _request_timeout=30
            )
            return response['spec']
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def get_endpoints(self, endpoint_name=None):
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            if not endpoint_name:
                return api.endpoints_get()
            else:
                return api.endpoints_by_name_get(endpoint_name)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def get_cluster_state(self, loggingEnabled=True):
        """
        Get cluster state
        """
        # Generate a request id
        request_id = str(uuid.uuid4())
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            result = api.get_bdc_status(
                request_id,
                'close',
                _request_timeout=30
            )

            return result['state']
        except ApiError as e:
            if loggingEnabled:
                if e.status == http_status_codes.not_found:
                    logger.info('Cannot find any cluster')
                elif e.status == http_status_codes.unauthorized:
                    logger.info('Failed to get the cluster information. '
                                'Invalid username or password')
                elif e.status == 0:
                    logger.info(e.reason)
                elif e.status == OPERATION_TIMED_OUT:
                    logger.debug('Failed to get cluster status with code [%d]'
                                 '' % e.status)
                else:
                    logger.info('Failed to get cluster status '
                                'with code [%d]' % e.status)
                    try:
                        logger.info('Reason: %s' %
                                    (json.loads(e.body)['reason']))
                    except (ValueError, TypeError):
                        pass
            logger.debug(e)
            return None
        except Exception as e:
            logger.debug(e)
            return None

    @RefreshRetry
    def get_bdc_status(self, service_name=None, resource_name=None,
                       loggingEnabled=True, **kwargs):
        """
        Get BDC status
        """
        try:
            request_id = str(uuid.uuid4())
            conn = 'close'
            time = 30
            api = ClusterRouterApi(api_client=self.api_client)
            if service_name:
                if resource_name:
                    return api.get_bdc_service_resource_status(
                        request_id,
                        conn,
                        service_name,
                        resource_name,
                        **kwargs,
                        _request_timeout=time
                    )
                else:
                    return api.get_bdc_service_status(
                        request_id,
                        conn,
                        service_name,
                        **kwargs,
                        _request_timeout=time
                    )
            elif resource_name:
                return api.get_bdc_resource_status(
                    request_id,
                    conn,
                    resource_name,
                    **kwargs,
                    _request_timeout=time
                )
            else:
                return api.get_bdc_status(
                    request_id,
                    conn,
                    **kwargs,
                    _request_timeout=time
                )
        except ApiError as e:
            if e.status == 503:
                raise ControllerError(
                    'Big data cluster is waiting for deployment to start. '
                    'Please check control '
                    'service status with `azdata bdc control status show`.'
                )
            raise ControllerError(e, status=e.status, response_json=e.response_json)


# ---------------------------------------------------------------------------- #
# Logs: /logs
# ---------------------------------------------------------------------------- #

class LogsResource(BaseResource):
    """
    LogsResource /logs
    """
    @RefreshRetry
    def logs_post(self, body):
        """
        Query and retrieve logs
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.logs_post(version="1", body=body)
        except ApiError as e:
            raise ControllerError(e, status=e.status)


# ---------------------------------------------------------------------------- #
# Metrics: /metrics
# ---------------------------------------------------------------------------- #

class MetricsResource(BaseResource):
    """
    MetricsResource /metrics
    """
    @RefreshRetry
    def metrics_post(self, body):
        """
        Query and retrieve metrics
        :return:
        """
        try:
            api = ClusterRouterApi(api_client=self.api_client)
            return api.metrics_post(version="1", body=body)
        except ApiError as e:
            raise ControllerError(e, status=e.status)


# ---------------------------------------------------------------------------- #
# Storage: /storage/*
# ---------------------------------------------------------------------------- #


class StorageResource(BaseResource):
    """Storage: /storage/*"""

    @RefreshRetry
    def create_mount(self, remote_uri, local_path, credentials):
        """
        Create a mount.
        :param remote_uri: URI of the store to mount.
        :param local_path: Local HDFS path which will serve as the mount point.
        :param credentials: the credentials to create the mount as a JSON.
        """

        api = ClusterRouterApi(api_client=self.api_client)

        # generate a request id
        request_id = str(uuid.uuid4())

        try:
            # Create a mount
            logger.debug(
                'Creating mount %s for remote uri %s.' %
                (local_path, remote_uri)
            )

            return api.create_mount(
                request_id,
                'close',
                remote_uri,
                local_path,
                credentials=credentials
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def delete_mount(self, local_path):
        """
        Delete a mount.
        :param local_path: mount point to delete
        """

        api = ClusterRouterApi(api_client=self.api_client)

        request_id = str(uuid.uuid4())
        try:
            # Delete the mount
            logger.debug('Deleting mount %s.' % local_path)
            return api.delete_mount(request_id, 'close', local_path)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def refresh_mount(self, local_path):
        """
        Refresh a mount.
        :param local_path: mount point to refresh
        """
        api = ClusterRouterApi(api_client=self.api_client)
        request_id = str(uuid.uuid4())
        try:
            # Delete the mount
            logger.debug('Refreshing mount %s.' % local_path)
            return api.refresh_mount(request_id, 'close', local_path)
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    @RefreshRetry
    def get_mount_status(self, local_path):
        """
        Get the status of mount(s).
        :param local_path: mount path
        :return: A Json for the status of a specified mount. If local_path is
        None, returns the status of all existing mounts.
        """
        api = ClusterRouterApi(api_client=self.api_client)
        request_id = str(uuid.uuid4())
        try:
            # TODO: server API should return proper content-type and not return
            # string's of JSON plain/text then force the client to marshal.
            # A proper model should be put together and use swagger for com.
            if local_path is None:
                logger.debug('Getting status of all mounts.')

                mounts = api.get_mounts(request_id, 'close') or []
                return [json.loads(mount) for mount in mounts]
            else:
                logger.debug('Getting status of mount %s.' % local_path)
                mount = api.get_mounts(request_id, 'close', mount=local_path)
                return json.loads(mount if mount else '{}')
        except (ValueError, TypeError):
            # TODO: see above. We also should not toggle response payloads
            # between JSON-strings and non-JSON strings (e.g sentences.)
            return []
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

# ---------------------------------------------------------------------------- #
# Registration router: /registration
#
# TODO: Remove this in place and replace with `arc` subgroup
# ---------------------------------------------------------------------------- #


class RegistrationResource (BaseResource):
    INSTANCE_NAME = 'instanceName'
    INSTANCE_TYPE = 'instanceType'
    SUBSCRIPTION_ID = 'subscriptionId'
    RESOURCE_GROUP_NAME = 'resourceGroupName'
    LOCATION = 'location'
    CUSTOM_OBJECT_NAME = 'customObjectName'
    DEFAULT_DATA_CONTROLLER_NAME = 'DataController'
    CLUSTER_CONNECTION_MODE_DISCONNECTED = 'disconnected'
    BILLING_MODEL_COMSUMPTION = 'consumption'
    INSTANCE_TYPE_DATA_CONTROLLER = 'dataControllers'
    GUID_REGEX = r'[0-9a-f]{8}\-([0-9a-f]{4}\-){3}[0-9a-f]{12}'

    def get_instance(self, name, instance_type, namespace, context):
        """
        Get instance registration record
        :param name: instance name
        :param instance_type: Azure instance type
        :param namespace: k8s namespace the instance belongs to
        :param context: active context in profile
        :return:
        """
        resource_name = name if namespace == context.namespace else \
            '{}_{}'.format(namespace, name)

        # Get list of custom objects and lookup given value
        instances = self.list_registration_instances(context.namespace)
        instances_by_name = list(
            filter(
                lambda x: x[self.INSTANCE_NAME] == resource_name
                      and x[self.INSTANCE_TYPE] == instance_type,
                instances))

        if instances_by_name:
            return {
                'resource_group_name':
                    instances_by_name[0][self.RESOURCE_GROUP_NAME],
                'subscription_id': instances_by_name[0][self.SUBSCRIPTION_ID],
                'location': instances_by_name[0][self.LOCATION],
                'reg_object_name': instances_by_name[0][self.CUSTOM_OBJECT_NAME]
            }
        else:
            return None

    def register_instance_by_configuration(self, config, namespace):
        from azdata.cli.core.models.registration_spec import RegistrationSpec
        import datetime

        # NOTE: This is the deciding factor that determins if we are pointing
        # to the ignite branch's controller or master.
        # TODO: This will be removed when this code space is moved to `arc`
        if 'dataController' not in config['spec']['controlPlane']['spec']:
            return None

        data_controller = \
            config['spec']['controlPlane']['spec']['dataController']

        display_name = data_controller['displayName']
        if display_name is None:
            display_name = self.DEFAULT_DATA_CONTROLLER_NAME

        subscription = data_controller['subscription']
        resource_group_name = data_controller['resourceGroup']
        location = data_controller['location']

        connection_mode = data_controller['connectionMode']

        if connection_mode is None:
            connection_mode = self.CLUSTER_CONNECTION_MODE_DISCONNECTED

        billing_mode = data_controller['billingMode']
        if billing_mode is None:
            billing_mode = self.BILLING_MODEL_COMSUMPTION

        if subscription and resource_group_name and location:

            create_timestamp = datetime.datetime.utcnow().strftime(
                "%m/%d/%Y %H:%M:%S")

            # TODO: should not be making print/stdout at this tier
            print('Creating data controller: "{}" CRD'.format(display_name))

            try:
                reg_spec = RegistrationSpec(
                    uid=str(uuid.uuid4()),
                    instance_name=display_name,
                    instance_type=self.INSTANCE_TYPE_DATA_CONTROLLER,
                    instance_namespace=namespace,
                    service_name='DataController',
                    vcores='-',
                    resource_group_name=resource_group_name,
                    subscription_id=subscription,
                    location=location,
                    create_timestamp=create_timestamp,
                    update_timestamp=create_timestamp,
                    properties=None,
                    connection_mode=connection_mode,
                    billing_mode=billing_mode
                )

                # Create a record in CRD with instance details
                api = RegistrationRouterApi(api_client=self.api_client)
                return api.create(
                    namespace=namespace,
                    name=reg_spec.instance_name,
                    spec=reg_spec
                )
            except ApiError as e:
                raise ControllerError(e, status=e.status, response_json=e.response_json)
        elif all(v is None for v in [subscription,
                                     resource_group_name,
                                     location]):
            return None
        else:
            raise ValueError('Azure subscription, resource group and location '
                             'must be not None all or None.')

    def register_instance(
            self,
            uid,
            name,
            instance_type,
            namespace,
            context,
            subscription_id,
            resource_group_name,
            location,
            service_name,
            endpoint,
            vcores='-',
            create_timestamp=None,
            update_timestamp=None,
            connection_mode=None,
            properties=None,
            billing_mode=None
    ):
        from azdata.cli.core.models.registration_spec import RegistrationSpec

        try:
            resource_name = name if namespace == context.namespace else \
                '{}_{}'.format(namespace, name)

            reg_spec = RegistrationSpec(
                uid=uid,
                instance_name=resource_name,
                instance_type=instance_type,
                instance_namespace=namespace,
                service_name=service_name,
                vcores=vcores,
                resource_group_name=resource_group_name,
                subscription_id=subscription_id,
                location=location,
                create_timestamp=create_timestamp,
                update_timestamp=update_timestamp,
                properties=properties,
                connection_mode=connection_mode,
                billing_mode=billing_mode
            )

            # Create a record in CRD with instance details
            api = RegistrationRouterApi(api_client=self.api_client)
            return api.create(
                namespace=context.namespace,
                name=resource_name,
                spec=reg_spec
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    def delete_instance(self, name, instance_type, namespace, context):
        """
        Delete the instance's shadow resource from Azure, and delete/update
        registration record
        :param name: instance name
        :param instance_type: Azure instance type
        :param namespace: k8s namespace the instance belongs to
        :param context: active context in profile
        """
        try:
            # -- Get list of custom objects and lookup given value --
            resource_name = name if namespace == context.namespace else \
                '{}_{}'.format(namespace, name)

            # Get list of custom objects and lookup given value
            instances = self.list_registration_instances(context.namespace)

            instances_by_name = list(
                filter(
                    lambda x: x[self.INSTANCE_NAME] == resource_name and
                    x[self.INSTANCE_TYPE] == instance_type, instances
                )
            )

            if instances_by_name:
                registration_info = {
                    'resource_group_name': instances_by_name[0][
                        self.RESOURCE_GROUP_NAME],
                    'subscription_id': instances_by_name[0][
                        self.SUBSCRIPTION_ID],
                    'location': instances_by_name[0][self.LOCATION],
                    'reg_object_name': instances_by_name[0][
                        self.CUSTOM_OBJECT_NAME]
                }

                # Don't delete CRD but mark it as IsDeleted
                reg_object_name = registration_info['reg_object_name']
                soft_delete = True
                api = RegistrationRouterApi(api_client=self.api_client)
                return api.registration_ns_name_delete(
                    context.namespace,
                    reg_object_name,
                    soft_delete
                )
            else:
                return None
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)

    def list_registration_instances(self, namespace):
        try:
            api = RegistrationRouterApi(api_client=self.api_client)
            return api.registration_list_resources_ns_get(namespace)
        except ApiError as e:
            if e.status == http_status_codes.not_found:
                return []
            else:
                raise ControllerError(e, status=e.status, response_json=e.response_json)

    def delete_registration_instance(self, name, namespace, soft_delete):
        try:
            api = RegistrationRouterApi(api_client=self.api_client)
            return api.registration_ns_name_delete(
                namespace,
                name,
                soft_delete
            )
        except ApiError as e:
            raise ControllerError(e, status=e.status, response_json=e.response_json)
