# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.logging import get_logger

import json
import os
import shlex
import subprocess
import sys

log = get_logger(__name__)


def beget(_):
    return Command


class Command(object):
    """
    A combined byte and text input and output coordinator that allows easy IO of
    bytes and text from a subprocess to execute a CLI command.
    """
    def run(commands, **kwargs):
        """
        A subprocess IO run factory based on commands.

        ```py
        from commandline_process import Command

        cmd = Command
        cmd.run('ls -la').stdout

        # -- or --
        Command.run('ls -la').stdout

        # -- or --
        cmd.run('ls -la', 'pwd', 'whoami').stdout

        # -- or pipe --
        cmd.run('grep -rn "joy division" *', data=cmd.run.stdin)
        ```

        :param commands: The commands to run in the subprocess.
        :return: The command to run.
        """
        return CommandLineProcess(commands, **kwargs)

    run = staticmethod(run)


class StdOutput(str):
    """
    StdOutput represents the IOSink for either `stdout` or `stderr`.
    It provides a blocking IOSink, so using this to write will block until the
    output is written.
    """

    @property
    def return_code(self):
        """
        Gets the return code of the process that produced this stdout.
        """
        return self._return_code

    @return_code.setter
    def return_code(self, return_code):
        """
        Sets the return code of the process that produced this stdout.
        """
        self._return_code = return_code

    @property
    def from_json(self):
        try:
            out = json.loads(self.split('\n')[0])
        except ValueError as e:
            msg = 'Invalid `json` {0}\n{1}'.format(type(e).__name__, str(e))
            out = {'stderr': self, 'message': msg}

        return out

    @property
    def lines(self):
        return self.split('\n')


class StdInput(type):
    """
    StdInput allows for synchronous reads from the standard input stream of the
    interpreter.
    """

    @property
    def stdin(cls):
        return sys.stdin


class CommandLineProcess(StdInput('base_CommandLineProcess', (StdOutput,), {})):
    STDOUT = 'stdout'
    """Defines the read state of the stdout stream."""

    STDERR = 'stderr'
    """Defines the read state of the stderr stream."""

    EOF = 'eof'
    """Defines the process is finished and done reading from stdout/stderr."""

    EXIT_SUCCESS = 0
    """ Signal general process success."""

    EXIT_FAILURE = 1
    """ Signal general process failure."""

    @classmethod
    def create_process(cls, command, stdin, cwd, env, shell):
        """
        The subprocess factory.
        :param command:
        :param stdin:
        :param cwd:
        :param env:
        :param shell:
        :return: The subprocess.
        """
        return subprocess.Popen(
            shlex.split(command),
            universal_newlines=True,
            shell=shell,
            cwd=cwd,
            env=env,
            stdin=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0
        )

    def __new__(cls, *args, **kwargs):
        """
        Defines the commandline process constructor.

        :param args: The commands to run.
        :param kwargs: A config `dict` of the following arguments and values:
               - env
               - cwd
               - shell
               - stdin
        """

        env = dict(os.environ)
        env.update(kwargs.get('env', {}))

        cwd = kwargs.get('cwd')
        shell = kwargs.get('shell', False)

        chain = []

        stdin = kwargs.get('stdin', subprocess.PIPE)

        for command in args:
            process = cls.create_process(
                command, stdin, cwd=cwd, env=env, shell=shell)

            stdin = process.stdout

            obj = super(CommandLineProcess, cls).__new__(CommandLineProcess,
                                                         command)

            obj.process = process
            obj.pid = process.pid
            obj.command = command

            chain.append(obj)
            obj.chain = chain[:]

        # -- some constructor defaults --
        obj._res = None

        return obj

    def kill(self):
        self.process.kill()

    def terminate(self):
        self.process.terminate()

    def communicate(self):
        self._res = self._io()

        log.debug('===========================================================')
        log.debug('Command: %s', self.command)
        log.debug('stdout: %s', self._res[0])
        log.debug('stderr: %s', self._res[1])
        log.debug('returncode: %s', self.process.returncode)
        log.debug('pid: %s', self.pid)
        log.debug('===========================================================')

        return self._res

    @property
    def status(self):
        self.communicate()
        return self.process.returncode

    @property
    def stdout(self):
        res = self.communicate()

        # Check here if we have both stderr and stdout present and the process
        # return-code code was a success. If this happens just suppress stderr
        # and use stdout. This can occur if the subprocess itself invokes
        # another subprocess and that subprocess does not deal with stderr
        # properly
        if res[1] and res[0] and self.process.returncode == \
                CommandLineProcess.EXIT_SUCCESS:
            out = StdOutput(res[0])
        else:
            out = StdOutput(res[1] if res[1] else res[0])  # stdout/stderr

        out.return_code = self.process.returncode

        return out

    @property
    def stderr(self):
        res = self.communicate()
        out = StdOutput(res[1])  # stdout/stderr
        out.return_code = self.process.returncode

        return out

    def _io(self):
        """
        Interact with the process. Send data to stdin, read data from stdout and
        stderr, until end-of-file is reached.

        By default this call will block and wait for process to terminate.
        If `polling=True` it will poll stdout/stderr pips for readlines to
        indicate progress.

        :return: A tuple (stdoutdata, stderrdata).
        """
        # -- do not re-run --
        if self._res:
            return self._res

        return self.process.communicate()

    def __repr__(self):
        return ' | '.join([e.command for e in self.chain])
