# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.events import EVENT_INVOKER_PRE_CMD_TBL_TRUNCATE
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.output import OutputProducer
from knack.log import CLI_LOGGER_NAME as BASE_CLI_LOGGER_NAME
from io import BytesIO
from six.moves import http_client as httplib

import knack.log
import sys
import os
import traceback
import logging
import datetime

__all__ = ['CLI_LOGGER_NAME', 'get_logger', 'get_logs', 'CliLogging']

CLI_LOGGER_NAME = Configuration().CLI_NAME
"""
The name of the parent logger is defined in CLI_LOGGER_NAME variable. Everything
descends from.
"""


def get_logger(mod_name):
    """
    :param mod_name:
    :return:
    """
    from knack.log import get_logger
    return get_logger(mod_name)


def get_logs():
    handlers = logging.getLogger(BASE_CLI_LOGGER_NAME).handlers
    logs = []
    if handlers and len(handlers) > 0:
        logs = handlers[1].flush() if len(handlers) > 2 else handlers[0].flush()

    return logs if logs else []


class CliLogging(knack.log.CLILogging):
    COMMAND_METADATA_LOGGER = '{}_command_data_logger'.format(CLI_LOGGER_NAME)
    CMD_LOG_LINE_PREFIX = 'CMD-LOG-LINE-BEGIN'
    UNKNOWN_COMMAND = 'unknown_command'

    def __init__(self, name, cli_ctx=None):
        super(CliLogging, self).__init__(name, cli_ctx)
        self.command_log_dir = os.path.join(cli_ctx.config.config_dir,
                                            'commands')
        self.command_logger_handler = None
        self.command_metadata_logger = None
        self.cli_ctx.register_event(
            EVENT_INVOKER_PRE_CMD_TBL_TRUNCATE,
            CliLogging.init_command_file_logging
        )

    def get_command_log_dir(self):
        return self.command_log_dir

    def end_cmd_metadata_logging(self, exit_code):
        if self.command_metadata_logger:
            self.command_metadata_logger.info('exit code: %s', exit_code)

            # We have finished metadata logging, remove handler and set
            # command_metadata_handler to None.
            # crucial to remove handler as in python logger objects are
            # shared which can affect testing of this logger
            # we do not want duplicate handlers to be added in subsequent
            # calls of _init_command_logfile_handlers
            self.command_metadata_logger.removeHandler(
                self.command_logger_handler)
            self.command_metadata_logger = None

    def add_error_handler(self):
        if OutputProducer.structured_output:
            cli_logger = logging.getLogger(BASE_CLI_LOGGER_NAME)
            if not hasattr(self, 'error_handler'):
                self.error_handler = StreamHandler(
                    30,
                    self.console_log_format[BASE_CLI_LOGGER_NAME]
                )
            cli_logger.addHandler(self.error_handler)

    @staticmethod
    def traceback(msg):
        """
        Log message and the traceback to log file only.
        :param msg: Error message to log.
        """
        logger = logging.getLogger(BASE_CLI_LOGGER_NAME)

        if not len(logger.handlers):
            return

        e_t, e_v, e_tb = sys.exc_info()
        trace_msg = '{}\n{}'. \
            format(msg, '\n'.join(traceback.format_exception(e_t, e_v, e_tb)))

        config = Configuration()

        # -- display traceback  --
        if not OutputProducer.structured_output and (
                    config.debug or config.verbose):
            logger.error(trace_msg)
        else:  # -- remove tracback from console-logger --
            console_handler = logger.handlers[0]
            logger.handlers = logger.handlers[1:]
            logger.error(trace_msg)  # log traceback to file
            logger.addHandler(console_handler)

    @staticmethod
    def init_command_file_logging(cli_ctx, **kwargs):
        def _delete_old_logs(log_dir):
            """
            Periodically delete the 5 oldest command log files, ensuring that
            only the history of the last 10 commands (or less) are kept.
            """
            # get log file names and sort them from newest to oldest file.
            log_file_names = [file for file in os.listdir(log_dir) if
                              file.endswith('.log')]
            sorted_files = sorted(log_file_names, reverse=True)

            # if we have too many files, delete the 5 last / oldest command
            # log files.
            if len(sorted_files) > 10:
                for file in sorted_files[-5:]:
                    try:
                        os.remove(os.path.join(log_dir, file))
                    except OSError:  # FileNotFoundError introduced in Python 3
                        continue

        # if tab-completion and not command don't log to file.
        # if not cli_ctx.data.get('completer_active', False):
        self = cli_ctx.logging
        args = kwargs['args']

        cmd_logger = logging.getLogger(CliLogging.COMMAND_METADATA_LOGGER)

        # pylint: disable=protected-access
        self._init_command_logfile_handlers(cmd_logger, args)

        get_logger(__name__).debug('metadata file logging enabled - writing '
                                   'logs to "%s".', self.command_log_dir)

        _delete_old_logs(self.command_log_dir)

    def _init_command_logfile_handlers(self, command_metadata_logger, args):
        os.makedirs(self.command_log_dir, exist_ok=True)

        # -- load commands and get command string --
        self.cli_ctx.invocation.commands_loader.load_command_table(args)
        command = self.cli_ctx.invocation._rudimentary_get_command(args) or \
                  CliLogging.UNKNOWN_COMMAND  # pylint: disable=protected-access
        command_str = command.replace(' ', '_')

        date_str = str(datetime.datetime.now().date())
        time = datetime.datetime.now().time()
        time_str = '{:02}-{:02}-{:02}'.\
            format(time.hour, time.minute, time.second)

        log_name = '{}.{}.{}.{}.{}'.format(date_str, time_str, command_str,
                                           os.getpid(), 'log')

        log_file_path = os.path.join(self.command_log_dir, log_name)
        logfile_handler = logging.FileHandler(log_file_path)

        lfmt = logging.Formatter(
            CliLogging.CMD_LOG_LINE_PREFIX +
            ' %(process)d | %(asctime)s | %(levelname)s'
            ' | %(name)s | %(message)s'
        )
        logfile_handler.setFormatter(lfmt)
        logfile_handler.setLevel(logging.DEBUG)
        command_metadata_logger.addHandler(logfile_handler)

        self.command_logger_handler = logfile_handler
        self.command_metadata_logger = command_metadata_logger

        args = self._sanitize_args(command_str, args)
        command_metadata_logger.info('command args: %s', ' '.join(args))

    def _sanitize_args(self, command, args):
        sanitized_args = []
        for arg in args:
            sanitized_args.append(arg)

        return sanitized_args

    def _init_logfile_handlers(self, root_logger, cli_logger):
        os.makedirs(self.log_dir, exist_ok=True)
        log_file_path = os.path.join(self.log_dir, self.logfile_name)
        logfile_handler = logging.handlers.RotatingFileHandler(
            log_file_path, maxBytes=10 * 1024 * 1024, backupCount=5)
        fmt = logging.Formatter(
            '%(process)d | %(asctime)s.%(msecs)04d UTC | %(levelname)s | '
            '%(name)s | %(message)s', '%Y-%m-%d %H:%M:%S')
        logfile_handler.setFormatter(fmt)
        logfile_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(logfile_handler)
        cli_logger.addHandler(logfile_handler)

    def _init_console_handlers(self, root_logger, cli_logger, log_level_config):
        config = Configuration()

        # -- set dependence logging levels to error (add more when needed) --
        if config.debug:
            httplib.HTTPConnection.debuglevel = 1
            logging.getLogger('requests').setLevel(logging.DEBUG)
            logging.getLogger('kubernetes').setLevel(logging.DEBUG)
            logging.getLogger('urllib3.connectionpool').setLevel(logging.DEBUG)
        else:
            httplib.HTTPConnection.debuglevel = 0
            logging.getLogger('requests').setLevel(logging.ERROR)
            logging.getLogger('kubernetes').setLevel(logging.ERROR)
            logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)

        if OutputProducer.structured_output:
            root_logger.addHandler(BufferedStreamHandler(
                log_level_config['root'],
                self.console_log_format['root'])
            )

            buffered_handler = BufferedStreamHandler(
                log_level_config[BASE_CLI_LOGGER_NAME],
                self.console_log_format[BASE_CLI_LOGGER_NAME]
            )
            cli_logger.addHandler(buffered_handler)

            self.error_handler = StreamHandler(
                log_level_config[BASE_CLI_LOGGER_NAME],
                self.console_log_format[BASE_CLI_LOGGER_NAME]
            )
        else:
            # -- console format for both color/non-color --
            if config.debug or config.verbose:
                fmt = '%(asctime)s.%(msecs)04d UTC | %(levelname)s | ' \
                      '%(message)s'

                self.console_log_format[BASE_CLI_LOGGER_NAME] = {
                    True: fmt,
                    False: fmt
                }

            root_logger.addHandler(
                StreamHandler(log_level_config['root'],
                              self.console_log_format['root']))
            cli_logger.addHandler(
                StreamHandler(log_level_config[BASE_CLI_LOGGER_NAME],
                              self.console_log_format[BASE_CLI_LOGGER_NAME]))


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #

class StreamHandler(logging.StreamHandler):
    """
    A custom console handler class which writes logging records, appropriately
    formatted and colored to the console stream.
    """
    COLOR_MAP = None

    def __init__(self, log_level_config, log_format):
        import platform
        import colorama

        logging.StreamHandler.__init__(self)
        self.setLevel(log_level_config)
        if platform.system() == 'Windows':
            self.stream = colorama.AnsiToWin32(self.stream).stream
        self.enable_color = self._should_enable_color()
        self.setFormatter(logging.Formatter(log_format[self.enable_color],
                                            '%Y-%m-%d %H:%M:%S'))

    def format(self, record):
        msg = logging.StreamHandler.format(self, record)
        if self.enable_color:
            try:
                msg = self.get_color_wrapper(record.levelno)(msg)
            except KeyError:
                pass
        return msg

    @classmethod
    def get_color_wrapper(cls, level):
        if not cls.COLOR_MAP:
            import colorama

            def _color_wrapper(color_marker):
                def wrap_msg_with_color(msg):
                    return '{}{}{}'.format(color_marker, msg,
                                           colorama.Style.RESET_ALL)
                return wrap_msg_with_color

            cls.COLOR_MAP = {
                logging.CRITICAL: _color_wrapper(colorama.Fore.LIGHTRED_EX),
                logging.ERROR: _color_wrapper(colorama.Fore.LIGHTRED_EX),
                logging.WARNING: _color_wrapper(colorama.Fore.YELLOW),
                logging.INFO: _color_wrapper(colorama.Fore.GREEN),
                logging.DEBUG: _color_wrapper(colorama.Fore.CYAN)
            }

        return cls.COLOR_MAP.get(level, None)

    def _should_enable_color(self):
        try:
            # Color if tty stream available
            if self.stream.isatty():
                return True
        except AttributeError:
            pass
        return False


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class BufferedStreamHandler(logging.Handler):
    """
    A handler class which writes logging records, appropriately formatted,
    to a stream. Note that this class does not close the stream, as
    sys.stdout or sys.stderr may be used.
    """

    terminator = '\n'

    def __init__(self, log_level_config, log_format):
        """
        Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        from azdata.cli.core.output import IOBuffer
        logging.Handler.__init__(self)

        self.setLevel(log_level_config)
        self.stream = IOBuffer(BytesIO(), sys.stdout.encoding)

    def flush(self):
        """
        Flushes the buffered stream.
        """
        self.acquire()
        try:
            self.stream.seek(0)
            buffer = self.stream.read().rstrip().split(self.terminator)
            self.stream.flush()
            return [] if len(buffer) == 1 and not buffer[0] else buffer
        finally:
            self.release()

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream has
        an 'encoding' attribute, it is used to determine how to do the output to
        the stream.
        """
        try:
            msg = self.format(record)
            stream = self.stream
            stream.write(msg)
        except Exception:
            self.handleError(record)
