import json
from azdata.cli.core.class_utils import (
    validatedclass,
    validator,
    enforcetype)
from abc import abstractmethod
from azdata.cli.core.models.dict_utils import SerializationUtils
from azdata.cli.core.models.volume_claim import VolumeClaim
from azdata.cli.core.models.service_spec import ServiceSpec
from azdata.cli.core.models.docker_spec import DockerSpec
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.util import (prune_dict,  trim_dict_entries)
from typing import Union, TYPE_CHECKING, Type
# KubernetesClient is only needed for typehints, but causes a circular import. This is the python provided workaround
if TYPE_CHECKING:
    from azdata.cli.core.clients.kubernetes_client import KubernetesClient

TYPE_ERROR = "Type \'{}\' is incompatible with property \'{}\'"


@validatedclass
class CustomResource(SerializationUtils):
    """
    Internal Custom Resource object to be used for deployments.
    """
    def __init__(
        self,
        api_version: str = None,
        kind: str = None,
        spec: SerializationUtils = None,
        metadata: SerializationUtils = None,
        status: SerializationUtils = None):
        """
        Initializes a CR object with the given json.
        :param cr_object: The custom resource json.
        """
        self.apiVersion = api_version
        self.kind = kind
        self.spec = spec if spec else self.Spec()
        self.metadata = metadata if metadata else self.Metadata()
        self.status = status if status else self.Status()

    @property
    def apiVersion(self) -> str:
        """
        A kubernetes required field - Which version of the kubernetes API is being used to create this object
        @see https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/#required-fields
        """
        return self._apiVersion

    @apiVersion.setter
    def apiVersion(self, v: str):
        self._apiVersion = v

    @property
    def kind(self) -> str:
        """
        A kubernetes required field - What kind of object this is
        @see https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/#required-fields
        """
        return self._kind

    @kind.setter
    def kind(self, k: str):
        self._kind = k

    @property
    def group(self):
        """
        A computed property equal to the first part of the apiVersion
        """
        return self.apiVersion.split("/")[0] if self.apiVersion else None

    @property
    def version(self):
        """
        A computed property equal to the second part of the apiVersion
        """
        return self.apiVersion.split("/")[1] if self.apiVersion else None

    class Metadata(SerializationUtils):
        """
        A kubernetes required field that contains unique information about a k8s object including name, UID, namespace etc
        @see https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/#required-fields

        CustomResource.metadata
        """
        def __init__(self, *args, **kwargs):
            if "name" in kwargs:
                self.name = kwargs["name"]

            if "namespace" in kwargs:
                self.namespace = kwargs["namespace"]

        @property
        def uid(self) -> str:
            """
            A unique identifier for a resource
            """
            return self._uid

        @uid.setter
        def uid(self, uid: str):
            self._uid = uid

        @property
        def name(self) -> str:
            """
            The name of the resource
            """
            return self._name

        @name.setter
        def name(self, n: str):
            self._name = n

        @property
        def namespace(self) -> str:
            """
            Which k8s namespace the resource resides in
            """
            return self._namespace

        @namespace.setter
        def namespace(self, ns: str):
            self._namespace = ns

        @property
        def resourceVersion(self) -> int:
            """
            What version of the resource this object was created from. Updates to a resource may be concurrent so Kubernetes
            wants to know what version of the resource you think you're updating when this is serialized and sent to the backend
            """
            return self._resourceVersion

        @resourceVersion.setter
        def resourceVersion(self, rv: Union[str, int]):
            try:
                self._resourceVersion = int(rv)
            except ValueError:
                raise ValueError("Valid resource versions are positive integers")

        def _to_dict(self):
            """
            @override
            """
            v = getattr(self, "resourceVersion", None)
            v_str = str(v) if v else None
            return {
                "name": getattr(self, "name", None),
                "namespace": getattr(self, "namespace", None),
                "resourceVersion": v_str,
                "uid": getattr(self, "uid", None)
            }

        def _hydrate(self, d: dict):
            """
            @override
            """
            if "name" in d:
                self.name = d["name"]
            if "namespace" in d:
                self.namespace = d["namespace"]
            if "resourceVersion" in d:
                self.resourceVersion = d["resourceVersion"]
            if "uid" in d:
                self.uid = d["uid"]

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @metadata.setter
    def metadata(self, m: Union[Metadata, tuple]):
        if type(m) is self.Metadata:
            self._metadata = m
        elif type(m) is tuple:
            self._metadata = self.Metadata(*m)
        else:
            raise TypeError(TYPE_ERROR.format(type(m), "metadata"))

    class Status(SerializationUtils):
        """
        Contains status information about the resource i.e. what the current state of the resource is

        CustomResource.status
        """
        def __init__(
                self,
                state: str = None,
                external_endpoint: str = None,
                log_search_dashboard: str = None,
                metrics_dashboard: str = None
        ):
            self.state = state
            self.externalEndpoint = external_endpoint
            self.log_search_dashboard = log_search_dashboard
            self.metrics_dashboard = metrics_dashboard

        @property
        def state(self) -> str:
            return self._state

        @state.setter
        def state(self, s: str):
            self._state = s

        @property
        def externalEndpoint(self) -> str:
            """
            What endpoint you can hit this resource at
            """
            return self._externalEndpoint

        @externalEndpoint.setter
        def externalEndpoint(self, e: str):
            self._externalEndpoint = e

        @property
        def log_search_dashboard(self) -> str:
            """
            The Log Search Dashboard endpoint for this custom resource.
            """
            return self._log_search_dashboard

        @log_search_dashboard.setter
        def log_search_dashboard(self, e: str):
            self._log_search_dashboard = e

        @property
        def metrics_dashboard(self) -> str:
            """
            The Metrics Dashboard endpoint for this custom resource.
            """
            return self._metrics_dashboard

        @metrics_dashboard.setter
        def metrics_dashboard(self, e: str):
            self._metrics_dashboard = e

        def _to_dict(self):
            """
            @override
            """
            return {
                "externalEndpoint": self.externalEndpoint,
                "logSearchDashboard": self.log_search_dashboard,
                "metricsDashboard": self.metrics_dashboard,
                "state": self.state
            }

        def _hydrate(self, d: dict):
            """
            @override
            """
            if "state" in d:
                self.state = d["state"]
            if "externalEndpoint" in d:
                self.externalEndpoint = d["externalEndpoint"]
            if "logSearchDashboard" in d:
                self.log_search_dashboard = d["logSearchDashboard"]
            if "metricsDashboard" in d:
                self.metrics_dashboard = d["metricsDashboard"]

    @property
    def status(self) -> Status:
        return self._status

    @status.setter
    def status(self, s: Union[Status, tuple]):
        if type(s) is self.Status:
            self._status = s
        elif type(s) is tuple:
            self._status = self.Status(*s)
        else:
            raise TypeError(TYPE_ERROR.format(type(s), "status"))

    class Spec(SerializationUtils):
        """
        A kubernetes required field that contains information about the desired state of this object
        @see https://kubernetes.io/docs/concepts/overview/working-with-objects/kubernetes-objects/#required-fields

        CustomResource.spec
        """
        def __init__(
                self,
                dev: bool = False,
                storage: 'Storage' = None,
                service: ServiceSpec = None,
                docker: DockerSpec = None):
            self.dev = dev
            self.storage = storage if storage else self.Storage()
            self.service = service if service else ServiceSpec()
            self.docker = docker if docker else DockerSpec()

        @property
        def dev(self) -> bool:
            """
            True if this is a dev object, false otherwise. Not a k8s thing, for internal use
            """
            return self._dev

        @dev.setter
        def dev(self, d: bool):
            self._dev = d

        @property
        def service(self) -> ServiceSpec:
            """
            @see ServiceSpec
            """
            return self._service

        @service.setter
        @enforcetype(ServiceSpec)
        def service(self, s: ServiceSpec):
            self._service = s

        class Storage(SerializationUtils):
            """
            Namespaces the properties of the storage spec for a custom resource. What volumes it should have, what type,
            and what size they are.

            CustomResource.spec.storage
            """
            def __init__(self, data: VolumeClaim = None, logs: VolumeClaim = None, backups: VolumeClaim = None):
                self.data = data if data else VolumeClaim()
                self.logs = logs if logs else VolumeClaim()
                self.backups = backups if backups else VolumeClaim()
                self.volumeClaimMounts = []

            @property
            def data(self) -> VolumeClaim:
                """
                The volume to hold application data
                @see VolumeClaim
                """
                return self._data

            @data.setter
            @enforcetype(VolumeClaim)
            def data(self, vc: VolumeClaim):
                self._data = vc

            @property
            def logs(self) -> VolumeClaim:
                """
                The volume to hold log data
                @see VolumeClaim
                """
                return self._logs

            @logs.setter
            @enforcetype(VolumeClaim)
            def logs(self, vc: VolumeClaim):
                self._logs = vc

            @property
            def backups(self) -> VolumeClaim:
                """
                The volume to hold backups
                @see VolumeClaim
                """
                return self._backups

            @backups.setter
            @enforcetype(VolumeClaim)
            def backups(self, vc):
                self._backups = vc

            @property
            def volumeClaimMounts(self) -> list:
                return self._volumeClaimMounts

            @volumeClaimMounts.setter
            def volumeClaimMounts(self, vcm: Union[list, str]):
                if type(vcm) is str:
                    self._volumeClaimMounts = self._parse_mount_string(vcm)
                elif type(vcm) is list:
                    self._volumeClaimMounts = vcm
                else:
                    raise ValueError(
                        "Type '{}' is incompatible with property volumeClaimMounts. Need 'str' or 'list'".format(type(vcm)))

            @staticmethod
            def _parse_mount_string(mount_str: str):
                def _map_mount(mount):
                    parts = mount.split(":")
                    if len(parts) < 2:
                        raise CliError(
                            "Volume claim mount must be in the form of [pvc-name]:[volume-type]:[optional-metadata]")
                    return {
                        "volumeClaimName": parts[0].strip(),
                        "volumeType": parts[1].strip()
                    }

                return [_map_mount(mount) for mount in mount_str.split(",")]

            def _hydrate(self, d: dict):
                """
                @override
                """
                if "data" in d:
                    self.data._hydrate(d["data"])
                if "logs" in d:
                    self.logs._hydrate(d["logs"])
                if "backups" in d:
                    self.backups._hydrate(d["backups"])
                if "volumeClaimMounts" in d:
                    self.volumeClaimMounts = d["volumeClaimMounts"]

            def _to_dict(self):
                """
                @override
                """
                return {
                    "data": self.data._to_dict(),
                    "logs": self.logs._to_dict(),
                    "backups": self.backups._to_dict(),
                    "volumeClaimMounts": self.volumeClaimMounts
                }

        @property
        def storage(self) -> Storage:
            """
            @see Storage
            """
            return self._storage

        @storage.setter
        @enforcetype(Storage)
        def storage(self, s: Storage):
            self._storage = s

        @property
        def docker(self) -> DockerSpec:
            """
            @see DockerSpec
            """
            return self._docker

        @docker.setter
        def docker(self, d: DockerSpec):
            self._docker = d

        def _hydrate(self, d: dict):
            """
            @override
            """
            if "dev" in d:
                self.dev = d["dev"]

            if "service" in d:
                self.service._hydrate(d["service"])

            if "storage" in d:
                self.storage._hydrate(d["storage"])

            if "docker" in d:
                self.docker._hydrate(d["docker"])

        def _to_dict(self):
            """
            @override
            """
            return {
                "dev": self.dev,
                "service": self.service._to_dict(),
                "storage": self.storage._to_dict(),
                "docker": self.docker._to_dict()
            }

    @property
    def spec(self) -> Spec:
        """
        @see Spec
        """
        return self._spec

    @spec.setter
    @enforcetype(Spec)
    def spec(self, s: Spec):
        self._spec = s

    def encode(self):
        """
        Serializes this object into a dictionary
        :return: Dictionary representation of this object
        """
        return prune_dict(self._to_dict())

    def encodes(self):
        """
        Serializes this objects into a JSON string
        :return: JSON string representing this object
        """
        return json.dumps(self.encode())

    @staticmethod
    def decode(cls: 'Type[CustomResource]', data: dict):
        """
        Takes JSON representation of this object and decodes into an object instance
        :param cls: The class to decode data into
        :param data: A dictionary used to hydrate a cls instance
        :returns: An instance of 'cls' hydrated by 'data'
        """
        if not issubclass(cls, CustomResource) and cls is not CustomResource:
            raise TypeError("\'{}\' is not a valid subclass of CustomResource".format(cls))

        ret = cls()
        ret._hydrate(trim_dict_entries(data))
        return ret

    @staticmethod
    def decodes(cls: 'CustomResource', data: str):
        """
        Takes a dictionary representation of this object and decodes into an object instance
        :param cls: The class to decode data into
        :param data: A JSON string used to hydrate a cls instance
        :return: An instance of 'cls' hydrated by 'data'
        """
        return CustomResource.decode(cls, json.loads(data))

    @validator
    def _validate_storage_classes(self, client: 'KubernetesClient'):
        """
        Ensures that all storage classes in the storage spec actually exist
        """
        STORAGE_CLASS_ERROR = "Storage class '{}' does not exist"

        logs = getattr(self.spec.storage, "logs", None)
        if logs and logs.className and not client.storage_class_exists(logs.className):
            raise ValueError(STORAGE_CLASS_ERROR.format(logs.className))

        data = getattr(self.spec.storage, "data", None)
        if data and data.className and not client.storage_class_exists(data.className):
            raise ValueError(STORAGE_CLASS_ERROR.format(data.className))

        backups = getattr(self.spec.storage, "backups", None)
        if backups and backups.className and not client.storage_class_exists(backups.className):
            raise ValueError(STORAGE_CLASS_ERROR.format(backups.className))

    @validator
    def _validate_volume_claim_mounts(self, client: 'KubernetesClient'):
        """
        Verifies various properties about volume claim mounts in the storage spec. Raises CliErrors if anything
        is amiss
        """
        # There can be at most one backup claim mount
        num_backups = sum(1 for m in self.spec.storage.volumeClaimMounts if m["volumeType"] == "backup")
        if num_backups > 1:
            raise CliError('Volume claim mount with backup volume type cannot be specified multiple times')

        # If there is a backup claim mount then should be no conflicting backup storage listed
        if num_backups == 1 and getattr(self.spec.storage.backups, "className", None) is not None:
            raise CliError('--storage-class-backups cannot be used with --volume-claim-mounts '
                           'when a volume claim mount is configured for backup purpose')

        # Check that all volume claims exist in this namespace
        for mount in self.spec.storage.volumeClaimMounts:
            if not client.persistent_volume_claim_exists(mount["volumeClaimName"], self.metadata.namespace):
                raise CliError('Persistent volume claim %s not found' % (mount['volumeClaimName']))

    def _hydrate(self, d: dict):
        """
        @override
        """
        if "apiVersion" in d:
            self.apiVersion = d["apiVersion"]

        if "kind" in d:
            self.kind = d["kind"]

        if "spec" in d:
            self.spec._hydrate(d["spec"])

        if "metadata" in d:
            self.metadata._hydrate(d["metadata"])

        if "status" in d:
            self.status._hydrate(d["status"])

    def _to_dict(self):
        """
        @override
        """
        return {
            "apiVersion": self.apiVersion,
            "kind": self.kind,
            "spec": self.spec._to_dict(),
            "metadata": self.metadata._to_dict(),
            "status": self.status._to_dict()
        }

    @abstractmethod
    def apply_args(self, **kwargs):
        self._set_if_provided(self.spec.storage.logs, "className", kwargs, "storage_class_logs")
        self._set_if_provided(self.spec.storage.data, "className", kwargs, "storage_class_data")
        self._set_if_provided(self.spec.storage.backups, "className", kwargs, "storage_class_backups")
        self._set_if_provided(self.spec.storage.logs, "size", kwargs, "storage_size_logs")
        self._set_if_provided(self.spec.storage.data, "size", kwargs, "storage_size_data")
        self._set_if_provided(self.spec.storage.backups, "size", kwargs, "storage_size_backups")
        self._set_if_provided(self.metadata, "name", kwargs, "name")
        self._set_if_provided(self.metadata, "namespace", kwargs, "namespace")
        self._set_if_provided(self.spec, "dev", kwargs, "dev")
        self._set_if_provided(self.spec.service, "port", kwargs, "port")
        self._set_if_provided(self.spec.storage, "volumeClaimMounts", kwargs, "volume_claim_mounts")

    @staticmethod
    def _set_if_provided(obj, key, args, args_key):
        if args_key in args and args[args_key] is not None:
            setattr(obj, key, args[args_key])