# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.constants import (
    DNS_NAME_REQUIREMENTS)
from azdata.cli.core.models.custom_resource import CustomResource
from azdata.cli.core.util import name_meets_dns_requirements
from azdata.cli.core.models.dict_utils import SerializationUtils
from azdata.cli.core.models.data_controller_volume import DataControllerVolume
from azdata.cli.core.models.security_spec import SecuritySpec
from azdata.cli.core.class_utils import enforcetype
from azdata.cli.core.models.endpoint_spec import EndpointSpec
from azdata.cli.commands.arc import data_controller_properties


class DataControllerCustomResource(CustomResource):
    """
    Internal Custom Resource object to be used for deployments.
    """

    def __init__(self):
        super().__init__()

    class Metadata(CustomResource.Metadata):
        """
        @override CustomResource.Metadata
        """
        arc_dc_name_max_length = 253
        arc_dc_namespace_max_length = 63

        def init(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @CustomResource.Metadata.name.setter
        def name(self, n: str):
            """
            @override
            """
            if not n:
                raise ValueError("Arc data controller name cannot be empty")

            if len(n) > self.arc_dc_name_max_length:
                raise ValueError("Arc data controller name '{}' exceeds {} character length limit".format(n, self.arc_dc_name_max_length))

            if not name_meets_dns_requirements(n):
                raise ValueError("Arc data controller name '{}' does not follow DNS requirements: {}".format(n, DNS_NAME_REQUIREMENTS))

            self._name = n

        @CustomResource.Metadata.namespace.setter
        def namespace(self, ns: str):
            """
            @override
            """
            if not ns:
                raise ValueError("Kubernetes namespace cannot be empty")

            if len(ns) > self.arc_dc_namespace_max_length:
                raise ValueError("Kubernetes namespace '{}' exceeds {} character name length limit".format(ns, self.arc_dc_namespace_max_length))

            if not name_meets_dns_requirements(ns):
                raise ValueError("Kubernetes namespace '{}' does not follow DNS requirements: {}".format(ns, DNS_NAME_REQUIREMENTS))

            self._namespace = ns
    
    class Spec(CustomResource.Spec):
        """
        @override CustomResource.Spec
        """
        def __init__(self):
            super().__init__()
            self.services = []
            self.settings = {}
            self.security = SecuritySpec()
            self.credentials = self.Credentials()

        class Credentials(SerializationUtils):
            """
            Contains "secret"/connection information for this data controller to communicate with a variety of resources

            CustomResource.Spec.Credentials
            """
            def __init__(
                    self,
                    domainServiceAccount: str = None,
                    serviceAccount: str = None,
                    dockerRegistry: str = None,
                    controllerAdmin: str = None):

                self.controllerAdmin = controllerAdmin
                self.domainServiceAccount = domainServiceAccount
                self.serviceAccount = serviceAccount
                self.dockerRegistry = dockerRegistry

            @property
            def domainServiceAccount(self) -> str:
                return self._domainServiceAccount

            @domainServiceAccount.setter
            def domainServiceAccount(self, dsa: str):
                self._domainServiceAccount = dsa

            @property
            def serviceAccount(self) -> str:
                return self._serviceAccount

            @serviceAccount.setter
            def serviceAccount(self, sa: str):
                self._serviceAccount = sa

            @property
            def dockerRegistry(self) -> str:
                return self._dockerRegistry

            @dockerRegistry.setter
            def dockerRegistry(self, dr: str):
                self._dockerRegistry = dr

            @property
            def controllerAdmin(self) -> str:
                return self._controllerAdmin

            @controllerAdmin.setter
            def controllerAdmin(self, ca: str):
                self._controllerAdmin = ca

            def _to_dict(self) -> dict:
                return {
                    "domainServiceAccount": getattr(self, "domainServiceAccount", None),
                    "serviceAccount": getattr(self, "serviceAccount", None),
                    "dockerRegistry": getattr(self, "dockerRegistry", None),
                    "controllerAdmin": getattr(self, "controllerAdmin", None)
                }

            def _hydrate(self, d: dict):
                if "domainServiceAccount" in d:
                    self.domainServiceAccount = d["domainServiceAccount"]

                if "controllerAdmin" in d:
                    self.controllerAdmin = d["controllerAdmin"]

                if "dockerRegistry" in d:
                    self.dockerRegistry = d["dockerRegistry"]

                if "serviceAccount" in d:
                    self.serviceAccount = d["serviceAccount"]

        @property
        def credentials(self) -> Credentials:
            return self._credentials

        @credentials.setter
        @enforcetype(Credentials)
        def credentials(self, c: Credentials):
            self._credentials = c

        class Storage(CustomResource.Spec.Storage):
            """
            @override CustomResource.Spec.Storage
            """
            def __init__(self):
                super().__init__(DataControllerVolume(), DataControllerVolume(), DataControllerVolume())

        @property
        def security(self) -> SecuritySpec:
            return self._security

        @security.setter
        @enforcetype(SecuritySpec)
        def security(self, s: SecuritySpec):
            self._security = s

        @property
        def services(self) -> list:
            return self._services

        @services.setter
        def services(self, s: list):
            self._services = s

        @property
        def settings(self):
            return self._settings

        @settings.setter
        def settings(self, s):
            self._settings = s

        def _to_dict(self):
            """
            @override
            """
            base = super()._to_dict()
            services = []
            for s in getattr(self, "services", []):
                services.append(s._to_dict())

            base["credentials"] = self.credentials._to_dict()
            base["services"] = services
            base["settings"] = getattr(self, "settings", None)
            base["security"] = self.security._to_dict()
            return base

        def _hydrate(self, d: dict):
            """
            @override
            """
            super()._hydrate(d)
            if "services" in d and d["services"] is not None:
                for s in d["services"]:
                    curr = EndpointSpec()
                    curr._hydrate(s)
                    self.services.append(curr)

            if "security" in d:
                self.security._hydrate(d["security"])

            if "settings" in d:
                self.settings = d["settings"]

            if "credentials" in d:
                self.credentials._hydrate(d["credentials"])

    class Status(CustomResource.Status):
        """
        CustomResource.Status
        """
        @property
        def phase(self):
            return self._phase

        @phase.setter
        def phase(self, p):
            self._phase = p

    def _to_dict(self) -> dict:
        return super()._to_dict()

    def _hydrate(self, d: dict):
        super()._hydrate(d)

    def get_controller_service(self) -> EndpointSpec:
        for s in self.spec.services:
            try:
                if s.name == "controller":
                    return s
            except AttributeError:
                continue

        return None

    def apply_args(self, **kwargs):
        """
        @override
        """
        super().apply_args(**kwargs)

        # Construct azure settings based on args
        self.spec.settings["azure"] = {
            data_controller_properties.CONNECTION_MODE: kwargs["connectivity_mode"].lower(),
            data_controller_properties.LOCATION: kwargs["location"].lower(),
            data_controller_properties.RESOURCE_GROUP: kwargs["resource_group"],
            data_controller_properties.SUBSCRIPTION: kwargs["subscription"]
        }

        # set the display name
        if "name" in kwargs:
            if "controller" not in self.spec.settings:
                self.spec.settings["controller"] = {}
            self.spec.settings["controller"][data_controller_properties.DISPLAY_NAME] = kwargs["name"]

        if "storage_class" in kwargs and kwargs["storage_class"] is not None:
            self.spec.storage.data.className = kwargs["storage_class"]
            self.spec.storage.logs.className = kwargs["storage_class"]
