# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.configuration import Configuration
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.util import (merge, singleton)
from azdata.cli.core.events import EVENT_CMD_OUTPUT
from knack.output import OutputProducer as KnackOutputProducer
from knack.util import CommandResultItem
from io import TextIOWrapper, BytesIO
from collections import OrderedDict

try:
    from types import SimpleNamespace as Namespace
except ImportError:
    # Python 2.x fallback
    from argparse import Namespace

import sys

__all__ = ['unicode', 'OutputProducer', 'OutputStream', 'IOBuffer']

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class _Unicode(Namespace):
    """Unicode catalog for output."""

    def __init__(self):
        def shim(code, fallback):
            import os
            return fallback if os.name == 'nt' else code

        super(_Unicode, self).__init__(**{
            'check': shim('\u2713', '-')
            # Add more when needed.
        })
unicode = _Unicode()

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


def structured_stdout(fn):
    def fn_wrapper(obj):
        from azdata.cli.core.logging import get_logs
        io = OutputStream().flush()
        obj.result = {
            'result': obj.result,
            'stdout': io[0],
            'stderr': io[1],
            'logs': get_logs()
        }

        return fn(obj)

    return fn_wrapper


def structured_stderr(fn):
    def fn_wrapper(ex, traceback=False, status=1):
        from azdata.cli.core.logging import get_logs
        ex = {
            'stderr': str(ex),
            'code': status,
            'logs': get_logs()
        }

        if traceback:
            from azdata.cli.core.util import traceback
            ex['tracback'] = '\n'.join(traceback())

        return fn(ex)
    return fn_wrapper


@structured_stdout
def format_json(obj):
    from knack.output import format_json
    return format_json(obj)


@structured_stdout
def format_json_color(obj):
    from knack.output import format_json_color
    return format_json_color(obj)


@structured_stdout
def format_table(obj):
    from knack.output import format_table
    tbl = []
    out = [{'stdout': out} for out in obj.result['stdout']]
    err = [{'stderr': err} for err in obj.result['stderr']]
    logs = [{'logs': log} for log in obj.result['logs']]

    if out:
        tbl.append(format_table(CommandResultItem(out, is_query_active=True)))
    if err:
        tbl.append(format_table(CommandResultItem(err, is_query_active=True)))
    if logs:
        tbl.append(format_table(CommandResultItem(logs, is_query_active=True)))
    if obj.result['result']:
        obj.result = obj.result['result']

    tbl.append(format_table(obj))
    return '\n'.join(tbl)


@structured_stdout
def format_tsv(obj):
    from knack.output import format_tsv
    tsv = []
    out = [{'stdout': out} for out in obj.result['stdout']]
    err = [{'stderr': err} for err in obj.result['stderr']]
    logs = [{'logs': log} for log in obj.result['logs']]

    if out:
        tsv.append(format_tsv(CommandResultItem(out, is_query_active=True)))
    if err:
        tsv.append(format_tsv(CommandResultItem(err, is_query_active=True)))
    if logs:
        tsv.append(format_tsv(CommandResultItem(logs, is_query_active=True)))
    if obj.result['result']:
        obj.result = obj.result['result']

    tsv.append(format_tsv(obj))
    return '\n'.join(tsv)


@structured_stderr
def format_error_json(ex):
    from knack.output import format_json
    return format_json(CommandResultItem(ex, is_query_active=True))


@structured_stderr
def format_error_table(ex):
    from knack.output import format_table
    table = []
    logs = [{'logs': log} for log in ex['logs']]
    ex.pop('logs', None)
    table.append(format_table(CommandResultItem(ex, is_query_active=True)))
    table.append(format_table(CommandResultItem(logs, is_query_active=True)))
    return '\n'.join(table)


@structured_stderr
def format_error_tsv(ex):
    from knack.output import format_tsv

    tsv = []
    logs = [{'logs': log} for log in ex['logs']]
    ex.pop('logs', None)
    tsv.append(format_tsv(CommandResultItem(ex, is_query_active=True)))
    tsv.append(format_tsv(CommandResultItem(logs, is_query_active=True)))
    return '\n'.join(tsv)


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class OutputProducer(KnackOutputProducer):
    _FORMATTER = merge(KnackOutputProducer._FORMAT_DICT, {
        'structured_json': format_json,
        'structured_jsonc': format_json_color,
        'structured_table': format_table,
        'structured_tsv': format_tsv
    })
    """Formatter"""

    _ERROR_FORMATTER = {
        'json': format_error_json,
        'table': format_error_table,
        'tsv': format_error_tsv
    }

    structured_output = False
    """
    """

    def __init__(self, cli_ctx=None):
        """
        Manages the production of output from the result of a command
        invocation.

        :param cli_ctx: CLI Context
        :type cli_ctx: knack.cli.CLI
        """
        super(OutputProducer, self).__init__(cli_ctx=cli_ctx)
        self.cli_ctx.register_event(EVENT_CMD_OUTPUT, lambda x: OutputStream())

    def get_formatter(self, format_type):
        invocation = self.cli_ctx.invocation.data
        cmd = invocation['command']
        structured = OutputProducer.structured_output

        # Set structured output to override if the given format type is listed
        override_output_types = invocation.get('override_output', {}).get(invocation['command'], [])
        override_output = True if format_type in override_output_types else False
        
        format_type = 'structured_' + format_type if structured else format_type
        formatter = OutputProducer._FORMATTER[format_type]
        custom_type = invocation.get('custom', False)

        # map current command to it's custom-type if it has one
        if custom_type:
            custom_type = cmd in custom_type and custom_type[cmd]

        # Custom output takes precedence over default, replace with custom
        # Structured output takes precedence over custom output
        if custom_type and (not structured or (structured and override_output)):
            custom_type = invocation['custom'].get(invocation['command'])
            is_fn = callable(custom_type)

            if is_fn:
                if structured and override_output:
                    OutputStream().flush()
                formatter = custom_type
            else:  # must be json|table|tsv
                if custom_type not in OutputProducer._FORMATTER:
                    raise ValueError('Output format not supported "{}".'.
                                     format(custom_type))
                formatter = OutputProducer._FORMATTER[custom_type]

        return formatter

    def error(self, ex, format_type):
        """
        The default exception handler
        """
        from knack.util import CLIError as KnackCliError
        from azdata.cli.core.logging import get_logger

        logger = get_logger(__name__)

        if OutputProducer.structured_output:
            formatter = self._ERROR_FORMATTER[format_type]
            if isinstance(ex, CliError):
                ex = formatter(ex, status=ex.status)
            elif isinstance(ex, KnackCliError):
                ex = formatter(ex)
            else:
                ex = formatter(ex, traceback=True)

            logger.error(ex)
        else:
            if isinstance(ex, KnackCliError):
                logger.error(ex)
            else:
                logger.exception(ex)

        return ex

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


@singleton
class OutputStream(object):
    __ORIGINAL_STDOUT__ = sys.stdout
    """
    """

    __ORIGINAL_STDERR__ = sys.stderr
    """
    """

    terminator = '\n'

    def __init__(self):
        if OutputProducer.structured_output:
            sys.stdout = stdout = IOBuffer(BytesIO(), sys.stdout.encoding)
            sys.stderr = stderr = IOBuffer(BytesIO(), sys.stderr.encoding)
        else:
            stdout = Stdout()
            stderr = Stderr()

        self._stdout = stdout
        self._stderr = stderr

    @property
    def stdout(self):
        return self._stdout

    @property
    def stderr(self):
        return self._stderr

    def flush(self):
        # -- capture any buffered stdout --
        sys.stdout.seek(0)
        stdout = sys.stdout.read().rstrip().split(self.terminator)
        sys.stdout = self.__ORIGINAL_STDOUT__
        sys.stdout.flush()
        self._stdout = [] if len(stdout) == 1 and stdout[0] == '' else stdout

        # -- capture any buffered stderr --
        sys.stderr.seek(0)
        stderr = sys.stderr.read().rstrip().split(self.terminator)
        sys.stderr = self.__ORIGINAL_STDERR__
        sys.stderr.flush()
        self._stderr = [] if len(stderr) == 1 and stderr[0] == '' else stderr

        return self._stdout, self._stderr

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class IOBuffer(TextIOWrapper):
    terminator = '\n'

    def write(self, obj):
        """
        Write string to stream.
        :param obj: The item written to the stream as a string.
        :return: Returns the number of characters written (which is always
                 equal to the length of the string).
        """
        from knack.output import (format_table, format_json)
        try:
            if isinstance(obj, list):
                obj = format_table(CommandResultItem(obj, is_query_active=True))
            elif isinstance(obj, dict) or isinstance(obj, OrderedDict):
                obj = format_json(CommandResultItem(obj, is_query_active=True))
            else:  # primitives
                obj = str(obj).rstrip()

            if obj:
                return super(IOBuffer, self).write(obj + self.terminator)
        except TypeError:
            # redirect encoded byte strings directly to buffer
            return super(IOBuffer, self).buffer.write(obj)

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class StdIO(str):
    """
    StdIO represents the base IOSink for `stdout` and `stderr`. It provides a
    blocking IOSink so using this to write will block until the output is
    written.
    """

    terminator = '\n'

    def __new__(cls, io=sys.stdout, color=None):
        obj = super(StdIO, cls).__new__(cls)
        obj.io = io
        obj.color = color
        return obj

    @property
    def lines(self):
        return self.split(self.terminator)

    @property
    def qlines(self):
        return [line.split() for line in self.split(self.terminator)]

    def write(self, obj, color=None, end='\n'):
        """
         A display stdout or stderro print hook equipped with a unicode
         fallback hook.

         Pseudo code originating from:
         https://docs.python.org/3/library/sys.html#sys.displayhook

         :param obj: The text string to write to stdout or stderr.
         :param color: Optional text color.
         :param end: Include an ending newline to stdout or stderr.
         """
        import colorama
        from knack.output import (format_json, format_table)

        def color_output(msg, c):
            enable_color = False
            try:
                # Color if tty stream available
                enable_color = (self.color or c) and sys.stdin.isatty()
            except AttributeError:
                pass

            if enable_color:
                try:
                    c = c or self.color
                    msg = '{}{}{}'.format(c, obj, colorama.Style.RESET_ALL)
                except KeyError:
                    pass

            return msg

        try:
            if isinstance(obj, list):
                obj = format_table(CommandResultItem(obj, is_query_active=True))
            elif isinstance(obj, dict) or isinstance(obj, OrderedDict):
                obj = format_json(CommandResultItem(obj, is_query_active=True))
            else:  # primitives
                obj = str(obj)

            self.io.write(color_output(obj, color))
        except ValueError:
            # -- pytest --
            sys.stdout.write(obj)
            if end:
                sys.stdout.write(end)
            return
        except UnicodeEncodeError:
            # Fallback shim to ascii with replace with backslashed escape
            # sequences
            bytes_string = obj.encode(self.io.encoding, 'backslashreplace')
            if hasattr(self.io, 'buffer'):
                self.io.buffer.write(bytes_string)
            else:
                self.io.write(bytes_string.decode(self.io.encoding, 'strict'))
        if end:
            self.io.write(end)

        self.io.flush()


class Stdout(StdIO):
    def __new__(cls):
        return super(Stdout, cls).__new__(cls, io=sys.stdout)


class Stderr(StdIO):
    def __new__(cls):
        from colorama import Fore
        return super(Stderr, cls).__new__(cls, io=sys.stderr, color=Fore.RED)



